@@ -21,7 +21,6 @@ import (
 	"path"
 	"regexp"
 	"strings"
-	"sync/atomic"
 	"time"
 
 	"github.com/ncw/go-acd"
@@ -31,7 +30,6 @@ import (
 	"github.com/ncw/rclone/pacer"
 	"github.com/ncw/rclone/rest"
 	"github.com/pkg/errors"
-	"github.com/spf13/pflag"
 	"golang.org/x/oauth2"
 )
 
@@ -51,7 +49,7 @@ const (
 var (
 	// Flags
 	tempLinkThreshold = fs.SizeSuffix(9 << 30) // Download files bigger than this via the tempLink
-	uploadWaitPerGB   = pflag.DurationP("acd-upload-wait-per-gb", "", 180*time.Second, "Additional time per GB to wait after a failed complete upload to see if it appears.")
+	uploadWaitPerGB   = fs.DurationP("acd-upload-wait-per-gb", "", 180*time.Second, "Additional time per GB to wait after a failed complete upload to see if it appears.")
 	// Description of how to auth for this app
 	acdConfig = &oauth2.Config{
 		Scopes: []string{"clouddrive:read_all", "clouddrive:write"},
@@ -85,20 +83,20 @@ func init() {
 			Help: "Amazon Application Client Secret - leave blank normally.",
 		}},
 	})
-	pflag.VarP(&tempLinkThreshold, "acd-templink-threshold", "", "Files >= this size will be downloaded via their tempLink.")
+	fs.VarP(&tempLinkThreshold, "acd-templink-threshold", "", "Files >= this size will be downloaded via their tempLink.")
 }
 
 // Fs represents a remote acd server
 type Fs struct {
-	name         string                 // name of this remote
-	c            *acd.Client            // the connection to the acd server
-	noAuthClient *http.Client           // unauthenticated http client
-	root         string                 // the path we are working on
-	dirCache     *dircache.DirCache     // Map of directory path to directory id
-	pacer        *pacer.Pacer           // pacer for API calls
-	ts           *oauthutil.TokenSource // token source for oauth
-	uploads      int32                  // number of uploads in progress - atomic access required
-	trueRootID   string                 // ID of true root directory
+	name         string             // name of this remote
+	features     *fs.Features       // optional features
+	c            *acd.Client        // the connection to the acd server
+	noAuthClient *http.Client       // unauthenticated http client
+	root         string             // the path we are working on
+	dirCache     *dircache.DirCache // Map of directory path to directory id
+	pacer        *pacer.Pacer       // pacer for API calls
+	trueRootID   string             // ID of true root directory
+	tokenRenewer *oauthutil.Renew   // renew the token on expiry
 }
 
 // Object describes a acd object
@@ -127,6 +125,11 @@ func (f *Fs) String() string {
 	return fmt.Sprintf("amazon drive root '%s'", f.root)
 }
 
+// Features returns the optional features of this Fs
+func (f *Fs) Features() *fs.Features {
+	return f.features
+}
+
 // Pattern to match a acd path
 var matcher = regexp.MustCompile(`^([^/]*)(.*)$`)
 
@@ -153,15 +156,15 @@ var retryErrorCodes = []int{
 func (f *Fs) shouldRetry(resp *http.Response, err error) (bool, error) {
 	if resp != nil {
 		if resp.StatusCode == 401 {
-			f.ts.Invalidate()
-			fs.Log(f, "401 error received - invalidating token")
+			f.tokenRenewer.Invalidate()
+			fs.Debugf(f, "401 error received - invalidating token")
 			return true, err
 		}
 		// Work around receiving this error sporadically on authentication
 		//
 		// HTTP code 403: "403 Forbidden", reponse body: {"message":"Authorization header requires 'Credential' parameter. Authorization header requires 'Signature' parameter. Authorization header requires 'SignedHeaders' parameter. Authorization header requires existence of either a 'X-Amz-Date' or a 'Date' header. Authorization=Bearer"}
 		if resp.StatusCode == 403 && strings.Contains(err.Error(), "Authorization header requires") {
-			fs.Log(f, "403 \"Authorization header requires...\" error received - retry")
+			fs.Debugf(f, "403 \"Authorization header requires...\" error received - retry")
 			return true, err
 		}
 	}
@@ -183,8 +186,8 @@ func NewFs(name, root string) (fs.Fs, error) {
 		c:            c,
 		pacer:        pacer.New().SetMinSleep(minSleep).SetPacer(pacer.AmazonCloudDrivePacer),
 		noAuthClient: fs.Config.Client(),
-		ts:           ts,
 	}
+	f.features = (&fs.Features{CaseInsensitive: true, ReadMimeType: true}).Fill(f)
 
 	// Update endpoints
 	var resp *http.Response
@@ -204,7 +207,10 @@ func NewFs(name, root string) (fs.Fs, error) {
 	f.trueRootID = *rootInfo.Id
 
 	// Renew the token in the background
-	go f.renewToken()
+	f.tokenRenewer = oauthutil.NewRenew(f.String(), ts, func() error {
+		_, err := f.getRootInfo()
+		return err
+	})
 
 	f.dirCache = dircache.New(root, f.trueRootID, f)
 
@@ -246,39 +252,6 @@ func (f *Fs) getRootInfo() (rootInfo *acd.Folder, err error) {
 	return rootInfo, err
 }
 
-// Renew the token - runs in the background
-//
-// Renews the token whenever it expires.  Useful when there are lots
-// of uploads in progress and the token doesn't get renewed.  Amazon
-// seem to cancel your uploads if you don't renew your token for 2hrs.
-func (f *Fs) renewToken() {
-	expiry := f.ts.OnExpiry()
-	for {
-		<-expiry
-		uploads := atomic.LoadInt32(&f.uploads)
-		if uploads != 0 {
-			fs.Debug(f, "Token expired - %d uploads in progress - refreshing", uploads)
-			// Do a transaction
-			_, err := f.getRootInfo()
-			if err == nil {
-				fs.Debug(f, "Token refresh successful")
-			} else {
-				fs.ErrorLog(f, "Token refresh failed: %v", err)
-			}
-		} else {
-			fs.Debug(f, "Token expired but no uploads in progress - doing nothing")
-		}
-	}
-}
-
-func (f *Fs) startUpload() {
-	atomic.AddInt32(&f.uploads, 1)
-}
-
-func (f *Fs) stopUpload() {
-	atomic.AddInt32(&f.uploads, -1)
-}
-
 // Return an Object from a path
 //
 // If it can't be found it returns the error fs.ErrorObjectNotFound.
@@ -307,7 +280,7 @@ func (f *Fs) NewObject(remote string) (fs.Object, error) {
 
 // FindLeaf finds a directory of name leaf in the folder with ID pathID
 func (f *Fs) FindLeaf(pathID, leaf string) (pathIDOut string, found bool, err error) {
-	//fs.Debug(f, "FindLeaf(%q, %q)", pathID, leaf)
+	//fs.Debugf(f, "FindLeaf(%q, %q)", pathID, leaf)
 	folder := acd.FolderFromId(pathID, f.c.Nodes)
 	var resp *http.Response
 	var subFolder *acd.Folder
@@ -317,18 +290,18 @@ func (f *Fs) FindLeaf(pathID, leaf string) (pathIDOut string, found bool, err er
 	})
 	if err != nil {
 		if err == acd.ErrorNodeNotFound {
-			//fs.Debug(f, "...Not found")
+			//fs.Debugf(f, "...Not found")
 			return "", false, nil
 		}
-		//fs.Debug(f, "...Error %v", err)
+		//fs.Debugf(f, "...Error %v", err)
 		return "", false, err
 	}
 	if subFolder.Status != nil && *subFolder.Status != statusAvailable {
-		fs.Debug(f, "Ignoring folder %q in state %q", leaf, *subFolder.Status)
+		fs.Debugf(f, "Ignoring folder %q in state %q", leaf, *subFolder.Status)
 		time.Sleep(1 * time.Second) // FIXME wait for problem to go away!
 		return "", false, nil
 	}
-	//fs.Debug(f, "...Found(%q, %v)", *subFolder.Id, leaf)
+	//fs.Debugf(f, "...Found(%q, %v)", *subFolder.Id, leaf)
 	return *subFolder.Id, true, nil
 }
 
@@ -424,7 +397,7 @@ func (f *Fs) listAll(dirID string, title string, directoriesOnly bool, filesOnly
 
 // ListDir reads the directory specified by the job into out, returning any more jobs
 func (f *Fs) ListDir(out fs.ListOpts, job dircache.ListDirJob) (jobs []dircache.ListDirJob, err error) {
-	fs.Debug(f, "Reading %q", job.Path)
+	fs.Debugf(f, "Reading %q", job.Path)
 	maxTries := fs.Config.LowLevelRetries
 	for tries := 1; tries <= maxTries; tries++ {
 		_, err = f.listAll(job.DirID, "", false, false, func(node *acd.Node) bool {
@@ -460,7 +433,7 @@ func (f *Fs) ListDir(out fs.ListOpts, job dircache.ListDirJob) (jobs []dircache.
 			return false
 		})
 		if fs.IsRetryError(err) {
-			fs.Debug(f, "Directory listing error for %q: %v - low level retry %d/%d", job.Path, err, tries, maxTries)
+			fs.Debugf(f, "Directory listing error for %q: %v - low level retry %d/%d", job.Path, err, tries, maxTries)
 			continue
 		}
 		if err != nil {
@@ -468,7 +441,7 @@ func (f *Fs) ListDir(out fs.ListOpts, job dircache.ListDirJob) (jobs []dircache.
 		}
 		break
 	}
-	fs.Debug(f, "Finished reading %q", job.Path)
+	fs.Debugf(f, "Finished reading %q", job.Path)
 	return jobs, err
 }
 
@@ -515,13 +488,13 @@ func (f *Fs) checkUpload(resp *http.Response, in io.Reader, src fs.ObjectInfo, i
 	buf := make([]byte, 1)
 	n, err := in.Read(buf)
 	if !(n == 0 && err == io.EOF) {
-		fs.Debug(src, "Upload error detected but didn't finish upload: %v (%q)", inErr, httpStatus)
+		fs.Debugf(src, "Upload error detected but didn't finish upload: %v (%q)", inErr, httpStatus)
 		return false, inInfo, inErr
 	}
 
 	// Don't wait for uploads - assume they will appear later
 	if *uploadWaitPerGB <= 0 {
-		fs.Debug(src, "Upload error detected but waiting disabled: %v (%q)", inErr, httpStatus)
+		fs.Debugf(src, "Upload error detected but waiting disabled: %v (%q)", inErr, httpStatus)
 		return false, inInfo, inErr
 	}
 
@@ -532,27 +505,27 @@ func (f *Fs) checkUpload(resp *http.Response, in io.Reader, src fs.ObjectInfo, i
 	const sleepTime = 5 * time.Second                        // sleep between tries
 	retries := int((timeToWait + sleepTime - 1) / sleepTime) // number of retries, rounded up
 
-	fs.Debug(src, "Error detected after finished upload - waiting to see if object was uploaded correctly: %v (%q)", inErr, httpStatus)
+	fs.Debugf(src, "Error detected after finished upload - waiting to see if object was uploaded correctly: %v (%q)", inErr, httpStatus)
 	remote := src.Remote()
 	for i := 1; i <= retries; i++ {
 		o, err := f.NewObject(remote)
 		if err == fs.ErrorObjectNotFound {
-			fs.Debug(src, "Object not found - waiting (%d/%d)", i, retries)
+			fs.Debugf(src, "Object not found - waiting (%d/%d)", i, retries)
 		} else if err != nil {
-			fs.Debug(src, "Object returned error - waiting (%d/%d): %v", i, retries, err)
+			fs.Debugf(src, "Object returned error - waiting (%d/%d): %v", i, retries, err)
 		} else {
 			if src.Size() == o.Size() {
-				fs.Debug(src, "Object found with correct size %d after waiting (%d/%d) - %v - returning with no error", src.Size(), i, retries, sleepTime*time.Duration(i-1))
+				fs.Debugf(src, "Object found with correct size %d after waiting (%d/%d) - %v - returning with no error", src.Size(), i, retries, sleepTime*time.Duration(i-1))
 				info = &acd.File{
 					Node: o.(*Object).info,
 				}
 				return true, info, nil
 			}
-			fs.Debug(src, "Object found but wrong size %d vs %d - waiting (%d/%d)", src.Size(), o.Size(), i, retries)
+			fs.Debugf(src, "Object found but wrong size %d vs %d - waiting (%d/%d)", src.Size(), o.Size(), i, retries)
 		}
 		time.Sleep(sleepTime)
 	}
-	fs.Debug(src, "Giving up waiting for object - returning original error: %v (%q)", inErr, httpStatus)
+	fs.Debugf(src, "Giving up waiting for object - returning original error: %v (%q)", inErr, httpStatus)
 	return false, inInfo, inErr
 }
 
@@ -580,21 +553,21 @@ func (f *Fs) Put(in io.Reader, src fs.ObjectInfo) (fs.Object, error) {
 		return nil, err
 	}
 	// If not create it
-	leaf, directoryID, err := f.dirCache.FindPath(remote, true)
+	leaf, directoryID, err := f.dirCache.FindRootAndPath(remote, true)
 	if err != nil {
 		return nil, err
 	}
 	if size > warnFileSize {
-		fs.Debug(f, "Warning: file %q may fail because it is too big. Use --max-size=%dM to skip large files.", remote, warnFileSize>>20)
+		fs.Logf(f, "Warning: file %q may fail because it is too big. Use --max-size=%dM to skip large files.", remote, warnFileSize>>20)
 	}
 	folder := acd.FolderFromId(directoryID, o.fs.c.Nodes)
 	var info *acd.File
 	var resp *http.Response
 	err = f.pacer.CallNoRetry(func() (bool, error) {
 		start := time.Now()
-		f.startUpload()
+		f.tokenRenewer.Start()
 		info, resp, err = folder.Put(in, leaf)
-		f.stopUpload()
+		f.tokenRenewer.Stop()
 		var ok bool
 		ok, info, err = f.checkUpload(resp, in, src, info, err, time.Since(start))
 		if ok {
@@ -634,7 +607,7 @@ func (f *Fs) Move(src fs.Object, remote string) (fs.Object, error) {
 	//  go test -v -run '^Test(Setup|Init|FsMkdir|FsPutFile1|FsPutFile2|FsUpdateFile1|FsMove)$'
 	srcObj, ok := src.(*Object)
 	if !ok {
-		fs.Debug(src, "Can't move - not same remote type")
+		fs.Debugf(src, "Can't move - not same remote type")
 		return nil, fs.ErrorCantMove
 	}
 
@@ -651,16 +624,36 @@ func (f *Fs) Move(src fs.Object, remote string) (fs.Object, error) {
 	if err != nil {
 		return nil, err
 	}
-	dstInfo, err := f.moveNode(srcObj.remote, dstLeaf, dstDirectoryID, srcObj.info, srcLeaf, srcDirectoryID, false)
+	err = f.moveNode(srcObj.remote, dstLeaf, dstDirectoryID, srcObj.info, srcLeaf, srcDirectoryID, false)
 	if err != nil {
 		return nil, err
 	}
-	dstObj := &Object{
-		fs:     f,
-		remote: remote,
-		info:   dstInfo,
+	// Wait for directory caching so we can no longer see the old
+	// object and see the new object
+	time.Sleep(200 * time.Millisecond) // enough time 90% of the time
+	var (
+		dstObj         fs.Object
+		srcErr, dstErr error
+	)
+	for i := 1; i <= fs.Config.LowLevelRetries; i++ {
+		_, srcErr = srcObj.fs.NewObject(srcObj.remote) // try reading the object
+		if srcErr != nil && srcErr != fs.ErrorObjectNotFound {
+			// exit if error on source
+			return nil, srcErr
+		}
+		dstObj, dstErr = f.NewObject(remote)
+		if dstErr != nil && dstErr != fs.ErrorObjectNotFound {
+			// exit if error on dst
+			return nil, dstErr
+		}
+		if srcErr == fs.ErrorObjectNotFound && dstErr == nil {
+			// finished if src not found and dst found
+			break
+		}
+		fs.Debugf(src, "Wait for directory listing to update after move %d/%d", i, fs.Config.LowLevelRetries)
+		time.Sleep(1 * time.Second)
 	}
-	return dstObj, nil
+	return dstObj, dstErr
 }
 
 // DirCacheFlush resets the directory cache - used in testing as an
@@ -669,74 +662,112 @@ func (f *Fs) DirCacheFlush() {
 	f.dirCache.ResetRoot()
 }
 
-// DirMove moves src directory to this remote using server side move
-// operations.
+// DirMove moves src, srcRemote to this remote at dstRemote
+// using server side move operations.
 //
 // Will only be called if src.Fs().Name() == f.Name()
 //
 // If it isn't possible then return fs.ErrorCantDirMove
 //
 // If destination exists then return fs.ErrorDirExists
-func (f *Fs) DirMove(src fs.Fs) (err error) {
-	// go test -v -run '^Test(Setup|Init|FsMkdir|FsPutFile1|FsPutFile2|FsUpdateFile1|FsDirMove)$
+func (f *Fs) DirMove(src fs.Fs, srcRemote, dstRemote string) (err error) {
 	srcFs, ok := src.(*Fs)
 	if !ok {
-		fs.Debug(src, "DirMove error: not same remote type")
+		fs.Debugf(src, "DirMove error: not same remote type")
 		return fs.ErrorCantDirMove
 	}
-
-	// Check if destination exists
-	if f.dirCache.FoundRoot() {
-		fs.Debug(src, "DirMove error: destination exists")
-		return fs.ErrorDirExists
-	}
+	srcPath := path.Join(srcFs.root, srcRemote)
+	dstPath := path.Join(f.root, dstRemote)
 
 	// Refuse to move to or from the root
-	if f.root == "" || srcFs.root == "" {
-		fs.Debug(src, "DirMove error: Can't move root")
+	if srcPath == "" || dstPath == "" {
+		fs.Debugf(src, "DirMove error: Can't move root")
 		return errors.New("can't move root directory")
 	}
 
-	// Find ID of parent
-	dstLeaf, dstDirectoryID, err := f.dirCache.FindPath(f.root, true)
-	if err != nil {
-		return err
-	}
-
-	// Find the ID of the source and make a node from it
+	// find the root src directory
 	err = srcFs.dirCache.FindRoot(false)
 	if err != nil {
-		fs.Debug(src, "DirMove error: error finding src root: %v", err)
 		return err
 	}
-	srcDirectoryID, err := srcFs.dirCache.RootParentID()
+
+	// find the root dst directory
+	if dstRemote != "" {
+		err = f.dirCache.FindRoot(true)
+		if err != nil {
+			return err
+		}
+	} else {
+		if f.dirCache.FoundRoot() {
+			return fs.ErrorDirExists
+		}
+	}
+
+	// Find ID of dst parent, creating subdirs if necessary
+	findPath := dstRemote
+	if dstRemote == "" {
+		findPath = f.root
+	}
+	dstLeaf, dstDirectoryID, err := f.dirCache.FindPath(findPath, true)
+	if err != nil {
+		return err
+	}
+
+	// Check destination does not exist
+	if dstRemote != "" {
+		_, err = f.dirCache.FindDir(dstRemote, false)
+		if err == fs.ErrorDirNotFound {
+			// OK
+		} else if err != nil {
+			return err
+		} else {
+			return fs.ErrorDirExists
+		}
+	}
+
+	// Find ID of src parent
+	findPath = srcRemote
+	var srcDirectoryID string
+	if srcRemote == "" {
+		srcDirectoryID, err = srcFs.dirCache.RootParentID()
+	} else {
+		_, srcDirectoryID, err = srcFs.dirCache.FindPath(findPath, false)
+	}
+	if err != nil {
+		return err
+	}
+	srcLeaf, _ := dircache.SplitPath(srcPath)
+
+	// Find ID of src
+	srcID, err := srcFs.dirCache.FindDir(srcRemote, false)
 	if err != nil {
-		fs.Debug(src, "DirMove error: error finding src RootParentID: %v", err)
 		return err
 	}
-	srcLeaf, _ := dircache.SplitPath(srcFs.root)
 
 	// FIXME make a proper node.UpdateMetadata command
-	srcInfo := acd.NodeFromId(srcFs.dirCache.RootID(), f.c.Nodes)
+	srcInfo := acd.NodeFromId(srcID, f.c.Nodes)
 	var jsonStr string
 	err = srcFs.pacer.Call(func() (bool, error) {
 		jsonStr, err = srcInfo.GetMetadata()
 		return srcFs.shouldRetry(nil, err)
 	})
 	if err != nil {
-		fs.Debug(src, "DirMove error: error reading src metadata: %v", err)
+		fs.Debugf(src, "DirMove error: error reading src metadata: %v", err)
 		return err
 	}
 	err = json.Unmarshal([]byte(jsonStr), &srcInfo)
 	if err != nil {
-		fs.Debug(src, "DirMove error: error reading unpacking src metadata: %v", err)
+		fs.Debugf(src, "DirMove error: error reading unpacking src metadata: %v", err)
 		return err
 	}
 
-	_, err = f.moveNode(srcFs.root, dstLeaf, dstDirectoryID, srcInfo, srcLeaf, srcDirectoryID, true)
+	err = f.moveNode(srcPath, dstLeaf, dstDirectoryID, srcInfo, srcLeaf, srcDirectoryID, true)
+	if err != nil {
+		return err
+	}
 
-	srcFs.dirCache.ResetRoot()
-	return err
+	srcFs.dirCache.FlushDir(srcRemote)
+	return nil
 }
 
 // purgeCheck remotes the root directory, if check is set then it
@@ -747,6 +778,10 @@ func (f *Fs) purgeCheck(dir string, check bool) error {
 		return errors.New("can't purge root directory")
 	}
 	dc := f.dirCache
+	err := dc.FindRoot(false)
+	if err != nil {
+		return err
+	}
 	rootID, err := dc.FindDir(dir, false)
 	if err != nil {
 		return err
@@ -764,7 +799,7 @@ func (f *Fs) purgeCheck(dir string, check bool) error {
 				empty = false
 				return true
 			default:
-				fs.Debug("Found ASSET %s", *node.Id)
+				fs.Debugf("Found ASSET %s", *node.Id)
 			}
 			return false
 		})
@@ -822,7 +857,7 @@ func (f *Fs) Hashes() fs.HashSet {
 //func (f *Fs) Copy(src fs.Object, remote string) (fs.Object, error) {
 // srcObj, ok := src.(*Object)
 // if !ok {
-// 	fs.Debug(src, "Can't copy - not same remote type")
+// 	fs.Debugf(src, "Can't copy - not same remote type")
 // 	return nil, fs.ErrorCantCopy
 // }
 // srcFs := srcObj.fs
@@ -890,7 +925,7 @@ func (o *Object) readMetaData() (err error) {
 	if o.info != nil {
 		return nil
 	}
-	leaf, directoryID, err := o.fs.dirCache.FindPath(o.remote, false)
+	leaf, directoryID, err := o.fs.dirCache.FindRootAndPath(o.remote, false)
 	if err != nil {
 		if err == fs.ErrorDirNotFound {
 			return fs.ErrorObjectNotFound
@@ -922,12 +957,12 @@ func (o *Object) readMetaData() (err error) {
 func (o *Object) ModTime() time.Time {
 	err := o.readMetaData()
 	if err != nil {
-		fs.Log(o, "Failed to read metadata: %v", err)
+		fs.Debugf(o, "Failed to read metadata: %v", err)
 		return time.Now()
 	}
 	modTime, err := time.Parse(timeFormat, *o.info.ModifiedDate)
 	if err != nil {
-		fs.Log(o, "Failed to read mtime from object: %v", err)
+		fs.Debugf(o, "Failed to read mtime from object: %v", err)
 		return time.Now()
 	}
 	return modTime
@@ -948,7 +983,7 @@ func (o *Object) Storable() bool {
 func (o *Object) Open(options ...fs.OpenOption) (in io.ReadCloser, err error) {
 	bigObject := o.Size() >= int64(tempLinkThreshold)
 	if bigObject {
-		fs.Debug(o, "Dowloading large object via tempLink")
+		fs.Debugf(o, "Dowloading large object via tempLink")
 	}
 	file := acd.File{Node: o.info}
 	var resp *http.Response
@@ -974,9 +1009,9 @@ func (o *Object) Update(in io.Reader, src fs.ObjectInfo) error {
 	var err error
 	err = o.fs.pacer.CallNoRetry(func() (bool, error) {
 		start := time.Now()
-		o.fs.startUpload()
+		o.fs.tokenRenewer.Start()
 		info, resp, err = file.Overwrite(in)
-		o.fs.stopUpload()
+		o.fs.tokenRenewer.Stop()
 		var ok bool
 		ok, info, err = o.fs.checkUpload(resp, in, src, info, err, time.Since(start))
 		if ok {
@@ -1008,25 +1043,23 @@ func (o *Object) Remove() error {
 }
 
 // Restore a node
-func (f *Fs) restoreNode(info *acd.Node) (*acd.Node, error) {
+func (f *Fs) restoreNode(info *acd.Node) (newInfo *acd.Node, err error) {
 	var resp *http.Response
-	var err error
 	err = f.pacer.Call(func() (bool, error) {
-		info, resp, err = info.Restore()
+		newInfo, resp, err = info.Restore()
 		return f.shouldRetry(resp, err)
 	})
-	return info, err
+	return newInfo, err
 }
 
 // Changes name of given node
-func (f *Fs) renameNode(info *acd.Node, newName string) (*acd.Node, error) {
+func (f *Fs) renameNode(info *acd.Node, newName string) (newInfo *acd.Node, err error) {
 	var resp *http.Response
-	var err error
 	err = f.pacer.Call(func() (bool, error) {
-		info, resp, err = info.Rename(newName)
+		newInfo, resp, err = info.Rename(newName)
 		return f.shouldRetry(resp, err)
 	})
-	return info, err
+	return newInfo, err
 }
 
 // Replaces one parent with another, effectively moving the file. Leaves other
@@ -1057,73 +1090,73 @@ func (f *Fs) removeParent(info *acd.Node, parentID string) error {
 
 // moveNode moves the node given from the srcLeaf,srcDirectoryID to
 // the dstLeaf,dstDirectoryID
-func (f *Fs) moveNode(name, dstLeaf, dstDirectoryID string, srcInfo *acd.Node, srcLeaf, srcDirectoryID string, useDirErrorMsgs bool) (dstInfo *acd.Node, err error) {
-	// fs.Debug(name, "moveNode dst(%q,%s) <- src(%q,%s)", dstLeaf, dstDirectoryID, srcLeaf, srcDirectoryID)
+func (f *Fs) moveNode(name, dstLeaf, dstDirectoryID string, srcInfo *acd.Node, srcLeaf, srcDirectoryID string, useDirErrorMsgs bool) (err error) {
+	// fs.Debugf(name, "moveNode dst(%q,%s) <- src(%q,%s)", dstLeaf, dstDirectoryID, srcLeaf, srcDirectoryID)
 	cantMove := fs.ErrorCantMove
 	if useDirErrorMsgs {
 		cantMove = fs.ErrorCantDirMove
 	}
 
 	if len(srcInfo.Parents) > 1 && srcLeaf != dstLeaf {
-		fs.Debug(name, "Move error: object is attached to multiple parents and should be renamed. This would change the name of the node in all parents.")
-		return nil, cantMove
+		fs.Debugf(name, "Move error: object is attached to multiple parents and should be renamed. This would change the name of the node in all parents.")
+		return cantMove
 	}
 
 	if srcLeaf != dstLeaf {
-		// fs.Debug(name, "renaming")
-		dstInfo, err = f.renameNode(srcInfo, dstLeaf)
+		// fs.Debugf(name, "renaming")
+		_, err = f.renameNode(srcInfo, dstLeaf)
 		if err != nil {
-			fs.Debug(name, "Move: quick path rename failed: %v", err)
+			fs.Debugf(name, "Move: quick path rename failed: %v", err)
 			goto OnConflict
 		}
 	}
 	if srcDirectoryID != dstDirectoryID {
-		// fs.Debug(name, "trying parent replace: %s -> %s", oldParentID, newParentID)
+		// fs.Debugf(name, "trying parent replace: %s -> %s", oldParentID, newParentID)
 		err = f.replaceParent(srcInfo, srcDirectoryID, dstDirectoryID)
 		if err != nil {
-			fs.Debug(name, "Move: quick path parent replace failed: %v", err)
-			return nil, err
+			fs.Debugf(name, "Move: quick path parent replace failed: %v", err)
+			return err
 		}
 	}
 
-	return dstInfo, nil
+	return nil
 
 OnConflict:
-	fs.Debug(name, "Could not directly rename file, presumably because there was a file with the same name already. Instead, the file will now be trashed where such operations do not cause errors. It will be restored to the correct parent after. If any of the subsequent calls fails, the rename/move will be in an invalid state.")
+	fs.Debugf(name, "Could not directly rename file, presumably because there was a file with the same name already. Instead, the file will now be trashed where such operations do not cause errors. It will be restored to the correct parent after. If any of the subsequent calls fails, the rename/move will be in an invalid state.")
 
-	// fs.Debug(name, "Trashing file")
+	// fs.Debugf(name, "Trashing file")
 	err = f.removeNode(srcInfo)
 	if err != nil {
-		fs.Debug(name, "Move: remove node failed: %v", err)
-		return nil, err
+		fs.Debugf(name, "Move: remove node failed: %v", err)
+		return err
 	}
-	// fs.Debug(name, "Renaming file")
+	// fs.Debugf(name, "Renaming file")
 	_, err = f.renameNode(srcInfo, dstLeaf)
 	if err != nil {
-		fs.Debug(name, "Move: rename node failed: %v", err)
-		return nil, err
+		fs.Debugf(name, "Move: rename node failed: %v", err)
+		return err
 	}
 	// note: replacing parent is forbidden by API, modifying them individually is
 	// okay though
-	// fs.Debug(name, "Adding target parent")
+	// fs.Debugf(name, "Adding target parent")
 	err = f.addParent(srcInfo, dstDirectoryID)
 	if err != nil {
-		fs.Debug(name, "Move: addParent failed: %v", err)
-		return nil, err
+		fs.Debugf(name, "Move: addParent failed: %v", err)
+		return err
 	}
-	// fs.Debug(name, "removing original parent")
+	// fs.Debugf(name, "removing original parent")
 	err = f.removeParent(srcInfo, srcDirectoryID)
 	if err != nil {
-		fs.Debug(name, "Move: removeParent failed: %v", err)
-		return nil, err
+		fs.Debugf(name, "Move: removeParent failed: %v", err)
+		return err
 	}
-	// fs.Debug(name, "Restoring")
-	dstInfo, err = f.restoreNode(srcInfo)
+	// fs.Debugf(name, "Restoring")
+	_, err = f.restoreNode(srcInfo)
 	if err != nil {
-		fs.Debug(name, "Move: restoreNode node failed: %v", err)
-		return nil, err
+		fs.Debugf(name, "Move: restoreNode node failed: %v", err)
+		return err
 	}
-	return dstInfo, nil
+	return nil
 }
 
 // MimeType of an Object if known, "" otherwise
