@@ -25,7 +25,6 @@ import (
 	"github.com/ncw/rclone/pacer"
 	"github.com/ncw/rclone/rest"
 	"github.com/pkg/errors"
-	"github.com/spf13/pflag"
 )
 
 const (
@@ -50,8 +49,8 @@ var (
 	minChunkSize       = fs.SizeSuffix(100E6)
 	chunkSize          = fs.SizeSuffix(96 * 1024 * 1024)
 	uploadCutoff       = fs.SizeSuffix(200E6)
-	b2TestMode         = pflag.StringP("b2-test-mode", "", "", "A flag string for X-Bz-Test-Mode header.")
-	b2Versions         = pflag.BoolP("b2-versions", "", false, "Include old versions in directory listings.")
+	b2TestMode         = fs.StringP("b2-test-mode", "", "", "A flag string for X-Bz-Test-Mode header.")
+	b2Versions         = fs.BoolP("b2-versions", "", false, "Include old versions in directory listings.")
 	errNotWithVersions = errors.New("can't modify or delete files in --b2-versions mode")
 )
 
@@ -73,13 +72,15 @@ func init() {
 		},
 		},
 	})
-	pflag.VarP(&uploadCutoff, "b2-upload-cutoff", "", "Cutoff for switching to chunked upload")
-	pflag.VarP(&chunkSize, "b2-chunk-size", "", "Upload chunk size. Must fit in memory.")
+	fs.VarP(&uploadCutoff, "b2-upload-cutoff", "", "Cutoff for switching to chunked upload")
+	fs.VarP(&chunkSize, "b2-chunk-size", "", "Upload chunk size. Must fit in memory.")
 }
 
 // Fs represents a remote b2 server
 type Fs struct {
 	name          string                       // name of this remote
+	root          string                       // the path we are working on if any
+	features      *fs.Features                 // optional features
 	account       string                       // account name
 	key           string                       // auth key
 	endpoint      string                       // name of the starting api endpoint
@@ -87,14 +88,12 @@ type Fs struct {
 	bucket        string                       // the bucket we are working on
 	bucketIDMutex sync.Mutex                   // mutex to protect _bucketID
 	_bucketID     string                       // the ID of the bucket we are working on
-	root          string                       // the path we are working on if any
 	info          api.AuthorizeAccountResponse // result of authorize call
 	uploadMu      sync.Mutex                   // lock for upload variable
 	uploads       []*api.GetUploadURLResponse  // result of get upload URL calls
 	authMu        sync.Mutex                   // lock for authorizing the account
 	pacer         *pacer.Pacer                 // To pace and retry the API calls
-	uploadTokens  chan struct{}                // control concurrency of uploads
-	extraTokens   chan struct{}                // extra tokens for multipart uploads
+	bufferTokens  chan []byte                  // control concurrency of multipart uploads
 }
 
 // Object describes a b2 object
@@ -131,6 +130,11 @@ func (f *Fs) String() string {
 	return fmt.Sprintf("B2 bucket %s path %s", f.bucket, f.root)
 }
 
+// Features returns the optional features of this Fs
+func (f *Fs) Features() *fs.Features {
+	return f.features
+}
+
 // Pattern to match a b2 path
 var matcher = regexp.MustCompile(`^([^/]*)(.*)$`)
 
@@ -169,12 +173,12 @@ func (f *Fs) shouldRetryNoReauth(resp *http.Response, err error) (bool, error) {
 			var err error
 			retryAfter, err = strconv.Atoi(retryAfterString)
 			if err != nil {
-				fs.ErrorLog(f, "Malformed %s header %q: %v", retryAfterHeader, retryAfterString, err)
+				fs.Errorf(f, "Malformed %s header %q: %v", retryAfterHeader, retryAfterString, err)
 			}
 		}
 		retryAfterDuration := time.Duration(retryAfter) * time.Second
 		if f.pacer.GetSleep() < retryAfterDuration {
-			fs.Debug(f, "Setting sleep to %v after error: %v", retryAfterDuration, err)
+			fs.Debugf(f, "Setting sleep to %v after error: %v", retryAfterDuration, err)
 			// We set 1/2 the value here because the pacer will double it immediately
 			f.pacer.SetSleep(retryAfterDuration / 2)
 		}
@@ -187,7 +191,7 @@ func (f *Fs) shouldRetryNoReauth(resp *http.Response, err error) (bool, error) {
 // deserve to be retried.  It returns the err as a convenience
 func (f *Fs) shouldRetry(resp *http.Response, err error) (bool, error) {
 	if resp != nil && resp.StatusCode == 401 {
-		fs.Debug(f, "Unauthorized: %v", err)
+		fs.Debugf(f, "Unauthorized: %v", err)
 		// Reauth
 		authErr := f.authorizeAccount()
 		if authErr != nil {
@@ -204,7 +208,7 @@ func errorHandler(resp *http.Response) error {
 	errResponse := new(api.Error)
 	err := rest.DecodeJSON(resp, &errResponse)
 	if err != nil {
-		fs.Debug(nil, "Couldn't decode error response: %v", err)
+		fs.Debugf(nil, "Couldn't decode error response: %v", err)
 	}
 	if errResponse.Code == "" {
 		errResponse.Code = "unknown"
@@ -230,15 +234,15 @@ func NewFs(name, root string) (fs.Fs, error) {
 	if err != nil {
 		return nil, err
 	}
-	account := fs.ConfigFile.MustValue(name, "account")
+	account := fs.ConfigFileGet(name, "account")
 	if account == "" {
 		return nil, errors.New("account not found")
 	}
-	key := fs.ConfigFile.MustValue(name, "key")
+	key := fs.ConfigFileGet(name, "key")
 	if key == "" {
 		return nil, errors.New("key not found")
 	}
-	endpoint := fs.ConfigFile.MustValue(name, "endpoint", defaultEndpoint)
+	endpoint := fs.ConfigFileGet(name, "endpoint", defaultEndpoint)
 	f := &Fs{
 		name:         name,
 		bucket:       bucket,
@@ -248,19 +252,18 @@ func NewFs(name, root string) (fs.Fs, error) {
 		endpoint:     endpoint,
 		srv:          rest.NewClient(fs.Config.Client()).SetErrorHandler(errorHandler),
 		pacer:        pacer.New().SetMinSleep(minSleep).SetMaxSleep(maxSleep).SetDecayConstant(decayConstant),
-		uploadTokens: make(chan struct{}, fs.Config.Transfers),
-		extraTokens:  make(chan struct{}, fs.Config.Transfers),
+		bufferTokens: make(chan []byte, fs.Config.Transfers),
 	}
+	f.features = (&fs.Features{ReadMimeType: true, WriteMimeType: true}).Fill(f)
 	// Set the test flag if required
 	if *b2TestMode != "" {
 		testMode := strings.TrimSpace(*b2TestMode)
 		f.srv.SetHeader(testModeHeader, testMode)
-		fs.Debug(f, "Setting test header \"%s: %s\"", testModeHeader, testMode)
+		fs.Debugf(f, "Setting test header \"%s: %s\"", testModeHeader, testMode)
 	}
-	// Fill up the upload and extra tokens
+	// Fill up the buffer tokens
 	for i := 0; i < fs.Config.Transfers; i++ {
-		f.returnUploadToken()
-		f.extraTokens <- struct{}{}
+		f.bufferTokens <- nil
 	}
 	err = f.authorizeAccount()
 	if err != nil {
@@ -364,77 +367,24 @@ func (f *Fs) clearUploadURL() {
 	f.uploadMu.Unlock()
 }
 
-// Gets an upload token to control the concurrency
-func (f *Fs) getUploadToken() {
-	<-f.uploadTokens
-}
-
-// Return an upload token
-func (f *Fs) returnUploadToken() {
-	f.uploadTokens <- struct{}{}
-}
-
-// Help count the multipart uploads
-type multipartUploadCounter struct {
-	f           *Fs
-	uploadToken chan struct{}
-}
-
-// Create a new upload counter. This gets an upload token for
-// exclusive use by this multipart upload - the other tokens are
-// shared between all the multipart uploads.
-//
-// Call .finished() when done to return the upload token.
-func (f *Fs) newMultipartUploadCounter() *multipartUploadCounter {
-	m := &multipartUploadCounter{
-		f:           f,
-		uploadToken: make(chan struct{}, 1),
+// getUploadBlock gets a block from the pool of size chunkSize
+func (f *Fs) getUploadBlock() []byte {
+	buf := <-f.bufferTokens
+	if buf == nil {
+		buf = make([]byte, chunkSize)
 	}
-	f.getUploadToken()
-	m.uploadToken <- struct{}{}
-	return m
+	// fs.Debugf(f, "Getting upload block %p", buf)
+	return buf
 }
 
-// Gets an upload token for a multipart upload
-//
-// This gets one token only from the first class tokens.  This means
-// that the multiplart upload is guaranteed at least one token as
-// there is one first class token per possible upload.
-//
-// Pass the return value to returnMultipartUploadToken
-func (m *multipartUploadCounter) getMultipartUploadToken() bool {
-	// get the upload token by preference
-	select {
-	case <-m.uploadToken:
-		return true
-	default:
+// putUploadBlock returns a block to the pool of size chunkSize
+func (f *Fs) putUploadBlock(buf []byte) {
+	buf = buf[:cap(buf)]
+	if len(buf) != int(chunkSize) {
+		panic("bad blocksize returned to pool")
 	}
-	// ...otherwise wait for the first one to appear.
-	//
-	// If both uploadToken and extraTokens are ready at this point
-	// (unlikely but possible) and we get an extraToken instead of
-	// an uploadToken this will not cause any harm - this
-	// multipart upload will get an extra upload slot temporarily.
-	select {
-	case <-m.uploadToken:
-		return true
-	case <-m.f.extraTokens:
-		return false
-	}
-}
-
-// Return a multipart upload token retreived from getMultipartUploadToken
-func (m *multipartUploadCounter) returnMultipartUploadToken(firstClass bool) {
-	if firstClass {
-		m.uploadToken <- struct{}{}
-	} else {
-		m.f.extraTokens <- struct{}{}
-	}
-}
-
-// Mark us finished with this upload counter
-func (m *multipartUploadCounter) finished() {
-	m.f.returnUploadToken()
+	// fs.Debugf(f, "Returning upload block %p", buf)
+	f.bufferTokens <- buf
 }
 
 // Return an Object from a path
@@ -517,7 +467,6 @@ func (f *Fs) list(dir string, level int, prefix string, limit int, hidden bool,
 	if prefix != "" {
 		request.StartFileName = prefix
 	}
-	var response api.ListFileNamesResponse
 	opts := rest.Opts{
 		Method: "POST",
 		Path:   "/b2_list_file_names",
@@ -526,6 +475,7 @@ func (f *Fs) list(dir string, level int, prefix string, limit int, hidden bool,
 		opts.Path = "/b2_list_file_versions"
 	}
 	for {
+		var response api.ListFileNamesResponse
 		err := f.pacer.Call(func() (bool, error) {
 			resp, err := f.srv.CallJSON(&opts, &request, &response)
 			return f.shouldRetry(resp, err)
@@ -540,7 +490,7 @@ func (f *Fs) list(dir string, level int, prefix string, limit int, hidden bool,
 				return nil
 			}
 			if !strings.HasPrefix(file.Name, f.root) {
-				fs.Log(f, "Odd name received %q", file.Name)
+				fs.Debugf(f, "Odd name received %q", file.Name)
 				continue
 			}
 			remote := file.Name[len(f.root):]
@@ -750,7 +700,7 @@ func (f *Fs) Mkdir(dir string) error {
 					return nil
 				}
 				if getBucketErr != fs.ErrorDirNotFound {
-					fs.Debug(f, "Error checking bucket exists: %v", getBucketErr)
+					fs.Debugf(f, "Error checking bucket exists: %v", getBucketErr)
 				}
 			}
 		}
@@ -857,13 +807,13 @@ func (f *Fs) purge(oldOnly bool) error {
 			fs.Stats.Checking(remote)
 			if oldOnly && last != remote {
 				if object.Action == "hide" {
-					fs.Debug(remote, "Deleting current version (id %q) as it is a hide marker", object.ID)
+					fs.Debugf(remote, "Deleting current version (id %q) as it is a hide marker", object.ID)
 					toBeDeleted <- object
 				} else {
-					fs.Debug(remote, "Not deleting current version (id %q) %q", object.ID, object.Action)
+					fs.Debugf(remote, "Not deleting current version (id %q) %q", object.ID, object.Action)
 				}
 			} else {
-				fs.Debug(remote, "Deleting (id %q)", object.ID)
+				fs.Debugf(remote, "Deleting (id %q)", object.ID)
 				toBeDeleted <- object
 			}
 			last = remote
@@ -1036,7 +986,7 @@ func (o *Object) parseTimeString(timeString string) (err error) {
 	}
 	unixMilliseconds, err := strconv.ParseInt(timeString, 10, 64)
 	if err != nil {
-		fs.Debug(o, "Failed to parse mod time string %q: %v", timeString, err)
+		fs.Debugf(o, "Failed to parse mod time string %q: %v", timeString, err)
 		return err
 	}
 	o.modTime = time.Unix(unixMilliseconds/1E3, (unixMilliseconds%1E3)*1E6).UTC()
@@ -1158,12 +1108,12 @@ func (o *Object) Open(options ...fs.OpenOption) (in io.ReadCloser, err error) {
 	// Read sha1 from header if it isn't set
 	if o.sha1 == "" {
 		o.sha1 = resp.Header.Get(sha1Header)
-		fs.Debug(o, "Reading sha1 from header - %q", o.sha1)
+		fs.Debugf(o, "Reading sha1 from header - %q", o.sha1)
 		// if sha1 header is "none" (in big files), then need
 		// to read it from the metadata
 		if o.sha1 == "none" {
 			o.sha1 = resp.Header.Get(sha1InfoHeader)
-			fs.Debug(o, "Reading sha1 from info - %q", o.sha1)
+			fs.Debugf(o, "Reading sha1 from info - %q", o.sha1)
 		}
 	}
 	// Don't check length or hash on partial content
@@ -1265,16 +1215,15 @@ func (o *Object) Update(in io.Reader, src fs.ObjectInfo) (err error) {
 		in = fd
 	}
 
-	// Get upload Token
-	o.fs.getUploadToken()
-	defer o.fs.returnUploadToken()
-
 	// Get upload URL
 	upload, err := o.fs.getUploadURL()
 	if err != nil {
 		return err
 	}
-	defer o.fs.returnUploadURL(upload)
+	defer func() {
+		// return it like this because we might nil it out
+		o.fs.returnUploadURL(upload)
+	}()
 
 	// Headers for upload file
 	//
@@ -1343,6 +1292,11 @@ func (o *Object) Update(in io.Reader, src fs.ObjectInfo) (err error) {
 		},
 		ContentLength: &size,
 	}
+	// for go1.8 (see release notes) we must nil the Body if we want a
+	// "Content-Length: 0" header which b2 requires for all files.
+	if size == 0 {
+		opts.Body = nil
+	}
 	var response api.FileInfo
 	// Don't retry, return a retry error instead
 	err = o.fs.pacer.CallNoRetry(func() (bool, error) {
@@ -1350,7 +1304,7 @@ func (o *Object) Update(in io.Reader, src fs.ObjectInfo) (err error) {
 		retry, err := o.fs.shouldRetry(resp, err)
 		// On retryable error clear UploadURL
 		if retry {
-			fs.Debug(o, "Clearing upload URL because of error: %v", err)
+			fs.Debugf(o, "Clearing upload URL because of error: %v", err)
 			upload = nil
 		}
 		return retry, err
