@@ -43,6 +43,17 @@ func newDir(f fs.Fs, fsDir *fs.Dir) *Dir {
 	}
 }
 
+// rename should be called after the directory is renamed
+//
+// Reset the directory to new state, discarding all the objects and
+// reading everything again
+func (d *Dir) rename(newParent *Dir, fsDir *fs.Dir) {
+	d.path = fsDir.Name
+	d.modTime = fsDir.When
+	d.items = nil
+	d.read = time.Time{}
+}
+
 // addObject adds a new object or directory to the directory
 //
 // note that we add new objects rather than updating old ones
@@ -70,15 +81,15 @@ func (d *Dir) readDir() error {
 	defer d.mu.Unlock()
 	when := time.Now()
 	if d.read.IsZero() {
-		fs.Debug(d.path, "Reading directory")
+		fs.Debugf(d.path, "Reading directory")
 	} else {
 		age := when.Sub(d.read)
 		if age < dirCacheTime {
 			return nil
 		}
-		fs.Debug(d.path, "Re-reading directory (%v old)", age)
+		fs.Debugf(d.path, "Re-reading directory (%v old)", age)
 	}
-	objs, dirs, err := fs.NewLister().SetLevel(1).Start(d.f, d.path).GetAll()
+	entries, err := fs.ListDirSorted(d.f, false, d.path)
 	if err == fs.ErrorDirNotFound {
 		// We treat directory not found as empty because we
 		// create directories on the fly
@@ -93,26 +104,34 @@ func (d *Dir) readDir() error {
 	oldItems := d.items
 
 	// Cache the items by name
-	d.items = make(map[string]*DirEntry, len(objs)+len(dirs))
-	for _, obj := range objs {
-		name := path.Base(obj.Remote())
-		d.items[name] = &DirEntry{
-			o:    obj,
-			node: nil,
-		}
-	}
-	for _, dir := range dirs {
-		name := path.Base(dir.Remote())
-		// Use old dir value if it exists
-		if oldItem, ok := oldItems[name]; ok {
-			if _, ok := oldItem.o.(*fs.Dir); ok {
-				d.items[name] = oldItem
-				continue
+	d.items = make(map[string]*DirEntry, len(entries))
+	for _, entry := range entries {
+		switch item := entry.(type) {
+		case fs.Object:
+			obj := item
+			name := path.Base(obj.Remote())
+			d.items[name] = &DirEntry{
+				o:    obj,
+				node: nil,
 			}
-		}
-		d.items[name] = &DirEntry{
-			o:    dir,
-			node: nil,
+		case *fs.Dir:
+			dir := item
+			name := path.Base(dir.Remote())
+			// Use old dir value if it exists
+			if oldItem, ok := oldItems[name]; ok {
+				if _, ok := oldItem.o.(*fs.Dir); ok {
+					d.items[name] = oldItem
+					continue
+				}
+			}
+			d.items[name] = &DirEntry{
+				o:    dir,
+				node: nil,
+			}
+		default:
+			err = errors.Errorf("unknown type %T", item)
+			fs.Errorf(d.path, "readDir error: %v", err)
+			return err
 		}
 	}
 	d.read = when
@@ -152,7 +171,6 @@ var _ fusefs.Node = (*Dir)(nil)
 
 // Attr updates the attribes of a directory
 func (d *Dir) Attr(ctx context.Context, a *fuse.Attr) error {
-	fs.Debug(d.path, "Dir.Attr")
 	a.Gid = gid
 	a.Uid = uid
 	a.Mode = os.ModeDir | dirPerms
@@ -160,7 +178,8 @@ func (d *Dir) Attr(ctx context.Context, a *fuse.Attr) error {
 	a.Mtime = d.modTime
 	a.Ctime = d.modTime
 	a.Crtime = d.modTime
-	// FIXME include Valid so get some caching? Also mtime
+	// FIXME include Valid so get some caching?
+	fs.Debugf(d.path, "Dir.Attr %+v", a)
 	return nil
 }
 
@@ -186,7 +205,7 @@ func (d *Dir) lookupNode(leaf string) (item *DirEntry, err error) {
 		return nil, err
 	}
 	item = d.addObject(item.o, node)
-	return item, err
+	return item, nil
 }
 
 // Check interface satisfied
@@ -200,15 +219,15 @@ var _ fusefs.NodeRequestLookuper = (*Dir)(nil)
 // Lookup need not to handle the names "." and "..".
 func (d *Dir) Lookup(ctx context.Context, req *fuse.LookupRequest, resp *fuse.LookupResponse) (node fusefs.Node, err error) {
 	path := path.Join(d.path, req.Name)
-	fs.Debug(path, "Dir.Lookup")
+	fs.Debugf(path, "Dir.Lookup")
 	item, err := d.lookupNode(req.Name)
 	if err != nil {
 		if err != fuse.ENOENT {
-			fs.ErrorLog(path, "Dir.Lookup error: %v", err)
+			fs.Errorf(path, "Dir.Lookup error: %v", err)
 		}
 		return nil, err
 	}
-	fs.Debug(path, "Dir.Lookup OK")
+	fs.Debugf(path, "Dir.Lookup OK")
 	return item.node, nil
 }
 
@@ -217,10 +236,10 @@ var _ fusefs.HandleReadDirAller = (*Dir)(nil)
 
 // ReadDirAll reads the contents of the directory
 func (d *Dir) ReadDirAll(ctx context.Context) (dirents []fuse.Dirent, err error) {
-	fs.Debug(d.path, "Dir.ReadDirAll")
+	fs.Debugf(d.path, "Dir.ReadDirAll")
 	err = d.readDir()
 	if err != nil {
-		fs.Debug(d.path, "Dir.ReadDirAll error: %v", err)
+		fs.Debugf(d.path, "Dir.ReadDirAll error: %v", err)
 		return nil, err
 	}
 	d.mu.RLock()
@@ -242,12 +261,12 @@ func (d *Dir) ReadDirAll(ctx context.Context) (dirents []fuse.Dirent, err error)
 			}
 		default:
 			err = errors.Errorf("unknown type %T", item)
-			fs.ErrorLog(d.path, "Dir.ReadDirAll error: %v", err)
+			fs.Errorf(d.path, "Dir.ReadDirAll error: %v", err)
 			return nil, err
 		}
 		dirents = append(dirents, dirent)
 	}
-	fs.Debug(d.path, "Dir.ReadDirAll OK with %d entries", len(dirents))
+	fs.Debugf(d.path, "Dir.ReadDirAll OK with %d entries", len(dirents))
 	return dirents, nil
 }
 
@@ -256,16 +275,16 @@ var _ fusefs.NodeCreater = (*Dir)(nil)
 // Create makes a new file
 func (d *Dir) Create(ctx context.Context, req *fuse.CreateRequest, resp *fuse.CreateResponse) (fusefs.Node, fusefs.Handle, error) {
 	path := path.Join(d.path, req.Name)
-	fs.Debug(path, "Dir.Create")
+	fs.Debugf(path, "Dir.Create")
 	src := newCreateInfo(d.f, path)
 	// This gets added to the directory when the file is written
 	file := newFile(d, nil)
 	fh, err := newWriteFileHandle(d, file, src)
 	if err != nil {
-		fs.ErrorLog(path, "Dir.Create error: %v", err)
+		fs.Errorf(path, "Dir.Create error: %v", err)
 		return nil, nil, err
 	}
-	fs.Debug(path, "Dir.Create OK")
+	fs.Debugf(path, "Dir.Create OK")
 	return file, fh, nil
 }
 
@@ -273,17 +292,20 @@ var _ fusefs.NodeMkdirer = (*Dir)(nil)
 
 // Mkdir creates a new directory
 func (d *Dir) Mkdir(ctx context.Context, req *fuse.MkdirRequest) (fusefs.Node, error) {
-	// We just pretend to have created the directory - rclone will
-	// actually create the directory if we write files into it
 	path := path.Join(d.path, req.Name)
-	fs.Debug(path, "Dir.Mkdir")
+	fs.Debugf(path, "Dir.Mkdir")
+	err := d.f.Mkdir(path)
+	if err != nil {
+		fs.Errorf(path, "Dir.Mkdir failed to create directory: %v", err)
+		return nil, err
+	}
 	fsDir := &fs.Dir{
 		Name: path,
 		When: time.Now(),
 	}
 	dir := newDir(d.f, fsDir)
 	d.addObject(fsDir, dir)
-	fs.Debug(path, "Dir.Mkdir OK")
+	fs.Debugf(path, "Dir.Mkdir OK")
 	return dir, nil
 }
 
@@ -294,42 +316,45 @@ var _ fusefs.NodeRemover = (*Dir)(nil)
 // may correspond to a file (unlink) or to a directory (rmdir).
 func (d *Dir) Remove(ctx context.Context, req *fuse.RemoveRequest) error {
 	path := path.Join(d.path, req.Name)
-	fs.Debug(path, "Dir.Remove")
+	fs.Debugf(path, "Dir.Remove")
 	item, err := d.lookupNode(req.Name)
 	if err != nil {
-		fs.ErrorLog(path, "Dir.Remove error: %v", err)
+		fs.Errorf(path, "Dir.Remove error: %v", err)
 		return err
 	}
 	switch x := item.o.(type) {
 	case fs.Object:
 		err = x.Remove()
 		if err != nil {
-			fs.ErrorLog(path, "Dir.Remove file error: %v", err)
+			fs.Errorf(path, "Dir.Remove file error: %v", err)
 			return err
 		}
 	case *fs.Dir:
-		// Do nothing for deleting directory - rclone can't
-		// currently remote a random directory
-		//
-		// Check directory is empty first though
+		// Check directory is empty first
 		dir := item.node.(*Dir)
 		empty, err := dir.isEmpty()
 		if err != nil {
-			fs.ErrorLog(path, "Dir.Remove dir error: %v", err)
+			fs.Errorf(path, "Dir.Remove dir error: %v", err)
 			return err
 		}
 		if !empty {
 			// return fuse.ENOTEMPTY - doesn't exist though so use EEXIST
-			fs.ErrorLog(path, "Dir.Remove not empty")
+			fs.Errorf(path, "Dir.Remove not empty")
 			return fuse.EEXIST
 		}
+		// remove directory
+		err = d.f.Rmdir(path)
+		if err != nil {
+			fs.Errorf(path, "Dir.Remove failed to remove directory: %v", err)
+			return err
+		}
 	default:
-		fs.ErrorLog(path, "Dir.Remove unknown type %T", item)
+		fs.Errorf(path, "Dir.Remove unknown type %T", item)
 		return errors.Errorf("unknown type %T", item)
 	}
 	// Remove the item from the directory listing
 	d.delObject(req.Name)
-	fs.Debug(path, "Dir.Remove OK")
+	fs.Debugf(path, "Dir.Remove OK")
 	return nil
 }
 
@@ -342,65 +367,87 @@ func (d *Dir) Rename(ctx context.Context, req *fuse.RenameRequest, newDir fusefs
 	destDir, ok := newDir.(*Dir)
 	if !ok {
 		err := errors.Errorf("Unknown Dir type %T", newDir)
-		fs.ErrorLog(oldPath, "Dir.Rename error: %v", err)
+		fs.Errorf(oldPath, "Dir.Rename error: %v", err)
 		return err
 	}
 	newPath := path.Join(destDir.path, req.NewName)
-	fs.Debug(oldPath, "Dir.Rename to %q", newPath)
+	fs.Debugf(oldPath, "Dir.Rename to %q", newPath)
 	oldItem, err := d.lookupNode(req.OldName)
 	if err != nil {
-		fs.ErrorLog(oldPath, "Dir.Rename error: %v", err)
+		fs.Errorf(oldPath, "Dir.Rename error: %v", err)
 		return err
 	}
 	var newObj fs.BasicInfo
+	oldNode := oldItem.node
 	switch x := oldItem.o.(type) {
 	case fs.Object:
 		oldObject := x
-		do, ok := d.f.(fs.Mover)
-		if !ok {
-			err := errors.Errorf("Fs %q can't Move files", d.f)
-			fs.ErrorLog(oldPath, "Dir.Rename error: %v", err)
+		// FIXME: could Copy then Delete if Move not available
+		// - though care needed if case insensitive...
+		doMove := d.f.Features().Move
+		if doMove == nil {
+			err := errors.Errorf("Fs %q can't rename files (no Move)", d.f)
+			fs.Errorf(oldPath, "Dir.Rename error: %v", err)
 			return err
 		}
-		newObject, err := do.Move(oldObject, newPath)
+		newObject, err := doMove(oldObject, newPath)
 		if err != nil {
-			fs.ErrorLog(oldPath, "Dir.Rename error: %v", err)
+			fs.Errorf(oldPath, "Dir.Rename error: %v", err)
 			return err
 		}
 		newObj = newObject
+		// Update the node with the new details
+		if oldNode != nil {
+			if oldFile, ok := oldNode.(*File); ok {
+				fs.Debugf(oldItem.o, "Updating file with %v %p", newObject, oldFile)
+				oldFile.rename(destDir, newObject)
+			}
+		}
 	case *fs.Dir:
-		oldDir := oldItem.node.(*Dir)
-		empty, err := oldDir.isEmpty()
-		if err != nil {
-			fs.ErrorLog(oldPath, "Dir.Rename dir error: %v", err)
+		doDirMove := d.f.Features().DirMove
+		if doDirMove == nil {
+			err := errors.Errorf("Fs %q can't rename directories (no DirMove)", d.f)
+			fs.Errorf(oldPath, "Dir.Rename error: %v", err)
 			return err
 		}
-		if !empty {
-			// return fuse.ENOTEMPTY - doesn't exist though so use EEXIST
-			fs.ErrorLog(oldPath, "Dir.Rename can't rename non empty directory")
-			return fuse.EEXIST
+		srcRemote := x.Name
+		dstRemote := newPath
+		err = doDirMove(d.f, srcRemote, dstRemote)
+		if err != nil {
+			fs.Errorf(oldPath, "Dir.Rename error: %v", err)
+			return err
 		}
-		newObj = &fs.Dir{
-			Name: newPath,
-			When: time.Now(),
+		newDir := new(fs.Dir)
+		*newDir = *x
+		newDir.Name = newPath
+		newObj = newDir
+		// Update the node with the new details
+		if oldNode != nil {
+			if oldDir, ok := oldNode.(*Dir); ok {
+				fs.Debugf(oldItem.o, "Updating dir with %v %p", newDir, oldDir)
+				oldDir.rename(destDir, newDir)
+			}
 		}
 	default:
 		err = errors.Errorf("unknown type %T", oldItem)
-		fs.ErrorLog(d.path, "Dir.ReadDirAll error: %v", err)
+		fs.Errorf(d.path, "Dir.ReadDirAll error: %v", err)
 		return err
 	}
 
 	// Show moved - delete from old dir and add to new
 	d.delObject(req.OldName)
-	destDir.addObject(newObj, nil)
+	destDir.addObject(newObj, oldNode)
 
-	// FIXME need to flush the dir also
+	fs.Debugf(newPath, "Dir.Rename renamed from %q", oldPath)
+	return nil
+}
 
-	// FIXME use DirMover to move a directory?
-	// or maybe use MoveDir which can move anything
-	// fallback to Copy/Delete if no Move?
-	// if dir is empty then can move it
+// Check interface satisfied
+var _ fusefs.NodeFsyncer = (*Dir)(nil)
 
-	fs.ErrorLog(newPath, "Dir.Rename renamed from %q", oldPath)
+// Fsync the directory
+//
+// Note that we don't do anything except return OK
+func (d *Dir) Fsync(ctx context.Context, req *fuse.FsyncRequest) error {
 	return nil
 }
