@@ -25,7 +25,7 @@ empty directories in.
 		cmd.CheckArgs(1, 1, command, args)
 		fdst := cmd.NewFsDst(args)
 		cmd.Run(true, false, command, func() error {
-			return fs.Rmdirs(fdst)
+			return fs.Rmdirs(fdst, "")
 		})
 	},
 }
