@@ -42,10 +42,12 @@ const (
 // Globals
 var (
 	// Flags
-	driveFullList      = pflag.BoolP("drive-full-list", "", false, "Use a full listing for directory list. More data but usually quicker. (obsolete)")
-	driveAuthOwnerOnly = pflag.BoolP("drive-auth-owner-only", "", false, "Only consider files owned by the authenticated user. Requires drive-full-list.")
-	driveUseTrash      = pflag.BoolP("drive-use-trash", "", false, "Send files to the trash instead of deleting permanently.")
-	driveExtensions    = pflag.StringP("drive-formats", "", defaultExtensions, "Comma separated list of preferred formats for downloading Google docs.")
+	driveFullList      = fs.BoolP("drive-full-list", "", false, "Use a full listing for directory list. More data but usually quicker. (obsolete)")
+	driveAuthOwnerOnly = fs.BoolP("drive-auth-owner-only", "", false, "Only consider files owned by the authenticated user. Requires drive-full-list.")
+	driveUseTrash      = fs.BoolP("drive-use-trash", "", false, "Send files to the trash instead of deleting permanently.")
+	driveSkipGdocs     = fs.BoolP("drive-skip-gdocs", "", false, "Skip google documents in all listings.")
+	driveExtensions    = fs.StringP("drive-formats", "", defaultExtensions, "Comma separated list of preferred formats for downloading Google docs.")
+	driveListChunk     = pflag.Int64P("drive-list-chunk", "", 1000, "Size of listing chunk 100-1000. 0 to disable.")
 	// chunkSize is the size of the chunks created during a resumable upload and should be a power of two.
 	// 1<<18 is the minimum size supported by the Google uploader, and there is no maximum.
 	chunkSize         = fs.SizeSuffix(8 * 1024 * 1024)
@@ -103,8 +105,8 @@ func init() {
 			Help: "Google Application Client Secret - leave blank normally.",
 		}},
 	})
-	pflag.VarP(&driveUploadCutoff, "drive-upload-cutoff", "", "Cutoff for switching to chunked upload")
-	pflag.VarP(&chunkSize, "drive-chunk-size", "", "Upload chunk size. Must a power of 2 >= 256k.")
+	fs.VarP(&driveUploadCutoff, "drive-upload-cutoff", "", "Cutoff for switching to chunked upload")
+	fs.VarP(&chunkSize, "drive-chunk-size", "", "Upload chunk size. Must a power of 2 >= 256k.")
 
 	// Invert mimeTypeToExtension
 	extensionToMimeType = make(map[string]string, len(mimeTypeToExtension))
@@ -116,8 +118,9 @@ func init() {
 // Fs represents a remote drive server
 type Fs struct {
 	name       string             // name of this remote
-	svc        *drive.Service     // the connection to the drive server
 	root       string             // the path we are working on
+	features   *fs.Features       // optional features
+	svc        *drive.Service     // the connection to the drive server
 	client     *http.Client       // authorized client
 	about      *drive.About       // information about the drive, including the root
 	dirCache   *dircache.DirCache // Map of directory path to directory id
@@ -155,6 +158,11 @@ func (f *Fs) String() string {
 	return fmt.Sprintf("Google drive root '%s'", f.root)
 }
 
+// Features returns the optional features of this Fs
+func (f *Fs) Features() *fs.Features {
+	return f.features
+}
+
 // shouldRetry determines whehter a given err rates being retried
 func shouldRetry(err error) (again bool, errOut error) {
 	again = false
@@ -195,25 +203,34 @@ type listAllFn func(*drive.File) bool
 // If the user fn ever returns true then it early exits with found = true
 //
 // Search params: https://developers.google.com/drive/search-parameters
-func (f *Fs) listAll(dirID string, title string, directoriesOnly bool, filesOnly bool, fn listAllFn) (found bool, err error) {
-	query := fmt.Sprintf("trashed=false")
+func (f *Fs) listAll(dirID string, title string, directoriesOnly bool, filesOnly bool, includeTrashed bool, fn listAllFn) (found bool, err error) {
+	var query []string
+	if !includeTrashed {
+		query = append(query, "trashed=false")
+	}
 	if dirID != "" {
-		query += fmt.Sprintf(" and '%s' in parents", dirID)
+		query = append(query, fmt.Sprintf("'%s' in parents", dirID))
 	}
 	if title != "" {
 		// Escaping the backslash isn't documented but seems to work
 		title = strings.Replace(title, `\`, `\\`, -1)
 		title = strings.Replace(title, `'`, `\'`, -1)
-		query += fmt.Sprintf(" and title='%s'", title)
+		query = append(query, fmt.Sprintf("title='%s'", title))
 	}
 	if directoriesOnly {
-		query += fmt.Sprintf(" and mimeType='%s'", driveFolderType)
+		query = append(query, fmt.Sprintf("mimeType='%s'", driveFolderType))
 	}
 	if filesOnly {
-		query += fmt.Sprintf(" and mimeType!='%s'", driveFolderType)
+		query = append(query, fmt.Sprintf("mimeType!='%s'", driveFolderType))
 	}
 	// fmt.Printf("listAll Query = %q\n", query)
-	list := f.svc.Files.List().Q(query).MaxResults(1000)
+	list := f.svc.Files.List()
+	if len(query) > 0 {
+		list = list.Q(strings.Join(query, " and "))
+	}
+	if *driveListChunk > 0 {
+		list = list.MaxResults(*driveListChunk)
+	}
 OUTER:
 	for {
 		var files *drive.FileList
@@ -295,6 +312,7 @@ func NewFs(name, path string) (fs.Fs, error) {
 		root:  root,
 		pacer: pacer.New().SetMinSleep(minSleep).SetPacer(pacer.GoogleDrivePacer),
 	}
+	f.features = (&fs.Features{DuplicateFiles: true, ReadMimeType: true, WriteMimeType: true}).Fill(f)
 
 	// Create a new authorized Drive client.
 	f.client = oAuthClient
@@ -378,7 +396,7 @@ func (f *Fs) NewObject(remote string) (fs.Object, error) {
 // FindLeaf finds a directory of name leaf in the folder with ID pathID
 func (f *Fs) FindLeaf(pathID, leaf string) (pathIDOut string, found bool, err error) {
 	// Find the leaf in pathID
-	found, err = f.listAll(pathID, leaf, true, false, func(item *drive.File) bool {
+	found, err = f.listAll(pathID, leaf, true, false, false, func(item *drive.File) bool {
 		if item.Title == leaf {
 			pathIDOut = item.Id
 			return true
@@ -428,7 +446,7 @@ func (f *Fs) findExportFormat(filepath string, item *drive.File) (extension, lin
 	// Warn about unknown export formats
 	for mimeType := range item.ExportLinks {
 		if _, ok := mimeTypeToExtension[mimeType]; !ok {
-			fs.Debug(filepath, "Unknown export type %q - ignoring", mimeType)
+			fs.Debugf(filepath, "Unknown export type %q - ignoring", mimeType)
 		}
 	}
 
@@ -446,8 +464,8 @@ func (f *Fs) findExportFormat(filepath string, item *drive.File) (extension, lin
 
 // ListDir reads the directory specified by the job into out, returning any more jobs
 func (f *Fs) ListDir(out fs.ListOpts, job dircache.ListDirJob) (jobs []dircache.ListDirJob, err error) {
-	fs.Debug(f, "Reading %q", job.Path)
-	_, err = f.listAll(job.DirID, "", false, false, func(item *drive.File) bool {
+	fs.Debugf(f, "Reading %q", job.Path)
+	_, err = f.listAll(job.DirID, "", false, false, false, func(item *drive.File) bool {
 		remote := job.Path + item.Title
 		switch {
 		case *driveAuthOwnerOnly && !isAuthOwned(item):
@@ -467,8 +485,8 @@ func (f *Fs) ListDir(out fs.ListOpts, job dircache.ListDirJob) (jobs []dircache.
 					jobs = append(jobs, dircache.ListDirJob{DirID: item.Id, Path: remote + "/", Depth: job.Depth - 1})
 				}
 			}
-		case item.Md5Checksum != "":
-			// If item has MD5 sum it is a file stored on drive
+		case item.Md5Checksum != "" || item.FileSize > 0:
+			// If item has MD5 sum or a length it is a file stored on drive
 			o, err := f.newObjectWithInfo(remote, item)
 			if err != nil {
 				out.SetError(err)
@@ -481,27 +499,31 @@ func (f *Fs) ListDir(out fs.ListOpts, job dircache.ListDirJob) (jobs []dircache.
 			// If item has export links then it is a google doc
 			extension, link := f.findExportFormat(remote, item)
 			if extension == "" {
-				fs.Debug(remote, "No export formats found")
+				fs.Debugf(remote, "No export formats found")
 			} else {
 				o, err := f.newObjectWithInfo(remote+"."+extension, item)
 				if err != nil {
 					out.SetError(err)
 					return true
 				}
-				obj := o.(*Object)
-				obj.isDocument = true
-				obj.url = link
-				obj.bytes = -1
-				if out.Add(o) {
-					return true
+				if !*driveSkipGdocs {
+					obj := o.(*Object)
+					obj.isDocument = true
+					obj.url = link
+					obj.bytes = -1
+					if out.Add(o) {
+						return true
+					}
+				} else {
+					fs.Debugf(f, "Skip google document: %q", remote)
 				}
 			}
 		default:
-			fs.Debug(remote, "Ignoring unknown object")
+			fs.Debugf(remote, "Ignoring unknown object")
 		}
 		return false
 	})
-	fs.Debug(f, "Finished reading %q", job.Path)
+	fs.Debugf(f, "Finished reading %q", job.Path)
 	return jobs, err
 }
 
@@ -522,7 +544,7 @@ func (f *Fs) createFileInfo(remote string, modTime time.Time, size int64) (*Obje
 		bytes:  size,
 	}
 
-	leaf, directoryID, err := f.dirCache.FindPath(remote, true)
+	leaf, directoryID, err := f.dirCache.FindRootAndPath(remote, true)
 	if err != nil {
 		return nil, nil, err
 	}
@@ -604,34 +626,42 @@ func (f *Fs) Mkdir(dir string) error {
 	return err
 }
 
-// Rmdir deletes the container
+// Rmdir deletes a directory
 //
 // Returns an error if it isn't empty
 func (f *Fs) Rmdir(dir string) error {
 	root := path.Join(f.root, dir)
 	dc := f.dirCache
-	rootID, err := dc.FindDir(dir, false)
+	directoryID, err := dc.FindDir(dir, false)
 	if err != nil {
 		return err
 	}
-	var children *drive.ChildList
-	err = f.pacer.Call(func() (bool, error) {
-		children, err = f.svc.Children.List(rootID).MaxResults(10).Do()
-		return shouldRetry(err)
+	var trashedFiles = false
+	found, err := f.listAll(directoryID, "", false, false, true, func(item *drive.File) bool {
+		if item.Labels == nil || !item.Labels.Trashed {
+			fs.Debugf(dir, "Rmdir: contains file: %q", item.Title)
+			return true
+		}
+		fs.Debugf(dir, "Rmdir: contains trashed file: %q", item.Title)
+		trashedFiles = true
+		return false
 	})
 	if err != nil {
 		return err
 	}
-	if len(children.Items) > 0 {
-		return errors.Errorf("directory not empty: %#v", children.Items)
+	if found {
+		return errors.Errorf("directory not empty")
 	}
 	// Delete the directory if it isn't the root
 	if root != "" {
 		err = f.pacer.Call(func() (bool, error) {
-			if *driveUseTrash {
-				_, err = f.svc.Files.Trash(rootID).Do()
+			// trash the directory if it had trashed files
+			// in or the user wants to trash, otherwise
+			// delete it.
+			if trashedFiles || *driveUseTrash {
+				_, err = f.svc.Files.Trash(directoryID).Do()
 			} else {
-				err = f.svc.Files.Delete(rootID).Do()
+				err = f.svc.Files.Delete(directoryID).Do()
 			}
 			return shouldRetry(err)
 		})
@@ -663,7 +693,7 @@ func (f *Fs) Precision() time.Duration {
 func (f *Fs) Copy(src fs.Object, remote string) (fs.Object, error) {
 	srcObj, ok := src.(*Object)
 	if !ok {
-		fs.Debug(src, "Can't copy - not same remote type")
+		fs.Debugf(src, "Can't copy - not same remote type")
 		return nil, fs.ErrorCantCopy
 	}
 	if srcObj.isDocument {
@@ -728,19 +758,13 @@ func (f *Fs) Purge() error {
 func (f *Fs) Move(src fs.Object, remote string) (fs.Object, error) {
 	srcObj, ok := src.(*Object)
 	if !ok {
-		fs.Debug(src, "Can't move - not same remote type")
+		fs.Debugf(src, "Can't move - not same remote type")
 		return nil, fs.ErrorCantMove
 	}
 	if srcObj.isDocument {
 		return nil, errors.New("can't move a Google document")
 	}
 
-	// create the destination directory if necessary
-	err := f.dirCache.FindRoot(true)
-	if err != nil {
-		return nil, err
-	}
-
 	// Temporary Object under construction
 	dstObj, dstInfo, err := f.createFileInfo(remote, srcObj.ModTime(), srcObj.bytes)
 	if err != nil {
@@ -761,34 +785,72 @@ func (f *Fs) Move(src fs.Object, remote string) (fs.Object, error) {
 	return dstObj, nil
 }
 
-// DirMove moves src directory to this remote using server side move
-// operations.
+// DirMove moves src, srcRemote to this remote at dstRemote
+// using server side move operations.
 //
 // Will only be called if src.Fs().Name() == f.Name()
 //
 // If it isn't possible then return fs.ErrorCantDirMove
 //
 // If destination exists then return fs.ErrorDirExists
-func (f *Fs) DirMove(src fs.Fs) error {
+func (f *Fs) DirMove(src fs.Fs, srcRemote, dstRemote string) error {
 	srcFs, ok := src.(*Fs)
 	if !ok {
-		fs.Debug(srcFs, "Can't move directory - not same remote type")
+		fs.Debugf(srcFs, "Can't move directory - not same remote type")
 		return fs.ErrorCantDirMove
 	}
-
-	// Check if destination exists
-	if f.dirCache.FoundRoot() {
-		return fs.ErrorDirExists
-	}
+	srcPath := path.Join(srcFs.root, srcRemote)
+	dstPath := path.Join(f.root, dstRemote)
 
 	// Refuse to move to or from the root
-	if f.root == "" || srcFs.root == "" {
-		fs.Debug(src, "DirMove error: Can't move root")
+	if srcPath == "" || dstPath == "" {
+		fs.Debugf(src, "DirMove error: Can't move root")
 		return errors.New("can't move root directory")
 	}
 
-	// Find ID of parent
-	leaf, directoryID, err := f.dirCache.FindPath(f.root, true)
+	// find the root src directory
+	err := srcFs.dirCache.FindRoot(false)
+	if err != nil {
+		return err
+	}
+
+	// find the root dst directory
+	if dstRemote != "" {
+		err = f.dirCache.FindRoot(true)
+		if err != nil {
+			return err
+		}
+	} else {
+		if f.dirCache.FoundRoot() {
+			return fs.ErrorDirExists
+		}
+	}
+
+	// Find ID of dst parent, creating subdirs if necessary
+	var leaf, directoryID string
+	findPath := dstRemote
+	if dstRemote == "" {
+		findPath = f.root
+	}
+	leaf, directoryID, err = f.dirCache.FindPath(findPath, true)
+	if err != nil {
+		return err
+	}
+
+	// Check destination does not exist
+	if dstRemote != "" {
+		_, err = f.dirCache.FindDir(dstRemote, false)
+		if err == fs.ErrorDirNotFound {
+			// OK
+		} else if err != nil {
+			return err
+		} else {
+			return fs.ErrorDirExists
+		}
+	}
+
+	// Find ID of src
+	srcID, err := srcFs.dirCache.FindDir(srcRemote, false)
 	if err != nil {
 		return err
 	}
@@ -799,13 +861,13 @@ func (f *Fs) DirMove(src fs.Fs) error {
 		Parents: []*drive.ParentReference{{Id: directoryID}},
 	}
 	err = f.pacer.Call(func() (bool, error) {
-		_, err = f.svc.Files.Patch(srcFs.dirCache.RootID(), &patch).Do()
+		_, err = f.svc.Files.Patch(srcID, &patch).Do()
 		return shouldRetry(err)
 	})
 	if err != nil {
 		return err
 	}
-	srcFs.dirCache.ResetRoot()
+	srcFs.dirCache.FlushDir(srcRemote)
 	return nil
 }
 
@@ -855,12 +917,12 @@ func (o *Object) Size() int64 {
 		// how big it is
 		_, res, err := o.httpResponse("HEAD", nil)
 		if err != nil {
-			fs.ErrorLog(o, "Error reading size: %v", err)
+			fs.Errorf(o, "Error reading size: %v", err)
 			return 0
 		}
 		_ = res.Body.Close()
 		o.bytes = res.ContentLength
-		// fs.Debug(o, "Read size of document: %v", o.bytes)
+		// fs.Debugf(o, "Read size of document: %v", o.bytes)
 	}
 	return o.bytes
 }
@@ -881,7 +943,7 @@ func (o *Object) readMetaData() (err error) {
 		return nil
 	}
 
-	leaf, directoryID, err := o.fs.dirCache.FindPath(o.remote, false)
+	leaf, directoryID, err := o.fs.dirCache.FindRootAndPath(o.remote, false)
 	if err != nil {
 		if err == fs.ErrorDirNotFound {
 			return fs.ErrorObjectNotFound
@@ -889,7 +951,7 @@ func (o *Object) readMetaData() (err error) {
 		return err
 	}
 
-	found, err := o.fs.listAll(directoryID, leaf, false, true, func(item *drive.File) bool {
+	found, err := o.fs.listAll(directoryID, leaf, false, true, false, func(item *drive.File) bool {
 		if item.Title == leaf {
 			o.setMetaData(item)
 			return true
@@ -913,12 +975,12 @@ func (o *Object) readMetaData() (err error) {
 func (o *Object) ModTime() time.Time {
 	err := o.readMetaData()
 	if err != nil {
-		fs.Log(o, "Failed to read metadata: %v", err)
+		fs.Debugf(o, "Failed to read metadata: %v", err)
 		return time.Now()
 	}
 	modTime, err := time.Parse(timeFormatIn, o.modifiedDate)
 	if err != nil {
-		fs.Log(o, "Failed to read mtime from object: %v", err)
+		fs.Debugf(o, "Failed to read mtime from object: %v", err)
 		return time.Now()
 	}
 	return modTime
@@ -996,7 +1058,7 @@ func (file *openFile) Read(p []byte) (n int, err error) {
 func (file *openFile) Close() (err error) {
 	// If end of file, update bytes read
 	if file.eof {
-		// fs.Debug(file.o, "Updating size of doc after download to %v", file.bytes)
+		// fs.Debugf(file.o, "Updating size of doc after download to %v", file.bytes)
 		file.o.bytes = file.bytes
 	}
 	return file.in.Close()
@@ -1087,7 +1149,7 @@ func (o *Object) Remove() error {
 func (o *Object) MimeType() string {
 	err := o.readMetaData()
 	if err != nil {
-		fs.Log(o, "Failed to read metadata: %v", err)
+		fs.Debugf(o, "Failed to read metadata: %v", err)
 		return ""
 	}
 	return o.mimeType
