@@ -22,7 +22,6 @@ import (
 	"github.com/ncw/rclone/fs"
 	"github.com/ncw/rclone/oauthutil"
 	"github.com/pkg/errors"
-	"github.com/spf13/pflag"
 	"github.com/stacktic/dropbox"
 )
 
@@ -58,13 +57,13 @@ func init() {
 			Help: "Dropbox App Secret - leave blank normally.",
 		}},
 	})
-	pflag.VarP(&uploadChunkSize, "dropbox-chunk-size", "", fmt.Sprintf("Upload chunk size. Max %v.", maxUploadChunkSize))
+	fs.VarP(&uploadChunkSize, "dropbox-chunk-size", "", fmt.Sprintf("Upload chunk size. Max %v.", maxUploadChunkSize))
 }
 
 // Configuration helper - called after the user has put in the defaults
 func configHelper(name string) {
 	// See if already have a token
-	token := fs.ConfigFile.MustValue(name, "token")
+	token := fs.ConfigFileGet(name, "token")
 	if token != "" {
 		fmt.Printf("Already have a dropbox token - refresh?\n")
 		if !fs.Confirm() {
@@ -87,9 +86,9 @@ func configHelper(name string) {
 	token = db.AccessToken()
 
 	// Stuff it in the config file if it has changed
-	old := fs.ConfigFile.MustValue(name, "token")
+	old := fs.ConfigFileGet(name, "token")
 	if token != old {
-		fs.ConfigFile.SetValue(name, "token", token)
+		fs.ConfigFileSet(name, "token", token)
 		fs.SaveConfig()
 	}
 }
@@ -97,8 +96,9 @@ func configHelper(name string) {
 // Fs represents a remote dropbox server
 type Fs struct {
 	name           string           // name of this remote
-	db             *dropbox.Dropbox // the connection to the dropbox server
 	root           string           // the path we are working on
+	features       *fs.Features     // optional features
+	db             *dropbox.Dropbox // the connection to the dropbox server
 	slashRoot      string           // root with "/" prefix, lowercase
 	slashRootSlash string           // root with "/" prefix and postfix, lowercase
 }
@@ -130,15 +130,20 @@ func (f *Fs) String() string {
 	return fmt.Sprintf("Dropbox root '%s'", f.root)
 }
 
+// Features returns the optional features of this Fs
+func (f *Fs) Features() *fs.Features {
+	return f.features
+}
+
 // Makes a new dropbox from the config
 func newDropbox(name string) (*dropbox.Dropbox, error) {
 	db := dropbox.NewDropbox()
 
-	appKey := fs.ConfigFile.MustValue(name, "app_key")
+	appKey := fs.ConfigFileGet(name, "app_key")
 	if appKey == "" {
 		appKey = rcloneAppKey
 	}
-	appSecret := fs.ConfigFile.MustValue(name, "app_secret")
+	appSecret := fs.ConfigFileGet(name, "app_secret")
 	if appSecret == "" {
 		appSecret = fs.MustReveal(rcloneEncryptedAppSecret)
 	}
@@ -160,10 +165,11 @@ func NewFs(name, root string) (fs.Fs, error) {
 		name: name,
 		db:   db,
 	}
+	f.features = (&fs.Features{CaseInsensitive: true, ReadMimeType: true}).Fill(f)
 	f.setRoot(root)
 
 	// Read the token from the config file
-	token := fs.ConfigFile.MustValue(name, "token")
+	token := fs.ConfigFileGet(name, "token")
 
 	// Set our custom context which enables our custom transport for timeouts etc
 	db.SetContext(oauthutil.Context())
@@ -206,13 +212,14 @@ func (f *Fs) newObjectWithInfo(remote string, info *dropbox.Entry) (fs.Object, e
 		fs:     f,
 		remote: remote,
 	}
+	var err error
 	if info != nil {
-		o.setMetadataFromEntry(info)
+		err = o.setMetadataFromEntry(info)
 	} else {
-		err := o.readEntryAndSetMetadata()
-		if err != nil {
-			return nil, err
-		}
+		err = o.readEntryAndSetMetadata()
+	}
+	if err != nil {
+		return nil, err
 	}
 	return o, nil
 }
@@ -235,8 +242,7 @@ func strip(path, root string) (string, error) {
 	} else if len(root) == 0 {
 		root = "/"
 	}
-	lowercase := strings.ToLower(path)
-	if !strings.HasPrefix(lowercase, root) {
+	if !strings.HasPrefix(strings.ToLower(path), strings.ToLower(root)) {
 		return "", errors.Errorf("path %q is not under root %q", path, root)
 	}
 	return path[len(root):], nil
@@ -273,7 +279,7 @@ func (f *Fs) list(out fs.ListOpts, dir string) {
 			out.SetError(err)
 			break
 		}
-		fs.Debug(f, "%d delta entries received", len(deltaPage.Entries))
+		fs.Debugf(f, "%d delta entries received", len(deltaPage.Entries))
 		for i := range deltaPage.Entries {
 			deltaEntry := &deltaPage.Entries[i]
 			entry := deltaEntry.Entry
@@ -281,7 +287,7 @@ func (f *Fs) list(out fs.ListOpts, dir string) {
 				// This notifies of a deleted object
 			} else {
 				if len(entry.Path) <= 1 || entry.Path[0] != '/' {
-					fs.Log(f, "dropbox API inconsistency: a path should always start with a slash and be at least 2 characters: %s", entry.Path)
+					fs.Debugf(f, "dropbox API inconsistency: a path should always start with a slash and be at least 2 characters: %s", entry.Path)
 					continue
 				}
 
@@ -376,11 +382,8 @@ func (f *Fs) listOneLevel(out fs.ListOpts, dir string) {
 	}
 	for i := range dirEntry.Contents {
 		entry := &dirEntry.Contents[i]
-		remote, err := strip(entry.Path, root)
-		if err != nil {
-			out.SetError(err)
-			return
-		}
+		// Normalise the path to the dir passed in
+		remote := path.Join(dir, path.Base(entry.Path))
 		if entry.IsDir {
 			dir := &fs.Dir{
 				Name:  remote,
@@ -494,7 +497,7 @@ func (f *Fs) Precision() time.Duration {
 func (f *Fs) Copy(src fs.Object, remote string) (fs.Object, error) {
 	srcObj, ok := src.(*Object)
 	if !ok {
-		fs.Debug(src, "Can't copy - not same remote type")
+		fs.Debugf(src, "Can't copy - not same remote type")
 		return nil, fs.ErrorCantCopy
 	}
 
@@ -510,7 +513,10 @@ func (f *Fs) Copy(src fs.Object, remote string) (fs.Object, error) {
 	if err != nil {
 		return nil, errors.Wrap(err, "copy failed")
 	}
-	dstObj.setMetadataFromEntry(entry)
+	err = dstObj.setMetadataFromEntry(entry)
+	if err != nil {
+		return nil, errors.Wrap(err, "copy failed")
+	}
 	return dstObj, nil
 }
 
@@ -537,7 +543,7 @@ func (f *Fs) Purge() error {
 func (f *Fs) Move(src fs.Object, remote string) (fs.Object, error) {
 	srcObj, ok := src.(*Object)
 	if !ok {
-		fs.Debug(src, "Can't move - not same remote type")
+		fs.Debugf(src, "Can't move - not same remote type")
 		return nil, fs.ErrorCantMove
 	}
 
@@ -553,23 +559,29 @@ func (f *Fs) Move(src fs.Object, remote string) (fs.Object, error) {
 	if err != nil {
 		return nil, errors.Wrap(err, "move failed")
 	}
-	dstObj.setMetadataFromEntry(entry)
+	err = dstObj.setMetadataFromEntry(entry)
+	if err != nil {
+		return nil, errors.Wrap(err, "move failed")
+	}
 	return dstObj, nil
 }
 
-// DirMove moves src to this remote using server side move operations.
+// DirMove moves src, srcRemote to this remote at dstRemote
+// using server side move operations.
 //
 // Will only be called if src.Fs().Name() == f.Name()
 //
 // If it isn't possible then return fs.ErrorCantDirMove
 //
 // If destination exists then return fs.ErrorDirExists
-func (f *Fs) DirMove(src fs.Fs) error {
+func (f *Fs) DirMove(src fs.Fs, srcRemote, dstRemote string) error {
 	srcFs, ok := src.(*Fs)
 	if !ok {
-		fs.Debug(srcFs, "Can't move directory - not same remote type")
+		fs.Debugf(srcFs, "Can't move directory - not same remote type")
 		return fs.ErrorCantDirMove
 	}
+	srcPath := path.Join(srcFs.slashRoot, srcRemote)
+	dstPath := path.Join(f.slashRoot, dstRemote)
 
 	// Check if destination exists
 	entry, err := f.db.Metadata(f.slashRoot, false, false, "", "", metadataLimit)
@@ -577,8 +589,11 @@ func (f *Fs) DirMove(src fs.Fs) error {
 		return fs.ErrorDirExists
 	}
 
+	// Make sure the parent directory exists
+	// ...apparently not necessary
+
 	// Do the move
-	_, err = f.db.Move(srcFs.slashRoot, f.slashRoot)
+	_, err = f.db.Move(srcPath, dstPath)
 	if err != nil {
 		return errors.Wrap(err, "MoveDir failed")
 	}
@@ -623,11 +638,15 @@ func (o *Object) Size() int64 {
 // setMetadataFromEntry sets the fs data from a dropbox.Entry
 //
 // This isn't a complete set of metadata and has an inacurate date
-func (o *Object) setMetadataFromEntry(info *dropbox.Entry) {
+func (o *Object) setMetadataFromEntry(info *dropbox.Entry) error {
+	if info.IsDir {
+		return errors.Wrapf(fs.ErrorNotAFile, "%q", o.remote)
+	}
 	o.bytes = info.Bytes
 	o.modTime = time.Time(info.ClientMtime)
 	o.mimeType = info.MimeType
 	o.hasMetadata = true
+	return nil
 }
 
 // Reads the entry from dropbox
@@ -654,8 +673,7 @@ func (o *Object) readEntryAndSetMetadata() error {
 	if err != nil {
 		return err
 	}
-	o.setMetadataFromEntry(entry)
-	return nil
+	return o.setMetadataFromEntry(entry)
 }
 
 // Returns the remote path for the object
@@ -693,7 +711,7 @@ func (o *Object) readMetaData() (err error) {
 func (o *Object) ModTime() time.Time {
 	err := o.readMetaData()
 	if err != nil {
-		fs.Log(o, "Failed to read metadata: %v", err)
+		fs.Debugf(o, "Failed to read metadata: %v", err)
 		return time.Now()
 	}
 	return o.modTime
@@ -722,7 +740,7 @@ func (o *Object) Open(options ...fs.OpenOption) (in io.ReadCloser, err error) {
 			offset = x.Offset
 		default:
 			if option.Mandatory() {
-				fs.Log(o, "Unsupported mandatory option: %v", option)
+				fs.Logf(o, "Unsupported mandatory option: %v", option)
 			}
 		}
 	}
@@ -746,15 +764,14 @@ func (o *Object) Open(options ...fs.OpenOption) (in io.ReadCloser, err error) {
 func (o *Object) Update(in io.Reader, src fs.ObjectInfo) error {
 	remote := o.remotePath()
 	if ignoredFiles.MatchString(remote) {
-		fs.Log(o, "File name disallowed - not uploading")
+		fs.Logf(o, "File name disallowed - not uploading")
 		return nil
 	}
 	entry, err := o.fs.db.UploadByChunk(ioutil.NopCloser(in), int(uploadChunkSize), remote, true, "")
 	if err != nil {
 		return errors.Wrap(err, "upload failed")
 	}
-	o.setMetadataFromEntry(entry)
-	return nil
+	return o.setMetadataFromEntry(entry)
 }
 
 // Remove an object
@@ -767,7 +784,7 @@ func (o *Object) Remove() error {
 func (o *Object) MimeType() string {
 	err := o.readMetaData()
 	if err != nil {
-		fs.Log(o, "Failed to read metadata: %v", err)
+		fs.Logf(o, "Failed to read metadata: %v", err)
 		return ""
 	}
 	return o.mimeType
