@@ -14,10 +14,10 @@ import (
 	"io"
 	"io/ioutil"
 	"log"
-	"math"
 	"os"
 	"os/user"
-	"path"
+	"path/filepath"
+	"regexp"
 	"sort"
 	"strconv"
 	"strings"
@@ -32,7 +32,8 @@ import (
 )
 
 const (
-	configFileName = ".rclone.conf"
+	configFileName       = "rclone.conf"
+	hiddenConfigFileName = "." + configFileName
 
 	// ConfigToken is the key used to store the token under
 	ConfigToken = "token"
@@ -47,49 +48,50 @@ const (
 	ConfigAutomatic = "config_automatic"
 )
 
-// SizeSuffix is parsed by flag with k/M/G suffixes
-type SizeSuffix int64
-
 // Global
 var (
-	// ConfigFile is the config file data structure
-	ConfigFile *goconfig.ConfigFile
-	// HomeDir is the home directory of the user
-	HomeDir = configHome()
+	// configData is the config file data structure
+	configData *goconfig.ConfigFile
 	// ConfigPath points to the config file
-	ConfigPath = path.Join(HomeDir, configFileName)
+	ConfigPath = makeConfigPath()
 	// Config is the global config
 	Config = &ConfigInfo{}
 	// Flags
-	verbose         = pflag.BoolP("verbose", "v", false, "Print lots more stuff")
-	quiet           = pflag.BoolP("quiet", "q", false, "Print as little stuff as possible")
-	modifyWindow    = pflag.DurationP("modify-window", "", time.Nanosecond, "Max time diff to be considered the same")
-	checkers        = pflag.IntP("checkers", "", 8, "Number of checkers to run in parallel.")
-	transfers       = pflag.IntP("transfers", "", 4, "Number of file transfers to run in parallel.")
-	configFile      = pflag.StringP("config", "", ConfigPath, "Config file.")
-	checkSum        = pflag.BoolP("checksum", "c", false, "Skip based on checksum & size, not mod-time & size")
-	sizeOnly        = pflag.BoolP("size-only", "", false, "Skip based on size only, not mod-time or checksum")
-	ignoreTimes     = pflag.BoolP("ignore-times", "I", false, "Don't skip files that match size and time - transfer all files")
-	ignoreExisting  = pflag.BoolP("ignore-existing", "", false, "Skip all files that exist on destination")
-	dryRun          = pflag.BoolP("dry-run", "n", false, "Do a trial run with no permanent changes")
-	connectTimeout  = pflag.DurationP("contimeout", "", 60*time.Second, "Connect timeout")
-	timeout         = pflag.DurationP("timeout", "", 5*60*time.Second, "IO idle timeout")
-	dumpHeaders     = pflag.BoolP("dump-headers", "", false, "Dump HTTP headers - may contain sensitive info")
-	dumpBodies      = pflag.BoolP("dump-bodies", "", false, "Dump HTTP headers and bodies - may contain sensitive info")
-	dumpAuth        = pflag.BoolP("dump-auth", "", false, "Dump HTTP headers with auth info")
-	skipVerify      = pflag.BoolP("no-check-certificate", "", false, "Do not verify the server SSL certificate. Insecure.")
-	AskPassword     = pflag.BoolP("ask-password", "", true, "Allow prompt for password for encrypted configuration.")
-	deleteBefore    = pflag.BoolP("delete-before", "", false, "When synchronizing, delete files on destination before transfering")
-	deleteDuring    = pflag.BoolP("delete-during", "", false, "When synchronizing, delete files during transfer (default)")
-	deleteAfter     = pflag.BoolP("delete-after", "", false, "When synchronizing, delete files on destination after transfering")
-	lowLevelRetries = pflag.IntP("low-level-retries", "", 10, "Number of low level retries to do.")
-	updateOlder     = pflag.BoolP("update", "u", false, "Skip files that are newer on the destination.")
-	noGzip          = pflag.BoolP("no-gzip-encoding", "", false, "Don't set Accept-Encoding: gzip.")
-	maxDepth        = pflag.IntP("max-depth", "", -1, "If set limits the recursion depth to this.")
-	ignoreSize      = pflag.BoolP("ignore-size", "", false, "Ignore size when skipping use mod-time or checksum.")
-	noTraverse      = pflag.BoolP("no-traverse", "", false, "Don't traverse destination file system on copy.")
-	noUpdateModTime = pflag.BoolP("no-update-modtime", "", false, "Don't update destination mod-time if files identical.")
-	bwLimit         SizeSuffix
+	verbose         = CountP("verbose", "v", "Print lots more stuff (repeat for more)")
+	quiet           = BoolP("quiet", "q", false, "Print as little stuff as possible")
+	logLevel        = StringP("log-level", "", "INFO", "Log level DEBUG|INFO|NOTICE|ERROR")
+	modifyWindow    = DurationP("modify-window", "", time.Nanosecond, "Max time diff to be considered the same")
+	checkers        = IntP("checkers", "", 8, "Number of checkers to run in parallel.")
+	transfers       = IntP("transfers", "", 4, "Number of file transfers to run in parallel.")
+	configFile      = StringP("config", "", ConfigPath, "Config file.")
+	checkSum        = BoolP("checksum", "c", false, "Skip based on checksum & size, not mod-time & size")
+	sizeOnly        = BoolP("size-only", "", false, "Skip based on size only, not mod-time or checksum")
+	ignoreTimes     = BoolP("ignore-times", "I", false, "Don't skip files that match size and time - transfer all files")
+	ignoreExisting  = BoolP("ignore-existing", "", false, "Skip all files that exist on destination")
+	dryRun          = BoolP("dry-run", "n", false, "Do a trial run with no permanent changes")
+	connectTimeout  = DurationP("contimeout", "", 60*time.Second, "Connect timeout")
+	timeout         = DurationP("timeout", "", 5*60*time.Second, "IO idle timeout")
+	dumpHeaders     = BoolP("dump-headers", "", false, "Dump HTTP headers - may contain sensitive info")
+	dumpBodies      = BoolP("dump-bodies", "", false, "Dump HTTP headers and bodies - may contain sensitive info")
+	dumpAuth        = BoolP("dump-auth", "", false, "Dump HTTP headers with auth info")
+	skipVerify      = BoolP("no-check-certificate", "", false, "Do not verify the server SSL certificate. Insecure.")
+	AskPassword     = BoolP("ask-password", "", true, "Allow prompt for password for encrypted configuration.")
+	deleteBefore    = BoolP("delete-before", "", false, "When synchronizing, delete files on destination before transfering")
+	deleteDuring    = BoolP("delete-during", "", false, "When synchronizing, delete files during transfer (default)")
+	deleteAfter     = BoolP("delete-after", "", false, "When synchronizing, delete files on destination after transfering")
+	trackRenames    = BoolP("track-renames", "", false, "When synchronizing, track file renames and do a server side move if possible")
+	lowLevelRetries = IntP("low-level-retries", "", 10, "Number of low level retries to do.")
+	updateOlder     = BoolP("update", "u", false, "Skip files that are newer on the destination.")
+	noGzip          = BoolP("no-gzip-encoding", "", false, "Don't set Accept-Encoding: gzip.")
+	maxDepth        = IntP("max-depth", "", -1, "If set limits the recursion depth to this.")
+	ignoreSize      = BoolP("ignore-size", "", false, "Ignore size when skipping use mod-time or checksum.")
+	ignoreChecksum  = BoolP("ignore-checksum", "", false, "Skip post copy check of checksums.")
+	noTraverse      = BoolP("no-traverse", "", false, "Don't traverse destination file system on copy.")
+	noUpdateModTime = BoolP("no-update-modtime", "", false, "Don't update destination mod-time if files identical.")
+	backupDir       = StringP("backup-dir", "", "", "Make backups into hierarchy based in DIR.")
+	suffix          = StringP("suffix", "", "", "Suffix for use with --backup-dir.")
+	bwLimit         BwTimetable
+	bufferSize      SizeSuffix = 16 << 20
 
 	// Key to use for password en/decryption.
 	// When nil, no encryption will be used for saving.
@@ -97,100 +99,10 @@ var (
 )
 
 func init() {
-	pflag.VarP(&bwLimit, "bwlimit", "", "Bandwidth limit in kBytes/s, or use suffix b|k|M|G")
+	VarP(&bwLimit, "bwlimit", "", "Bandwidth limit in kBytes/s, or use suffix b|k|M|G or a full timetable.")
+	VarP(&bufferSize, "buffer-size", "", "Buffer size when copying files.")
 }
 
-// Turn SizeSuffix into a string and a suffix
-func (x SizeSuffix) string() (string, string) {
-	scaled := float64(0)
-	suffix := ""
-	switch {
-	case x < 0:
-		return "off", ""
-	case x == 0:
-		return "0", ""
-	case x < 1024:
-		scaled = float64(x)
-		suffix = ""
-	case x < 1024*1024:
-		scaled = float64(x) / 1024
-		suffix = "k"
-	case x < 1024*1024*1024:
-		scaled = float64(x) / 1024 / 1024
-		suffix = "M"
-	default:
-		scaled = float64(x) / 1024 / 1024 / 1024
-		suffix = "G"
-	}
-	if math.Floor(scaled) == scaled {
-		return fmt.Sprintf("%.0f", scaled), suffix
-	}
-	return fmt.Sprintf("%.3f", scaled), suffix
-}
-
-// String turns SizeSuffix into a string
-func (x SizeSuffix) String() string {
-	val, suffix := x.string()
-	return val + suffix
-}
-
-// Unit turns SizeSuffix into a string with a unit
-func (x SizeSuffix) Unit(unit string) string {
-	val, suffix := x.string()
-	if val == "off" {
-		return val
-	}
-	return val + " " + suffix + unit
-}
-
-// Set a SizeSuffix
-func (x *SizeSuffix) Set(s string) error {
-	if len(s) == 0 {
-		return errors.New("empty string")
-	}
-	if strings.ToLower(s) == "off" {
-		*x = -1
-		return nil
-	}
-	suffix := s[len(s)-1]
-	suffixLen := 1
-	var multiplier float64
-	switch suffix {
-	case '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.':
-		suffixLen = 0
-		multiplier = 1 << 10
-	case 'b', 'B':
-		multiplier = 1
-	case 'k', 'K':
-		multiplier = 1 << 10
-	case 'm', 'M':
-		multiplier = 1 << 20
-	case 'g', 'G':
-		multiplier = 1 << 30
-	default:
-		return errors.Errorf("bad suffix %q", suffix)
-	}
-	s = s[:len(s)-suffixLen]
-	value, err := strconv.ParseFloat(s, 64)
-	if err != nil {
-		return err
-	}
-	if value < 0 {
-		return errors.Errorf("size can't be negative %q", s)
-	}
-	value *= multiplier
-	*x = SizeSuffix(value)
-	return nil
-}
-
-// Type of the value
-func (x *SizeSuffix) Type() string {
-	return "int64"
-}
-
-// Check it satisfies the interface
-var _ pflag.Value = (*SizeSuffix)(nil)
-
 // crypt internals
 var (
 	cryptKey = []byte{
@@ -274,8 +186,7 @@ func MustReveal(x string) string {
 
 // ConfigInfo is filesystem config options
 type ConfigInfo struct {
-	Verbose            bool
-	Quiet              bool
+	LogLevel           LogLevel
 	DryRun             bool
 	CheckSum           bool
 	SizeOnly           bool
@@ -291,47 +202,141 @@ type ConfigInfo struct {
 	DumpAuth           bool
 	Filter             *Filter
 	InsecureSkipVerify bool // Skip server certificate verification
-	DeleteBefore       bool // Delete before checking
-	DeleteDuring       bool // Delete during checking/transfer
-	DeleteAfter        bool // Delete after successful transfer.
+	DeleteMode         DeleteMode
+	TrackRenames       bool // Track file renames.
 	LowLevelRetries    int
 	UpdateOlder        bool // Skip files that are newer on the destination
 	NoGzip             bool // Disable compression
 	MaxDepth           int
 	IgnoreSize         bool
+	IgnoreChecksum     bool
 	NoTraverse         bool
 	NoUpdateModTime    bool
 	DataRateUnit       string
+	BackupDir          string
+	Suffix             string
+	BufferSize         SizeSuffix
 }
 
-// Find the config directory
-func configHome() string {
-	// Find users home directory
+// Return the path to the configuration file
+func makeConfigPath() string {
+	// Find user's home directory
 	usr, err := user.Current()
+	var homedir string
 	if err == nil {
-		return usr.HomeDir
+		homedir = usr.HomeDir
+	} else {
+		// Fall back to reading $HOME - work around user.Current() not
+		// working for cross compiled binaries on OSX.
+		// https://github.com/golang/go/issues/6376
+		homedir = os.Getenv("HOME")
 	}
-	// Fall back to reading $HOME - work around user.Current() not
-	// working for cross compiled binaries on OSX.
-	// https://github.com/golang/go/issues/6376
-	home := os.Getenv("HOME")
-	if home != "" {
-		return home
+
+	// Possibly find the user's XDG config paths
+	// See XDG Base Directory specification
+	// https://specifications.freedesktop.org/basedir-spec/latest/
+	xdgdir := os.Getenv("XDG_CONFIG_HOME")
+	var xdgcfgdir string
+	if xdgdir != "" {
+		xdgcfgdir = filepath.Join(xdgdir, "rclone")
+	} else if homedir != "" {
+		xdgdir = filepath.Join(homedir, ".config")
+		xdgcfgdir = filepath.Join(xdgdir, "rclone")
 	}
-	ErrorLog(nil, "Couldn't find home directory or read HOME environment variable.")
-	ErrorLog(nil, "Defaulting to storing config in current directory.")
-	ErrorLog(nil, "Use -config flag to workaround.")
-	ErrorLog(nil, "Error was: %v", err)
-	return ""
+
+	// Use $XDG_CONFIG_HOME/rclone/rclone.conf if already existing
+	var xdgconf string
+	if xdgcfgdir != "" {
+		xdgconf = filepath.Join(xdgcfgdir, configFileName)
+		_, err := os.Stat(xdgconf)
+		if err == nil {
+			return xdgconf
+		}
+	}
+
+	// Use $HOME/.rclone.conf if already existing
+	var homeconf string
+	if homedir != "" {
+		homeconf = filepath.Join(homedir, hiddenConfigFileName)
+		_, err := os.Stat(homeconf)
+		if err == nil {
+			return homeconf
+		}
+	}
+
+	// Try to create $XDG_CONFIG_HOME/rclone/rclone.conf
+	if xdgconf != "" {
+		// xdgconf != "" implies xdgcfgdir != ""
+		err := os.MkdirAll(xdgcfgdir, os.ModePerm)
+		if err == nil {
+			return xdgconf
+		}
+	}
+
+	// Try to create $HOME/.rclone.conf
+	if homeconf != "" {
+		return homeconf
+	}
+
+	// Default to ./.rclone.conf (current working directory)
+	Errorf(nil, "Couldn't find home directory or read HOME or XDG_CONFIG_HOME environment variables.")
+	Errorf(nil, "Defaulting to storing config in current directory.")
+	Errorf(nil, "Use -config flag to workaround.")
+	Errorf(nil, "Error was: %v", err)
+	return hiddenConfigFileName
 }
 
+// DeleteMode describes the possible delete modes in the config
+type DeleteMode byte
+
+// DeleteMode constants
+const (
+	DeleteModeOff DeleteMode = iota
+	DeleteModeBefore
+	DeleteModeDuring
+	DeleteModeAfter
+	DeleteModeOnly
+	DeleteModeDefault = DeleteModeAfter
+)
+
 // LoadConfig loads the config file
 func LoadConfig() {
 	// Read some flags if set
 	//
 	// FIXME read these from the config file too
-	Config.Verbose = *verbose
-	Config.Quiet = *quiet
+	Config.LogLevel = LogLevelNotice
+	if *verbose >= 2 {
+		Config.LogLevel = LogLevelDebug
+	} else if *verbose >= 1 {
+		Config.LogLevel = LogLevelInfo
+	}
+	if *quiet {
+		if *verbose > 0 {
+			log.Fatalf("Can't set -v and -q")
+		}
+		Config.LogLevel = LogLevelError
+	}
+	logLevelFlag := pflag.Lookup("log-level")
+	if logLevelFlag != nil && logLevelFlag.Changed {
+		if *verbose > 0 {
+			log.Fatalf("Can't set -v and --log-level")
+		}
+		if *quiet {
+			log.Fatalf("Can't set -q and --log-level")
+		}
+		switch strings.ToUpper(*logLevel) {
+		case "ERROR":
+			Config.LogLevel = LogLevelError
+		case "NOTICE":
+			Config.LogLevel = LogLevelNotice
+		case "INFO":
+			Config.LogLevel = LogLevelInfo
+		case "DEBUG":
+			Config.LogLevel = LogLevelDebug
+		default:
+			log.Fatalf("Unknown --log-level %q", *logLevel)
+		}
+	}
 	Config.ModifyWindow = *modifyWindow
 	Config.Checkers = *checkers
 	Config.Transfers = *transfers
@@ -351,35 +356,45 @@ func LoadConfig() {
 	Config.NoGzip = *noGzip
 	Config.MaxDepth = *maxDepth
 	Config.IgnoreSize = *ignoreSize
+	Config.IgnoreChecksum = *ignoreChecksum
 	Config.NoTraverse = *noTraverse
 	Config.NoUpdateModTime = *noUpdateModTime
+	Config.BackupDir = *backupDir
+	Config.Suffix = *suffix
+	Config.BufferSize = bufferSize
 
 	ConfigPath = *configFile
 
-	Config.DeleteBefore = *deleteBefore
-	Config.DeleteDuring = *deleteDuring
-	Config.DeleteAfter = *deleteAfter
+	Config.TrackRenames = *trackRenames
 
 	switch {
 	case *deleteBefore && (*deleteDuring || *deleteAfter),
 		*deleteDuring && *deleteAfter:
 		log.Fatalf(`Only one of --delete-before, --delete-during or --delete-after can be used.`)
-
-	// If none are specified, use "during".
-	case !*deleteBefore && !*deleteDuring && !*deleteAfter:
-		Config.DeleteDuring = true
+	case *deleteBefore:
+		Config.DeleteMode = DeleteModeBefore
+	case *deleteDuring:
+		Config.DeleteMode = DeleteModeDuring
+	case *deleteAfter:
+		Config.DeleteMode = DeleteModeAfter
+	default:
+		Config.DeleteMode = DeleteModeDefault
 	}
 
 	if Config.IgnoreSize && Config.SizeOnly {
 		log.Fatalf(`Can't use --size-only and --ignore-size together.`)
 	}
 
+	if Config.Suffix != "" && Config.BackupDir == "" {
+		log.Fatalf(`Can only use --suffix with --backup-dir.`)
+	}
+
 	// Load configuration file.
 	var err error
-	ConfigFile, err = loadConfigFile()
+	configData, err = loadConfigFile()
 	if err == errorConfigFileNotFound {
-		Log(nil, "Config file %q not found - using defaults", ConfigPath)
-		ConfigFile, _ = goconfig.LoadFromReader(&bytes.Buffer{})
+		Logf(nil, "Config file %q not found - using defaults", ConfigPath)
+		configData, _ = goconfig.LoadFromReader(&bytes.Buffer{})
 	} else if err != nil {
 		log.Fatalf("Failed to load config file %q: %v", ConfigPath, err)
 	}
@@ -392,6 +407,9 @@ func LoadConfig() {
 
 	// Start the token bucket limiter
 	startTokenBucket()
+
+	// Start the bandwidth update ticker
+	startTokenTicker()
 }
 
 var errorConfigFileNotFound = errors.New("config file not found")
@@ -450,7 +468,7 @@ func loadConfigFile() (*goconfig.ConfigFile, error) {
 				fmt.Println("Using RCLONE_CONFIG_PASS returned:", err)
 				envpw = ""
 			} else {
-				Debug(nil, "Using RCLONE_CONFIG_PASS password.")
+				Debugf(nil, "Using RCLONE_CONFIG_PASS password.")
 			}
 		}
 		if len(configKey) == 0 {
@@ -474,7 +492,7 @@ func loadConfigFile() (*goconfig.ConfigFile, error) {
 		}
 
 		// Retry
-		ErrorLog(nil, "Couldn't decrypt configuration, most likely wrong password.")
+		Errorf(nil, "Couldn't decrypt configuration, most likely wrong password.")
 		configKey = nil
 		envpw = ""
 	}
@@ -572,18 +590,18 @@ func changeConfigPassword() {
 // if configKey has been set, the file will be encrypted.
 func SaveConfig() {
 	if len(configKey) == 0 {
-		err := goconfig.SaveConfigFile(ConfigFile, ConfigPath)
+		err := goconfig.SaveConfigFile(configData, ConfigPath)
 		if err != nil {
 			log.Fatalf("Failed to save config file: %v", err)
 		}
 		err = os.Chmod(ConfigPath, 0600)
 		if err != nil {
-			ErrorLog(nil, "Failed to set permissions on config file: %v", err)
+			Errorf(nil, "Failed to set permissions on config file: %v", err)
 		}
 		return
 	}
 	var buf bytes.Buffer
-	err := goconfig.SaveConfigData(ConfigFile, &buf)
+	err := goconfig.SaveConfigData(configData, &buf)
 	if err != nil {
 		log.Fatalf("Failed to save config file: %v", err)
 	}
@@ -625,7 +643,7 @@ func SaveConfig() {
 
 	err = os.Chmod(ConfigPath, 0600)
 	if err != nil {
-		ErrorLog(nil, "Failed to set permissions on config file: %v", err)
+		Errorf(nil, "Failed to set permissions on config file: %v", err)
 	}
 }
 
@@ -634,7 +652,7 @@ func SaveConfig() {
 // disk first and overwrites the given value only.
 func ConfigSetValueAndSave(name, key, value string) (err error) {
 	// Set the value in config in case we fail to reload it
-	ConfigFile.SetValue(name, key, value)
+	configData.SetValue(name, key, value)
 	// Reload the config file
 	reloadedConfigFile, err := loadConfigFile()
 	if err == errorConfigFileNotFound {
@@ -649,7 +667,7 @@ func ConfigSetValueAndSave(name, key, value string) (err error) {
 		return err
 	}
 	// Update the config file with the reloaded version
-	ConfigFile = reloadedConfigFile
+	configData = reloadedConfigFile
 	// Set the value in the reloaded version
 	reloadedConfigFile.SetValue(name, key, value)
 	// Save it again
@@ -659,7 +677,7 @@ func ConfigSetValueAndSave(name, key, value string) (err error) {
 
 // ShowRemotes shows an overview of the config file
 func ShowRemotes() {
-	remotes := ConfigFile.GetSectionList()
+	remotes := configData.GetSectionList()
 	if len(remotes) == 0 {
 		return
 	}
@@ -667,13 +685,13 @@ func ShowRemotes() {
 	fmt.Printf("%-20s %s\n", "Name", "Type")
 	fmt.Printf("%-20s %s\n", "====", "====")
 	for _, remote := range remotes {
-		fmt.Printf("%-20s %s\n", remote, ConfigFile.MustValue(remote, "type"))
+		fmt.Printf("%-20s %s\n", remote, ConfigFileGet(remote, "type"))
 	}
 }
 
 // ChooseRemote chooses a remote name
 func ChooseRemote() string {
-	remotes := ConfigFile.GetSectionList()
+	remotes := configData.GetSectionList()
 	sort.Strings(remotes)
 	return Choose("remote", remotes, nil, false)
 }
@@ -717,11 +735,11 @@ func Confirm() bool {
 
 // Choose one of the defaults or type a new string if newOk is set
 func Choose(what string, defaults, help []string, newOk bool) string {
-	fmt.Printf("Choose a number from below")
+	valueDescripton := "an existing"
 	if newOk {
-		fmt.Printf(", or type in your own value")
+		valueDescripton = "your own"
 	}
-	fmt.Println()
+	fmt.Printf("Choose a number from below, or type in %s value\n", valueDescripton)
 	for i, text := range defaults {
 		var lines []string
 		if help != nil {
@@ -760,6 +778,11 @@ func Choose(what string, defaults, help []string, newOk bool) string {
 			if newOk {
 				return result
 			}
+			for _, v := range defaults {
+				if result == v {
+					return result
+				}
+			}
 			continue
 		}
 		if i >= 1 && i <= len(defaults) {
@@ -792,7 +815,7 @@ func ShowRemote(name string) {
 	fmt.Printf("--------------------\n")
 	fmt.Printf("[%s]\n", name)
 	fs := MustFindByName(name)
-	for _, key := range ConfigFile.GetKeyList(name) {
+	for _, key := range configData.GetKeyList(name) {
 		isPassword := false
 		for _, option := range fs.Options {
 			if option.Name == key && option.IsPassword {
@@ -800,7 +823,7 @@ func ShowRemote(name string) {
 				break
 			}
 		}
-		value := ConfigFile.MustValue(name, key)
+		value := ConfigFileGet(name, key)
 		if isPassword && value != "" {
 			fmt.Printf("%s = *** ENCRYPTED ***\n", key)
 		} else {
@@ -819,10 +842,10 @@ func OkRemote(name string) bool {
 	case 'e':
 		return false
 	case 'd':
-		ConfigFile.DeleteSection(name)
+		configData.DeleteSection(name)
 		return true
 	default:
-		ErrorLog(nil, "Bad choice %c", i)
+		Errorf(nil, "Bad choice %c", i)
 	}
 	return false
 }
@@ -830,7 +853,7 @@ func OkRemote(name string) bool {
 // MustFindByName finds the RegInfo for the remote name passed in or
 // exits with a fatal error.
 func MustFindByName(name string) *RegInfo {
-	fsType := ConfigFile.MustValue(name, "type")
+	fsType := ConfigFileGet(name, "type")
 	if fsType == "" {
 		log.Fatalf("Couldn't find type of fs for %q", name)
 	}
@@ -881,7 +904,7 @@ func ChooseOption(o *Option) string {
 		case 'n':
 			return ""
 		default:
-			ErrorLog(nil, "Bad choice %c", i)
+			Errorf(nil, "Bad choice %c", i)
 		}
 		return MustObscure(password)
 	}
@@ -915,13 +938,32 @@ func fsOption() *Option {
 	return o
 }
 
+// NewRemoteName asks the user for a name for a remote
+func NewRemoteName() (name string) {
+	for {
+		fmt.Printf("name> ")
+		name = ReadLine()
+		parts := matcher.FindStringSubmatch(name + ":")
+		switch {
+		case name == "":
+			fmt.Printf("Can't use empty name.\n")
+		case isDriveLetter(name):
+			fmt.Printf("Can't use %q as it can be confused a drive letter.\n", name)
+		case parts == nil:
+			fmt.Printf("Can't use %q as it has invalid characters in it.\n", name)
+		default:
+			return name
+		}
+	}
+}
+
 // NewRemote make a new remote from its name
 func NewRemote(name string) {
 	newType := ChooseOption(fsOption())
-	ConfigFile.SetValue(name, "type", newType)
+	configData.SetValue(name, "type", newType)
 	fs := MustFind(newType)
 	for _, option := range fs.Options {
-		ConfigFile.SetValue(name, option.Name, ChooseOption(&option))
+		configData.SetValue(name, option.Name, ChooseOption(&option))
 	}
 	RemoteConfig(name)
 	if OkRemote(name) {
@@ -938,12 +980,12 @@ func EditRemote(fs *RegInfo, name string) {
 	for {
 		for _, option := range fs.Options {
 			key := option.Name
-			value := ConfigFile.MustValue(name, key)
+			value := ConfigFileGet(name, key)
 			fmt.Printf("Value %q = %q\n", key, value)
 			fmt.Printf("Edit? (y/n)>\n")
 			if Confirm() {
 				newValue := ChooseOption(&option)
-				ConfigFile.SetValue(name, key, newValue)
+				configData.SetValue(name, key, newValue)
 			}
 		}
 		RemoteConfig(name)
@@ -956,15 +998,41 @@ func EditRemote(fs *RegInfo, name string) {
 
 // DeleteRemote gets the user to delete a remote
 func DeleteRemote(name string) {
-	ConfigFile.DeleteSection(name)
+	configData.DeleteSection(name)
+	SaveConfig()
+}
+
+// copyRemote asks the user for a new remote name and copies name into
+// it
+func copyRemote(name string) {
+	newName := NewRemoteName()
+	// Copy the keys
+	for _, key := range configData.GetKeyList(name) {
+		value := configData.MustValue(name, key, "")
+		configData.SetValue(newName, key, value)
+	}
+}
+
+// RenameRemote renames a config section
+func RenameRemote(name string) {
+	fmt.Printf("Enter new name for %q remote.\n", name)
+	copyRemote(name)
+	configData.DeleteSection(name)
+	SaveConfig()
+}
+
+// CopyRemote copies a config section
+func CopyRemote(name string) {
+	fmt.Printf("Enter name for copy of %q remote.\n", name)
+	copyRemote(name)
 	SaveConfig()
 }
 
 // EditConfig edits the config file interactively
 func EditConfig() {
 	for {
-		haveRemotes := len(ConfigFile.GetSectionList()) != 0
-		what := []string{"eEdit existing remote", "nNew remote", "dDelete remote", "sSet configuration password", "qQuit config"}
+		haveRemotes := len(configData.GetSectionList()) != 0
+		what := []string{"eEdit existing remote", "nNew remote", "dDelete remote", "rRename remote", "cCopy remote", "sSet configuration password", "qQuit config"}
 		if haveRemotes {
 			fmt.Printf("Current remotes:\n\n")
 			ShowRemotes()
@@ -979,26 +1047,14 @@ func EditConfig() {
 			fs := MustFindByName(name)
 			EditRemote(fs, name)
 		case 'n':
-		nameLoop:
-			for {
-				fmt.Printf("name> ")
-				name := ReadLine()
-				parts := matcher.FindStringSubmatch(name + ":")
-				switch {
-				case name == "":
-					fmt.Printf("Can't use empty name\n")
-				case isDriveLetter(name):
-					fmt.Printf("Can't use %q as it can be confused a drive letter\n", name)
-				case len(parts) != 3 || parts[2] != "":
-					fmt.Printf("Can't use %q as it has invalid characters in it %v\n", name, parts)
-				default:
-					NewRemote(name)
-					break nameLoop
-				}
-			}
+			NewRemote(NewRemoteName())
 		case 'd':
 			name := ChooseRemote()
 			DeleteRemote(name)
+		case 'r':
+			RenameRemote(ChooseRemote())
+		case 'c':
+			CopyRemote(ChooseRemote())
 		case 's':
 			SetPassword()
 		case 'q':
@@ -1070,10 +1126,87 @@ func Authorize(args []string) {
 	defer DeleteRemote(name)
 
 	// Indicate that we want fully automatic configuration.
-	ConfigFile.SetValue(name, ConfigAutomatic, "yes")
+	configData.SetValue(name, ConfigAutomatic, "yes")
 	if len(args) == 3 {
-		ConfigFile.SetValue(name, ConfigClientID, args[1])
-		ConfigFile.SetValue(name, ConfigClientSecret, args[2])
+		configData.SetValue(name, ConfigClientID, args[1])
+		configData.SetValue(name, ConfigClientSecret, args[2])
 	}
 	fs.Config(name)
 }
+
+// configToEnv converts an config section and name, eg ("myremote",
+// "ignore-size") into an environment name
+// "RCLONE_CONFIG_MYREMOTE_IGNORE_SIZE"
+func configToEnv(section, name string) string {
+	return "RCLONE_CONFIG_" + strings.ToUpper(strings.Replace(section+"_"+name, "-", "_", -1))
+}
+
+// ConfigFileGet gets the config key under section returning the
+// default or empty string if not set.
+//
+// It looks up defaults in the environment if they are present
+func ConfigFileGet(section, key string, defaultVal ...string) string {
+	envKey := configToEnv(section, key)
+	newValue, found := os.LookupEnv(envKey)
+	if found {
+		defaultVal = []string{newValue}
+	}
+	return configData.MustValue(section, key, defaultVal...)
+}
+
+// ConfigFileGetBool gets the config key under section returning the
+// default or false if not set.
+//
+// It looks up defaults in the environment if they are present
+func ConfigFileGetBool(section, key string, defaultVal ...bool) bool {
+	envKey := configToEnv(section, key)
+	newValue, found := os.LookupEnv(envKey)
+	if found {
+		newBool, err := strconv.ParseBool(newValue)
+		if err != nil {
+			Errorf(nil, "Couldn't parse %q into bool - ignoring: %v", envKey, err)
+		} else {
+			defaultVal = []bool{newBool}
+		}
+	}
+	return configData.MustBool(section, key, defaultVal...)
+}
+
+// ConfigFileGetInt gets the config key under section returning the
+// default or 0 if not set.
+//
+// It looks up defaults in the environment if they are present
+func ConfigFileGetInt(section, key string, defaultVal ...int) int {
+	envKey := configToEnv(section, key)
+	newValue, found := os.LookupEnv(envKey)
+	if found {
+		newInt, err := strconv.Atoi(newValue)
+		if err != nil {
+			Errorf(nil, "Couldn't parse %q into int - ignoring: %v", envKey, err)
+		} else {
+			defaultVal = []int{newInt}
+		}
+	}
+	return configData.MustInt(section, key, defaultVal...)
+}
+
+// ConfigFileSet sets the key in section to value.  It doesn't save
+// the config file.
+func ConfigFileSet(section, key, value string) {
+	configData.SetValue(section, key, value)
+}
+
+var matchEnv = regexp.MustCompile(`^RCLONE_CONFIG_(.*?)_TYPE=.*$`)
+
+// ConfigFileSections returns the sections in the config file
+// including any defined by environment variables.
+func ConfigFileSections() []string {
+	sections := configData.GetSectionList()
+	for _, item := range os.Environ() {
+		matches := matchEnv.FindStringSubmatch(item)
+		if len(matches) == 2 {
+			sections = append(sections, strings.ToLower(matches[1]))
+		}
+	}
+	return sections
+}
