@@ -3,6 +3,7 @@
 package fs
 
 import (
+	"bytes"
 	"fmt"
 	"io"
 	"log"
@@ -32,12 +33,12 @@ func CalculateModifyWindow(fs ...Fs) {
 				Config.ModifyWindow = precision
 			}
 			if precision == ModTimeNotSupported {
-				Debug(f, "Modify window not supported")
+				Infof(f, "Modify window not supported")
 				return
 			}
 		}
 	}
-	Debug(fs[0], "Modify window is %s", Config.ModifyWindow)
+	Infof(fs[0], "Modify window is %s", Config.ModifyWindow)
 }
 
 // HashEquals checks to see if src == dst, but ignores empty strings
@@ -64,7 +65,7 @@ func HashEquals(src, dst string) bool {
 // If an error is returned it will return equal as false
 func CheckHashes(src, dst Object) (equal bool, hash HashType, err error) {
 	common := src.Fs().Hashes().Overlap(dst.Fs().Hashes())
-	// Debug(nil, "Shared hashes: %v", common)
+	// Debugf(nil, "Shared hashes: %v", common)
 	if common.Count() == 0 {
 		return true, HashNone, nil
 	}
@@ -72,7 +73,7 @@ func CheckHashes(src, dst Object) (equal bool, hash HashType, err error) {
 	srcHash, err := src.Hash(hash)
 	if err != nil {
 		Stats.Error()
-		ErrorLog(src, "Failed to calculate src hash: %v", err)
+		Errorf(src, "Failed to calculate src hash: %v", err)
 		return false, hash, err
 	}
 	if srcHash == "" {
@@ -81,12 +82,16 @@ func CheckHashes(src, dst Object) (equal bool, hash HashType, err error) {
 	dstHash, err := dst.Hash(hash)
 	if err != nil {
 		Stats.Error()
-		ErrorLog(dst, "Failed to calculate dst hash: %v", err)
+		Errorf(dst, "Failed to calculate dst hash: %v", err)
 		return false, hash, err
 	}
 	if dstHash == "" {
 		return true, HashNone, nil
 	}
+	if srcHash != dstHash {
+		Debugf(src, "%v = %s (%v)", hash, srcHash, src.Fs())
+		Debugf(dst, "%v = %s (%v)", hash, dstHash, dst.Fs())
+	}
 	return srcHash == dstHash, hash, nil
 }
 
@@ -109,38 +114,42 @@ func CheckHashes(src, dst Object) (equal bool, hash HashType, err error) {
 // Otherwise the file is considered to be not equal including if there
 // were errors reading info.
 func Equal(src, dst Object) bool {
+	return equal(src, dst, Config.SizeOnly, Config.CheckSum)
+}
+
+func equal(src, dst Object, sizeOnly, checkSum bool) bool {
 	if !Config.IgnoreSize {
 		if src.Size() != dst.Size() {
-			Debug(src, "Sizes differ")
+			Debugf(src, "Sizes differ")
 			return false
 		}
 	}
-	if Config.SizeOnly {
-		Debug(src, "Sizes identical")
+	if sizeOnly {
+		Debugf(src, "Sizes identical")
 		return true
 	}
 
 	// Assert: Size is equal or being ignored
 
 	// If checking checksum and not modtime
-	if Config.CheckSum {
+	if checkSum {
 		// Check the hash
 		same, hash, _ := CheckHashes(src, dst)
 		if !same {
-			Debug(src, "%v differ", hash)
+			Debugf(src, "%v differ", hash)
 			return false
 		}
 		if hash == HashNone {
-			Debug(src, "Size of src and dst objects identical")
+			Debugf(src, "Size of src and dst objects identical")
 		} else {
-			Debug(src, "Size and %v of src and dst objects identical", hash)
+			Debugf(src, "Size and %v of src and dst objects identical", hash)
 		}
 		return true
 	}
 
 	// Sizes the same so check the mtime
 	if Config.ModifyWindow == ModTimeNotSupported {
-		Debug(src, "Sizes identical")
+		Debugf(src, "Sizes identical")
 		return true
 	}
 	srcModTime := src.ModTime()
@@ -148,16 +157,16 @@ func Equal(src, dst Object) bool {
 	dt := dstModTime.Sub(srcModTime)
 	ModifyWindow := Config.ModifyWindow
 	if dt < ModifyWindow && dt > -ModifyWindow {
-		Debug(src, "Size and modification time the same (differ by %s, within tolerance %s)", dt, ModifyWindow)
+		Debugf(src, "Size and modification time the same (differ by %s, within tolerance %s)", dt, ModifyWindow)
 		return true
 	}
 
-	Debug(src, "Modification times differ by %s: %v, %v", dt, srcModTime, dstModTime)
+	Debugf(src, "Modification times differ by %s: %v, %v", dt, srcModTime, dstModTime)
 
 	// Check if the hashes are the same
 	same, hash, _ := CheckHashes(src, dst)
 	if !same {
-		Debug(src, "%v differ", hash)
+		Debugf(src, "%v differ", hash)
 		return false
 	}
 	if hash == HashNone {
@@ -167,17 +176,21 @@ func Equal(src, dst Object) bool {
 
 	// mod time differs but hash is the same to reset mod time if required
 	if !Config.NoUpdateModTime {
-		// Size and hash the same but mtime different so update the
-		// mtime of the dst object here
-		err := dst.SetModTime(srcModTime)
-		if err == ErrorCantSetModTime {
-			Debug(src, "src and dst identical but can't set mod time without re-uploading")
-			return false
-		} else if err != nil {
-			Stats.Error()
-			ErrorLog(dst, "Failed to set modification time: %v", err)
+		if Config.DryRun {
+			Logf(src, "Not updating modification time as --dry-run")
 		} else {
-			Debug(src, "Updated modification time in destination")
+			// Size and hash the same but mtime different so update the
+			// mtime of the dst object here
+			err := dst.SetModTime(srcModTime)
+			if err == ErrorCantSetModTime {
+				Debugf(src, "src and dst identical but can't set mod time without re-uploading")
+				return false
+			} else if err != nil {
+				Stats.Error()
+				Errorf(dst, "Failed to set modification time: %v", err)
+			} else {
+				Infof(src, "Updated modification time in destination")
+			}
 		}
 	}
 	return true
@@ -198,7 +211,7 @@ func MimeType(o ObjectInfo) (mimeType string) {
 	// Read the MimeType from the optional interface if available
 	if do, ok := o.(MimeTyper); ok {
 		mimeType = do.MimeType()
-		Debug(o, "Read MimeType as %q", mimeType)
+		// Debugf(o, "Read MimeType as %q", mimeType)
 		if mimeType != "" {
 			return mimeType
 		}
@@ -213,10 +226,10 @@ func removeFailedCopy(dst Object) bool {
 	if dst == nil {
 		return false
 	}
-	Debug(dst, "Removing failed copy")
+	Infof(dst, "Removing failed copy")
 	removeErr := dst.Remove()
 	if removeErr != nil {
-		Debug(dst, "Failed to remove failed copy: %s", removeErr)
+		Infof(dst, "Failed to remove failed copy: %s", removeErr)
 		return false
 	}
 	return true
@@ -233,11 +246,23 @@ func (o *overrideRemoteObject) Remote() string {
 	return o.remote
 }
 
+// MimeType returns the mime type of the underlying object or "" if it
+// can't be worked out
+func (o *overrideRemoteObject) MimeType() string {
+	if do, ok := o.Object.(MimeTyper); ok {
+		return do.MimeType()
+	}
+	return ""
+}
+
+// Check interface is satisfied
+var _ MimeTyper = (*overrideRemoteObject)(nil)
+
 // Copy src object to dst or f if nil.  If dst is nil then it uses
 // remote as the name of the new object.
 func Copy(f Fs, dst Object, remote string, src Object) (err error) {
 	if Config.DryRun {
-		Log(src, "Not copying as --dry-run")
+		Logf(src, "Not copying as --dry-run")
 		return nil
 	}
 	maxTries := Config.LowLevelRetries
@@ -248,9 +273,9 @@ func Copy(f Fs, dst Object, remote string, src Object) (err error) {
 		// Try server side copy first - if has optional interface and
 		// is same underlying remote
 		actionTaken = "Copied (server side copy)"
-		if fCopy, ok := f.(Copier); ok && src.Fs().Name() == f.Name() {
+		if doCopy := f.Features().Copy; doCopy != nil && SameConfig(src.Fs(), f) {
 			var newDst Object
-			newDst, err = fCopy.Copy(src, remote)
+			newDst, err = doCopy(src, remote)
 			if err == nil {
 				dst = newDst
 			}
@@ -264,9 +289,12 @@ func Copy(f Fs, dst Object, remote string, src Object) (err error) {
 			if err != nil {
 				err = errors.Wrap(err, "failed to open source object")
 			} else {
-				in := NewAccountWithBuffer(in0, src) // account and buffer the transfer
-
-				wrappedSrc := &overrideRemoteObject{Object: src, remote: remote}
+				in := NewAccount(in0, src).WithBuffer() // account and buffer the transfer
+				var wrappedSrc ObjectInfo = src
+				// We try to pass the original object if possible
+				if src.Remote() != remote {
+					wrappedSrc = &overrideRemoteObject{Object: src, remote: remote}
+				}
 				if doUpdate {
 					actionTaken = "Copied (replaced existing)"
 					err = dst.Update(in, wrappedSrc)
@@ -286,7 +314,7 @@ func Copy(f Fs, dst Object, remote string, src Object) (err error) {
 		}
 		// Retry if err returned a retry error
 		if IsRetryError(err) || ShouldRetry(err) {
-			Debug(src, "Received error: %v - low level retry %d/%d", err, tries, maxTries)
+			Debugf(src, "Received error: %v - low level retry %d/%d", err, tries, maxTries)
 			continue
 		}
 		// otherwise finish
@@ -294,7 +322,7 @@ func Copy(f Fs, dst Object, remote string, src Object) (err error) {
 	}
 	if err != nil {
 		Stats.Error()
-		ErrorLog(src, "Failed to copy: %v", err)
+		Errorf(src, "Failed to copy: %v", err)
 		return err
 	}
 
@@ -302,7 +330,7 @@ func Copy(f Fs, dst Object, remote string, src Object) (err error) {
 	if !Config.IgnoreSize && src.Size() != dst.Size() {
 		Stats.Error()
 		err = errors.Errorf("corrupted on transfer: sizes differ %d vs %d", src.Size(), dst.Size())
-		ErrorLog(dst, "%v", err)
+		Errorf(dst, "%v", err)
 		removeFailedCopy(dst)
 		return err
 	}
@@ -311,7 +339,7 @@ func Copy(f Fs, dst Object, remote string, src Object) (err error) {
 	// TODO(klauspost): This could be extended, so we always create a hash type matching
 	// the destination, and calculate it while sending.
 	common := src.Fs().Hashes().Overlap(dst.Fs().Hashes())
-	// Debug(src, "common hashes: %v", common)
+	// Debugf(src, "common hashes: %v", common)
 	if !Config.SizeOnly && common.Count() > 0 {
 		// Get common hash type
 		hashType := common.GetOne()
@@ -320,24 +348,24 @@ func Copy(f Fs, dst Object, remote string, src Object) (err error) {
 		srcSum, err = src.Hash(hashType)
 		if err != nil {
 			Stats.Error()
-			ErrorLog(src, "Failed to read src hash: %v", err)
+			Errorf(src, "Failed to read src hash: %v", err)
 		} else if srcSum != "" {
 			var dstSum string
 			dstSum, err = dst.Hash(hashType)
 			if err != nil {
 				Stats.Error()
-				ErrorLog(dst, "Failed to read hash: %v", err)
-			} else if !HashEquals(srcSum, dstSum) {
+				Errorf(dst, "Failed to read hash: %v", err)
+			} else if !Config.IgnoreChecksum && !HashEquals(srcSum, dstSum) {
 				Stats.Error()
 				err = errors.Errorf("corrupted on transfer: %v hash differ %q vs %q", hashType, srcSum, dstSum)
-				ErrorLog(dst, "%v", err)
+				Errorf(dst, "%v", err)
 				removeFailedCopy(dst)
 				return err
 			}
 		}
 	}
 
-	Debug(src, actionTaken)
+	Infof(src, actionTaken)
 	return err
 }
 
@@ -345,11 +373,11 @@ func Copy(f Fs, dst Object, remote string, src Object) (err error) {
 // remote as the name of the new object.
 func Move(fdst Fs, dst Object, remote string, src Object) (err error) {
 	if Config.DryRun {
-		Log(src, "Not moving as --dry-run")
+		Logf(src, "Not moving as --dry-run")
 		return nil
 	}
 	// See if we have Move available
-	if do, ok := fdst.(Mover); ok && src.Fs().Name() == fdst.Name() {
+	if doMove := fdst.Features().Move; doMove != nil && SameConfig(src.Fs(), fdst) {
 		// Delete destination if it exists
 		if dst != nil {
 			err = DeleteFile(dst)
@@ -358,49 +386,88 @@ func Move(fdst Fs, dst Object, remote string, src Object) (err error) {
 			}
 		}
 		// Move dst <- src
-		_, err := do.Move(src, remote)
+		_, err := doMove(src, remote)
 		switch err {
 		case nil:
-			Debug(src, "Moved (server side)")
+			Infof(src, "Moved (server side)")
 			return nil
 		case ErrorCantMove:
-			Debug(src, "Can't move, switching to copy")
+			Debugf(src, "Can't move, switching to copy")
 		default:
 			Stats.Error()
-			ErrorLog(dst, "Couldn't move: %v", err)
+			Errorf(src, "Couldn't move: %v", err)
 			return err
 		}
 	}
 	// Move not found or didn't work so copy dst <- src
 	err = Copy(fdst, dst, remote, src)
 	if err != nil {
-		ErrorLog(src, "Not deleting source as copy failed: %v", err)
+		Errorf(src, "Not deleting source as copy failed: %v", err)
 		return err
 	}
 	// Delete src if no error on copy
 	return DeleteFile(src)
 }
 
-// DeleteFile deletes a single file respecting --dry-run and accumulating stats and errors.
-func DeleteFile(dst Object) (err error) {
-	if Config.DryRun {
-		Log(dst, "Not deleting as --dry-run")
-	} else {
-		Stats.Checking(dst.Remote())
-		err = dst.Remove()
-		Stats.DoneChecking(dst.Remote())
-		if err != nil {
-			Stats.Error()
-			ErrorLog(dst, "Couldn't delete: %v", err)
-		} else {
-			Debug(dst, "Deleted")
-		}
+// CanServerSideMove returns true if fdst support server side moves or
+// server side copies
+//
+// Some remotes simulate rename by server-side copy and delete, so include
+// remotes that implements either Mover or Copier.
+func CanServerSideMove(fdst Fs) bool {
+	canMove := fdst.Features().Move != nil
+	canCopy := fdst.Features().Copy != nil
+	return canMove || canCopy
+}
+
+// deleteFileWithBackupDir deletes a single file respecting --dry-run
+// and accumulating stats and errors.
+//
+// If backupDir is set then it moves the file to there instead of
+// deleting
+func deleteFileWithBackupDir(dst Object, backupDir Fs) (err error) {
+	Stats.Checking(dst.Remote())
+	action, actioned, actioning := "delete", "Deleted", "deleting"
+	if backupDir != nil {
+		action, actioned, actioning = "move into backup dir", "Moved into backup dir", "moving into backup dir"
 	}
+	if Config.DryRun {
+		Logf(dst, "Not %s as --dry-run", actioning)
+	} else if backupDir != nil {
+		if !SameConfig(dst.Fs(), backupDir) {
+			err = errors.New("parameter to --backup-dir has to be on the same remote as destination")
+		} else {
+			remoteWithSuffix := dst.Remote() + Config.Suffix
+			overwritten, _ := backupDir.NewObject(remoteWithSuffix)
+			err = Move(backupDir, overwritten, remoteWithSuffix, dst)
+		}
+	} else {
+		err = dst.Remove()
+	}
+	if err != nil {
+		Stats.Error()
+		Errorf(dst, "Couldn't %s: %v", action, err)
+	} else {
+		Infof(dst, actioned)
+	}
+	Stats.DoneChecking(dst.Remote())
 	return err
 }
 
-// DeleteFiles removes all the files passed in the channel
-func DeleteFiles(toBeDeleted ObjectsChan) error {
+// DeleteFile deletes a single file respecting --dry-run and accumulating stats and errors.
+//
+// If useBackupDir is set and --backup-dir is in effect then it moves
+// the file to there instead of deleting
+func DeleteFile(dst Object) (err error) {
+	return deleteFileWithBackupDir(dst, nil)
+}
+
+// deleteFilesWithBackupDir removes all the files passed in the
+// channel
+//
+// If backupDir is set the files will be placed into that directory
+// instead of being deleted.
+func deleteFilesWithBackupDir(toBeDeleted ObjectsChan, backupDir Fs) error {
 	var wg sync.WaitGroup
 	wg.Add(Config.Transfers)
 	var errorCount int32
@@ -408,14 +475,14 @@ func DeleteFiles(toBeDeleted ObjectsChan) error {
 		go func() {
 			defer wg.Done()
 			for dst := range toBeDeleted {
-				err := DeleteFile(dst)
+				err := deleteFileWithBackupDir(dst, backupDir)
 				if err != nil {
 					atomic.AddInt32(&errorCount, 1)
 				}
 			}
 		}()
 	}
-	Log(nil, "Waiting for deletions to finish")
+	Infof(nil, "Waiting for deletions to finish")
 	wg.Wait()
 	if errorCount > 0 {
 		return errors.Errorf("failed to delete %d files", errorCount)
@@ -423,6 +490,11 @@ func DeleteFiles(toBeDeleted ObjectsChan) error {
 	return nil
 }
 
+// DeleteFiles removes all the files passed in the channel
+func DeleteFiles(toBeDeleted ObjectsChan) error {
+	return deleteFilesWithBackupDir(toBeDeleted, nil)
+}
+
 // Read a Objects into add() for the given Fs.
 // dir is the start directory, "" for root
 // If includeAll is specified all files will be added,
@@ -453,10 +525,73 @@ func readFilesFn(fs Fs, includeAll bool, dir string, add func(Object) error) (er
 				list.SetError(err)
 			}
 		} else {
-			Debug(o, "Excluded from sync (and deletion)")
+			Debugf(o, "Excluded from sync (and deletion)")
 		}
 	}
-	return nil
+	return list.Error()
+}
+
+// DirEntries is a slice of Object or *Dir
+type DirEntries []BasicInfo
+
+// Len is part of sort.Interface.
+func (ds DirEntries) Len() int {
+	return len(ds)
+}
+
+// Swap is part of sort.Interface.
+func (ds DirEntries) Swap(i, j int) {
+	ds[i], ds[j] = ds[j], ds[i]
+}
+
+// Less is part of sort.Interface.
+func (ds DirEntries) Less(i, j int) bool {
+	return ds[i].Remote() < ds[j].Remote()
+}
+
+// ListDirSorted reads Object and *Dir into entries for the given Fs.
+//
+// dir is the start directory, "" for root
+//
+// If includeAll is specified all files will be added, otherwise only
+// files passing the filter will be added.
+//
+// Files will be returned in sorted order
+func ListDirSorted(fs Fs, includeAll bool, dir string) (entries DirEntries, err error) {
+	list := NewLister().SetLevel(1)
+	if !includeAll {
+		list.SetFilter(Config.Filter)
+	}
+	list.Start(fs, dir)
+	for {
+		o, dir, err := list.Get()
+		if err != nil {
+			return nil, err
+		} else if o != nil {
+			// Make sure we don't delete excluded files if not required
+			if includeAll || Config.Filter.IncludeObject(o) {
+				entries = append(entries, o)
+			} else {
+				Debugf(o, "Excluded from sync (and deletion)")
+			}
+		} else if dir != nil {
+			if includeAll || Config.Filter.IncludeDirectory(dir.Remote()) {
+				entries = append(entries, dir)
+			} else {
+				Debugf(dir, "Excluded from sync (and deletion)")
+			}
+		} else {
+			// finishd since err, o, dir == nil
+			break
+		}
+	}
+	err = list.Error()
+	if err != nil {
+		return nil, err
+	}
+	// sort the directory entries by Remote
+	sort.Sort(entries)
+	return entries, nil
 }
 
 // Read a map of Object.Remote to Object for the given Fs.
@@ -474,10 +609,10 @@ func readFilesMap(fs Fs, includeAll bool, dir string) (files map[string]Object,
 		if _, ok := files[remote]; !ok {
 			files[remote] = o
 			if _, ok := normalised[normalisedRemote]; ok {
-				Log(o, "Warning: File found with same name but different case on %v", o.Fs())
+				Logf(o, "File found with same name but different case on %v", o.Fs())
 			}
 		} else {
-			Log(o, "Duplicate file detected")
+			Logf(o, "Duplicate file detected")
 		}
 		normalised[normalisedRemote] = struct{}{}
 		return nil
@@ -496,13 +631,13 @@ func readFilesMaps(fdst Fs, fdstIncludeAll bool, fsrc Fs, fsrcIncludeAll bool, d
 
 	list := func(fs Fs, includeAll bool, pMap *map[string]Object, pErr *error) {
 		defer wg.Done()
-		Log(fs, "Building file list")
+		Infof(fs, "Building file list")
 		files, listErr := readFilesMap(fs, includeAll, dir)
 		if listErr != nil {
-			ErrorLog(fs, "Error building file list: %v", listErr)
+			Errorf(fs, "Error building file list: %v", listErr)
 			*pErr = listErr
 		} else {
-			Debug(fs, "Done building file list")
+			Debugf(fs, "Done building file list")
 			*pMap = files
 		}
 	}
@@ -521,15 +656,34 @@ func readFilesMaps(fdst Fs, fdstIncludeAll bool, fsrc Fs, fsrcIncludeAll bool, d
 	return dstFiles, srcFiles, err
 }
 
+// SameConfig returns true if fdst and fsrc are using the same config
+// file entry
+func SameConfig(fdst, fsrc Info) bool {
+	return fdst.Name() == fsrc.Name()
+}
+
 // Same returns true if fdst and fsrc point to the same underlying Fs
-func Same(fdst, fsrc Fs) bool {
-	return fdst.Name() == fsrc.Name() && fdst.Root() == fsrc.Root()
+func Same(fdst, fsrc Info) bool {
+	return SameConfig(fdst, fsrc) && fdst.Root() == fsrc.Root()
 }
 
 // Overlapping returns true if fdst and fsrc point to the same
-// underlying Fs or they overlap.
-func Overlapping(fdst, fsrc Fs) bool {
-	return fdst.Name() == fsrc.Name() && (strings.HasPrefix(fdst.Root(), fsrc.Root()) || strings.HasPrefix(fsrc.Root(), fdst.Root()))
+// underlying Fs and they overlap.
+func Overlapping(fdst, fsrc Info) bool {
+	if !SameConfig(fdst, fsrc) {
+		return false
+	}
+	// Return the Root with a trailing / if not empty
+	fixedRoot := func(f Info) string {
+		s := strings.Trim(f.Root(), "/")
+		if s != "" {
+			s += "/"
+		}
+		return s
+	}
+	fdstRoot := fixedRoot(fdst)
+	fsrcRoot := fixedRoot(fsrc)
+	return strings.HasPrefix(fdstRoot, fsrcRoot) || strings.HasPrefix(fsrcRoot, fdstRoot)
 }
 
 // checkIdentical checks to see if dst and src are identical
@@ -537,34 +691,30 @@ func Overlapping(fdst, fsrc Fs) bool {
 // it returns true if differences were found
 // it also returns whether it couldn't be hashed
 func checkIdentical(dst, src Object) (differ bool, noHash bool) {
-	Stats.Checking(src.Remote())
-	defer Stats.DoneChecking(src.Remote())
-	if src.Size() != dst.Size() {
-		Stats.Error()
-		ErrorLog(src, "Sizes differ")
+	same, hash, err := CheckHashes(src, dst)
+	if err != nil {
+		// CheckHashes will log and count errors
 		return true, false
 	}
-	if !Config.SizeOnly {
-		same, hash, err := CheckHashes(src, dst)
-		if err != nil {
-			// CheckHashes will log and count errors
-			return true, false
-		}
-		if hash == HashNone {
-			return false, true
-		}
-		if !same {
-			Stats.Error()
-			ErrorLog(src, "%v differ", hash)
-			return true, false
-		}
+	if hash == HashNone {
+		return false, true
+	}
+	if !same {
+		Stats.Error()
+		Errorf(src, "%v differ", hash)
+		return true, false
 	}
-	Debug(src, "OK")
 	return false, false
 }
 
-// Check the files in fsrc and fdst according to Size and hash
-func Check(fdst, fsrc Fs) error {
+// CheckFn checks the files in fsrc and fdst according to Size and
+// hash using checkFunction on each file to check the hashes.
+//
+// checkFunction sees if dst and src are identical
+//
+// it returns true if differences were found
+// it also returns whether it couldn't be hashed
+func CheckFn(fdst, fsrc Fs, checkFunction func(a, b Object) (differ bool, noHash bool)) error {
 	dstFiles, srcFiles, err := readFilesMaps(fdst, false, fsrc, false, "")
 	if err != nil {
 		return err
@@ -577,30 +727,30 @@ func Check(fdst, fsrc Fs) error {
 
 	// Move all the common files into commonFiles and delete then
 	// from srcFiles and dstFiles
-	commonFiles := make(map[string][]Object)
+	commonFiles := make(map[string][2]Object)
 	for remote, src := range srcFiles {
 		if dst, ok := dstFiles[remote]; ok {
-			commonFiles[remote] = []Object{dst, src}
+			commonFiles[remote] = [2]Object{dst, src}
 			delete(srcFiles, remote)
 			delete(dstFiles, remote)
 		}
 	}
 
-	Log(fdst, "%d files not in %v", len(dstFiles), fsrc)
+	Logf(fdst, "%d files not in %v", len(dstFiles), fsrc)
 	for _, dst := range dstFiles {
 		Stats.Error()
-		ErrorLog(dst, "File not in %v", fsrc)
+		Errorf(dst, "File not in %v", fsrc)
 		atomic.AddInt32(&differences, 1)
 	}
 
-	Log(fsrc, "%d files not in %s", len(srcFiles), fdst)
+	Logf(fsrc, "%d files not in %s", len(srcFiles), fdst)
 	for _, src := range srcFiles {
 		Stats.Error()
-		ErrorLog(src, "File not in %v", fdst)
+		Errorf(src, "File not in %v", fdst)
 		atomic.AddInt32(&differences, 1)
 	}
 
-	checks := make(chan []Object, Config.Transfers)
+	checks := make(chan [2]Object, Config.Transfers)
 	go func() {
 		for _, check := range commonFiles {
 			checks <- check
@@ -608,6 +758,20 @@ func Check(fdst, fsrc Fs) error {
 		close(checks)
 	}()
 
+	checkIdentical := func(dst, src Object) (differ bool, noHash bool) {
+		Stats.Checking(src.Remote())
+		defer Stats.DoneChecking(src.Remote())
+		if src.Size() != dst.Size() {
+			Stats.Error()
+			Errorf(src, "Sizes differ")
+			return true, false
+		}
+		if Config.SizeOnly {
+			return false, false
+		}
+		return checkFunction(dst, src)
+	}
+
 	var checkerWg sync.WaitGroup
 	checkerWg.Add(Config.Checkers)
 	for i := 0; i < Config.Checkers; i++ {
@@ -617,6 +781,8 @@ func Check(fdst, fsrc Fs) error {
 				differ, noHash := checkIdentical(check[0], check[1])
 				if differ {
 					atomic.AddInt32(&differences, 1)
+				} else {
+					Debugf(check[0], "OK")
 				}
 				if noHash {
 					atomic.AddInt32(&noHashes, 1)
@@ -625,11 +791,11 @@ func Check(fdst, fsrc Fs) error {
 		}()
 	}
 
-	Log(fdst, "Waiting for checks to finish")
+	Infof(fdst, "Waiting for checks to finish")
 	checkerWg.Wait()
-	Log(fdst, "%d differences found", Stats.GetErrors())
+	Logf(fdst, "%d differences found", Stats.GetErrors())
 	if noHashes > 0 {
-		Log(fdst, "%d hashes could not be checked", noHashes)
+		Logf(fdst, "%d hashes could not be checked", noHashes)
 	}
 	if differences > 0 {
 		return errors.Errorf("%d differences found", differences)
@@ -637,6 +803,93 @@ func Check(fdst, fsrc Fs) error {
 	return nil
 }
 
+// Check the files in fsrc and fdst according to Size and hash
+func Check(fdst, fsrc Fs) error {
+	return CheckFn(fdst, fsrc, checkIdentical)
+}
+
+// ReadFill reads as much data from r into buf as it can
+//
+// It reads until the buffer is full or r.Read returned an error.
+//
+// This is io.ReadFull but when you just want as much data as
+// possible, not an exact size of block.
+func ReadFill(r io.Reader, buf []byte) (n int, err error) {
+	var nn int
+	for n < len(buf) && err == nil {
+		nn, err = r.Read(buf[n:])
+		n += nn
+	}
+	return n, err
+}
+
+// CheckEqualReaders checks to see if in1 and in2 have the same
+// content when read.
+//
+// it returns true if differences were found
+func CheckEqualReaders(in1, in2 io.Reader) (differ bool, err error) {
+	const bufSize = 64 * 1024
+	buf1 := make([]byte, bufSize)
+	buf2 := make([]byte, bufSize)
+	for {
+		n1, err1 := ReadFill(in1, buf1)
+		n2, err2 := ReadFill(in2, buf2)
+		// check errors
+		if err1 != nil && err1 != io.EOF {
+			return true, err1
+		} else if err2 != nil && err2 != io.EOF {
+			return true, err2
+		}
+		// err1 && err2 are nil or io.EOF here
+		// process the data
+		if n1 != n2 || !bytes.Equal(buf1[:n1], buf2[:n2]) {
+			return true, nil
+		}
+		// if both streams finished the we have finished
+		if err1 == io.EOF && err2 == io.EOF {
+			break
+		}
+	}
+	return false, nil
+}
+
+// CheckIdentical checks to see if dst and src are identical by
+// reading all their bytes if necessary.
+//
+// it returns true if differences were found
+func CheckIdentical(dst, src Object) (differ bool, err error) {
+	in1, err := dst.Open()
+	if err != nil {
+		return true, errors.Wrapf(err, "failed to open %q", dst)
+	}
+	in1 = NewAccount(in1, dst).WithBuffer() // account and buffer the transfer
+	defer CheckClose(in1, &err)
+
+	in2, err := src.Open()
+	if err != nil {
+		return true, errors.Wrapf(err, "failed to open %q", src)
+	}
+	in2 = NewAccount(in2, src).WithBuffer() // account and buffer the transfer
+	defer CheckClose(in2, &err)
+
+	return CheckEqualReaders(in1, in2)
+}
+
+// CheckDownload checks the files in fsrc and fdst according to Size
+// and the actual contents of the files.
+func CheckDownload(fdst, fsrc Fs) error {
+	check := func(a, b Object) (differ bool, noHash bool) {
+		differ, err := CheckIdentical(a, b)
+		if err != nil {
+			Stats.Error()
+			Errorf(a, "Failed to download: %v", err)
+			return true, true
+		}
+		return differ, false
+	}
+	return CheckFn(fdst, fsrc, check)
+}
+
 // ListFn lists the Fs to the supplied function
 //
 // Lists in parallel which may get them out of order
@@ -650,7 +903,10 @@ func ListFn(f Fs, fn func(Object)) error {
 			for {
 				o, err := list.GetObject()
 				if err != nil {
-					log.Fatal(err)
+					// The error will be persisted within the Lister object and
+					// we'll get an opportunity to return it as we leave this
+					// function.
+					return
 				}
 				// check if we are finished
 				if o == nil {
@@ -663,7 +919,7 @@ func ListFn(f Fs, fn func(Object)) error {
 		}()
 	}
 	wg.Wait()
-	return nil
+	return list.Error()
 }
 
 // mutex for synchronized output
@@ -730,7 +986,7 @@ func hashLister(ht HashType, f Fs, w io.Writer) error {
 		if err == ErrHashUnsupported {
 			sum = "UNSUPPORTED"
 		} else if err != nil {
-			Debug(o, "Failed to read %v: %v", ht, err)
+			Debugf(o, "Failed to read %v: %v", ht, err)
 			sum = "ERROR"
 		}
 		syncFprintf(w, "%*s  %s\n", HashWidth[ht], sum, o.Remote())
@@ -771,7 +1027,7 @@ func ListDir(f Fs, w io.Writer) error {
 // Mkdir makes a destination directory or container
 func Mkdir(f Fs, dir string) error {
 	if Config.DryRun {
-		Log(f, "Not making directory as dry run is set")
+		Logf(f, "Not making directory as dry run is set")
 		return nil
 	}
 	err := f.Mkdir(dir)
@@ -787,9 +1043,9 @@ func Mkdir(f Fs, dir string) error {
 func TryRmdir(f Fs, dir string) error {
 	if Config.DryRun {
 		if dir != "" {
-			Log(dir, "Not deleting as dry run is set")
+			Logf(dir, "Not deleting as dry run is set")
 		} else {
-			Log(f, "Not deleting as dry run is set")
+			Logf(f, "Not deleting as dry run is set")
 		}
 		return nil
 	}
@@ -807,17 +1063,15 @@ func Rmdir(f Fs, dir string) error {
 }
 
 // Purge removes a container and all of its contents
-//
-// FIXME doesn't delete local directories
 func Purge(f Fs) error {
 	doFallbackPurge := true
 	var err error
-	if purger, ok := f.(Purger); ok {
+	if doPurge := f.Features().Purge; doPurge != nil {
 		doFallbackPurge = false
 		if Config.DryRun {
-			Log(f, "Not purging as --dry-run set")
+			Logf(f, "Not purging as --dry-run set")
 		} else {
-			err = purger.Purge()
+			err = doPurge()
 			if err == ErrorCantPurge {
 				doFallbackPurge = true
 			}
@@ -830,7 +1084,7 @@ func Purge(f Fs) error {
 		if err != nil {
 			return err
 		}
-		err = Rmdir(f, "")
+		err = Rmdirs(f, "")
 	}
 	if err != nil {
 		Stats.Error()
@@ -861,8 +1115,8 @@ func Delete(f Fs) error {
 // dedupeRename renames the objs slice to different names
 func dedupeRename(remote string, objs []Object) {
 	f := objs[0].Fs()
-	mover, ok := f.(Mover)
-	if !ok {
+	doMove := f.Features().Move
+	if doMove == nil {
 		log.Fatalf("Fs %v doesn't support Move", f)
 	}
 	ext := path.Ext(remote)
@@ -870,15 +1124,15 @@ func dedupeRename(remote string, objs []Object) {
 	for i, o := range objs {
 		newName := fmt.Sprintf("%s-%d%s", base, i+1, ext)
 		if !Config.DryRun {
-			newObj, err := mover.Move(o, newName)
+			newObj, err := doMove(o, newName)
 			if err != nil {
 				Stats.Error()
-				ErrorLog(o, "Failed to rename: %v", err)
+				Errorf(o, "Failed to rename: %v", err)
 				continue
 			}
-			Log(newObj, "renamed from: %v", o)
+			Infof(newObj, "renamed from: %v", o)
 		} else {
-			Log(remote, "Not renaming to %q as --dry-run", newName)
+			Logf(remote, "Not renaming to %q as --dry-run", newName)
 		}
 	}
 }
@@ -891,7 +1145,7 @@ func dedupeDeleteAllButOne(keep int, remote string, objs []Object) {
 		}
 		_ = DeleteFile(o)
 	}
-	Log(remote, "Deleted %d extra copies", len(objs)-1)
+	Logf(remote, "Deleted %d extra copies", len(objs)-1)
 }
 
 // dedupeDeleteIdentical deletes all but one of identical (by hash) copies
@@ -909,7 +1163,7 @@ func dedupeDeleteIdentical(remote string, objs []Object) []Object {
 	objs = nil
 	for md5sum, hashObjs := range byHash {
 		if len(hashObjs) > 1 {
-			Log(remote, "Deleting %d/%d identical duplicates (md5sum %q)", len(hashObjs)-1, len(hashObjs), md5sum)
+			Logf(remote, "Deleting %d/%d identical duplicates (md5sum %q)", len(hashObjs)-1, len(hashObjs), md5sum)
 			for _, o := range hashObjs[1:] {
 				_ = DeleteFile(o)
 			}
@@ -1012,7 +1266,7 @@ var _ pflag.Value = (*DeduplicateMode)(nil)
 // delete all but one or rename them to be different. Only useful with
 // Google Drive which can have duplicate file names.
 func Deduplicate(f Fs, mode DeduplicateMode) error {
-	Log(f, "Looking for duplicates using %v mode.", mode)
+	Infof(f, "Looking for duplicates using %v mode.", mode)
 	files := map[string][]Object{}
 	list := NewLister().Start(f, "")
 	for {
@@ -1029,10 +1283,10 @@ func Deduplicate(f Fs, mode DeduplicateMode) error {
 	}
 	for remote, objs := range files {
 		if len(objs) > 1 {
-			Log(remote, "Found %d duplicates - deleting identical copies", len(objs))
+			Logf(remote, "Found %d duplicates - deleting identical copies", len(objs))
 			objs = dedupeDeleteIdentical(remote, objs)
 			if len(objs) <= 1 {
-				Log(remote, "All duplicates removed")
+				Logf(remote, "All duplicates removed")
 				continue
 			}
 			switch mode {
@@ -1073,7 +1327,7 @@ func listToChan(list *Lister) ObjectsChan {
 			if err != nil {
 				if err != ErrorDirNotFound {
 					Stats.Error()
-					ErrorLog(nil, "Failed to list: %v", err)
+					Errorf(nil, "Failed to list: %v", err)
 				}
 				return
 			}
@@ -1091,19 +1345,37 @@ func listToChan(list *Lister) ObjectsChan {
 
 // CleanUp removes the trash for the Fs
 func CleanUp(f Fs) error {
-	fc, ok := f.(CleanUpper)
-	if !ok {
+	doCleanUp := f.Features().CleanUp
+	if doCleanUp == nil {
 		return errors.Errorf("%v doesn't support cleanup", f)
 	}
 	if Config.DryRun {
-		Log(f, "Not running cleanup as --dry-run set")
+		Logf(f, "Not running cleanup as --dry-run set")
 		return nil
 	}
-	return fc.CleanUp()
+	return doCleanUp()
+}
+
+// wrap a Reader and a Closer together into a ReadCloser
+type readCloser struct {
+	io.Reader
+	Closer io.Closer
+}
+
+// Close the Closer
+func (r *readCloser) Close() error {
+	return r.Closer.Close()
 }
 
 // Cat any files to the io.Writer
-func Cat(f Fs, w io.Writer) error {
+//
+// if offset == 0 it will be ignored
+// if offset > 0 then the file will be seeked to that offset
+// if offset < 0 then the file will be seeked that far from the end
+//
+// if count < 0 then it will be ignored
+// if count >= 0 then only that many characters will be output
+func Cat(f Fs, w io.Writer, offset, count int64) error {
 	var mu sync.Mutex
 	return ListFn(f, func(o Object) {
 		var err error
@@ -1111,41 +1383,60 @@ func Cat(f Fs, w io.Writer) error {
 		defer func() {
 			Stats.DoneTransferring(o.Remote(), err == nil)
 		}()
-		mu.Lock()
-		defer mu.Unlock()
-		in, err := o.Open()
+		size := o.Size()
+		thisOffset := offset
+		if thisOffset < 0 {
+			thisOffset += size
+		}
+		// size remaining is now reduced by thisOffset
+		size -= thisOffset
+		var options []OpenOption
+		if thisOffset > 0 {
+			options = append(options, &SeekOption{Offset: thisOffset})
+		}
+		in, err := o.Open(options...)
 		if err != nil {
 			Stats.Error()
-			ErrorLog(o, "Failed to open: %v", err)
+			Errorf(o, "Failed to open: %v", err)
 			return
 		}
+		if count >= 0 {
+			in = &readCloser{Reader: &io.LimitedReader{R: in, N: count}, Closer: in}
+			// reduce remaining size to count
+			if size > count {
+				size = count
+			}
+		}
+		in = NewAccountSizeName(in, size, o.Remote()).WithBuffer() // account and buffer the transfer
 		defer func() {
 			err = in.Close()
 			if err != nil {
 				Stats.Error()
-				ErrorLog(o, "Failed to close: %v", err)
+				Errorf(o, "Failed to close: %v", err)
 			}
 		}()
-		inAccounted := NewAccountWithBuffer(in, o) // account and buffer the transfer
-		_, err = io.Copy(w, inAccounted)
+		// take the lock just before we output stuff, so at the last possible moment
+		mu.Lock()
+		defer mu.Unlock()
+		_, err = io.Copy(w, in)
 		if err != nil {
 			Stats.Error()
-			ErrorLog(o, "Failed to send to output: %v", err)
+			Errorf(o, "Failed to send to output: %v", err)
 		}
 	})
 }
 
 // Rmdirs removes any empty directories (or directories only
 // containing empty directories) under f, including f.
-func Rmdirs(f Fs) error {
-	list := NewLister().Start(f, "")
+func Rmdirs(f Fs, dir string) error {
+	list := NewLister().Start(f, dir)
 	dirEmpty := make(map[string]bool)
 	dirEmpty[""] = true
 	for {
 		o, dir, err := list.Get()
 		if err != nil {
 			Stats.Error()
-			ErrorLog(f, "Failed to list: %v", err)
+			Errorf(f, "Failed to list: %v", err)
 			return err
 		} else if dir != nil {
 			// add a new directory as empty
@@ -1187,7 +1478,7 @@ func Rmdirs(f Fs) error {
 		err := TryRmdir(f, dir)
 		if err != nil {
 			Stats.Error()
-			ErrorLog(dir, "Failed to rmdir: %v", err)
+			Errorf(dir, "Failed to rmdir: %v", err)
 			return err
 		}
 	}
