@@ -21,13 +21,17 @@ package fs_test
 
 import (
 	"bytes"
+	"errors"
 	"flag"
+	"fmt"
+	"io"
 	"io/ioutil"
 	"log"
 	"os"
 	"path"
 	"path/filepath"
 	"regexp"
+	"sort"
 	"strings"
 	"testing"
 	"time"
@@ -103,8 +107,9 @@ func newRun() *Run {
 	// "RCLONE_CONFIG_PASS=hunter2" (or your password)
 	*fs.AskPassword = false
 	fs.LoadConfig()
-	fs.Config.Verbose = *Verbose
-	fs.Config.Quiet = !*Verbose
+	if *Verbose {
+		fs.Config.LogLevel = fs.LogLevelDebug
+	}
 	fs.Config.DumpHeaders = *DumpHeaders
 	fs.Config.DumpBodies = *DumpBodies
 	fs.Config.LowLevelRetries = *LowLevelRetries
@@ -127,6 +132,13 @@ func newRun() *Run {
 	return r
 }
 
+// dirsToRemove sorts by string length
+type dirsToRemove []string
+
+func (d dirsToRemove) Len() int           { return len(d) }
+func (d dirsToRemove) Swap(i, j int)      { d[i], d[j] = d[j], d[i] }
+func (d dirsToRemove) Less(i, j int) bool { return len(d[i]) > len(d[j]) }
+
 // NewRun initialise the remote and local for testing and returns a
 // run object.  Call this from the tests.
 //
@@ -143,19 +155,31 @@ func NewRun(t *testing.T) *Run {
 		r = new(Run)
 		*r = *oneRun
 		r.cleanRemote = func() {
+			var toDelete dirsToRemove
 			list := fs.NewLister().Start(r.fremote, "")
 			for {
-				o, err := list.GetObject()
+				o, dir, err := list.Get()
 				if err != nil {
+					if err == fs.ErrorDirNotFound {
+						break
+					}
 					t.Fatalf("Error listing: %v", err)
-				}
-				// Check if we are finished
-				if o == nil {
+				} else if o != nil {
+					err = o.Remove()
+					if err != nil {
+						t.Errorf("Error removing file %q: %v", o.Remote(), err)
+					}
+				} else if dir != nil {
+					toDelete = append(toDelete, dir.Remote())
+				} else {
 					break
 				}
-				err = o.Remove()
+			}
+			sort.Sort(toDelete)
+			for _, dir := range toDelete {
+				err := r.fremote.Rmdir(dir)
 				if err != nil {
-					t.Errorf("Error removing file: %v", err)
+					t.Errorf("Error removing dir %q: %v", dir, err)
 				}
 			}
 			// Check remote is empty
@@ -168,6 +192,19 @@ func NewRun(t *testing.T) *Run {
 	return r
 }
 
+// Rename a file in local
+func (r *Run) RenameFile(item fstest.Item, newpath string) fstest.Item {
+	oldFilepath := path.Join(r.localName, item.Path)
+	newFilepath := path.Join(r.localName, newpath)
+	if err := os.Rename(oldFilepath, newFilepath); err != nil {
+		r.Fatalf("Failed to rename file from %q to %q: %v", item.Path, newpath, err)
+	}
+
+	item.Path = newpath
+
+	return item
+}
+
 // Write a file to local
 func (r *Run) WriteFile(filePath, content string, t time.Time) fstest.Item {
 	item := fstest.NewItem(filePath, content, t)
@@ -209,9 +246,8 @@ func (r *Run) Mkdir(f fs.Fs) {
 func (r *Run) WriteObjectTo(f fs.Fs, remote, content string, modTime time.Time, useUnchecked bool) fstest.Item {
 	put := f.Put
 	if useUnchecked {
-		if fPutUnchecked, ok := f.(fs.PutUncheckeder); ok {
-			put = fPutUnchecked.PutUnchecked
-		} else {
+		put = f.Features().PutUnchecked
+		if put == nil {
 			r.Fatalf("Fs doesn't support PutUnchecked")
 		}
 	}
@@ -435,14 +471,14 @@ func TestDelete(t *testing.T) {
 	fstest.CheckItems(t, r.fremote, file3)
 }
 
-func TestCheck(t *testing.T) {
+func testCheck(t *testing.T, checkFunction func(fdst, fsrc fs.Fs) error) {
 	r := NewRun(t)
 	defer r.Finalise()
 
 	check := func(i int, wantErrors int64) {
-		fs.Debug(r.fremote, "%d: Starting check test", i)
+		fs.Debugf(r.fremote, "%d: Starting check test", i)
 		oldErrors := fs.Stats.GetErrors()
-		err := fs.Check(r.flocal, r.fremote)
+		err := checkFunction(r.flocal, r.fremote)
 		gotErrors := fs.Stats.GetErrors() - oldErrors
 		if wantErrors == 0 && err != nil {
 			t.Errorf("%d: Got error when not expecting one: %v", i, err)
@@ -453,7 +489,7 @@ func TestCheck(t *testing.T) {
 		if wantErrors != gotErrors {
 			t.Errorf("%d: Expecting %d errors but got %d", i, wantErrors, gotErrors)
 		}
-		fs.Debug(r.fremote, "%d: Ending check test", i)
+		fs.Debugf(r.fremote, "%d: Ending check test", i)
 	}
 
 	file1 := r.WriteBoth("rutabaga", "is tasty", t3)
@@ -483,6 +519,14 @@ func TestCheck(t *testing.T) {
 	check(5, 0)
 }
 
+func TestCheck(t *testing.T) {
+	testCheck(t, fs.Check)
+}
+
+func TestCheckDownload(t *testing.T) {
+	testCheck(t, fs.CheckDownload)
+}
+
 func TestCheckSizeOnly(t *testing.T) {
 	fs.Config.SizeOnly = true
 	defer func() { fs.Config.SizeOnly = false }()
@@ -500,12 +544,22 @@ func (r *Run) checkWithDuplicates(t *testing.T, items ...fstest.Item) {
 	assert.Equal(t, wantSize, size)
 }
 
-func TestDeduplicateInteractive(t *testing.T) {
-	if *RemoteName != "TestDrive:" {
-		t.Skip("Can only test deduplicate on google drive")
+func skipIfCantDedupe(t *testing.T, f fs.Fs) {
+	if f.Features().PutUnchecked == nil {
+		t.Skip("Can't test deduplicate - no PutUnchecked")
 	}
+	if !f.Features().DuplicateFiles {
+		t.Skip("Can't test deduplicate - no duplicate files possible")
+	}
+	if !f.Hashes().Contains(fs.HashMD5) {
+		t.Skip("Can't test deduplicate - MD5 not supported")
+	}
+}
+
+func TestDeduplicateInteractive(t *testing.T) {
 	r := NewRun(t)
 	defer r.Finalise()
+	skipIfCantDedupe(t, r.fremote)
 
 	file1 := r.WriteUncheckedObject("one", "This is one", t1)
 	file2 := r.WriteUncheckedObject("one", "This is one", t1)
@@ -519,11 +573,9 @@ func TestDeduplicateInteractive(t *testing.T) {
 }
 
 func TestDeduplicateSkip(t *testing.T) {
-	if *RemoteName != "TestDrive:" {
-		t.Skip("Can only test deduplicate on google drive")
-	}
 	r := NewRun(t)
 	defer r.Finalise()
+	skipIfCantDedupe(t, r.fremote)
 
 	file1 := r.WriteUncheckedObject("one", "This is one", t1)
 	file2 := r.WriteUncheckedObject("one", "This is one", t1)
@@ -537,11 +589,9 @@ func TestDeduplicateSkip(t *testing.T) {
 }
 
 func TestDeduplicateFirst(t *testing.T) {
-	if *RemoteName != "TestDrive:" {
-		t.Skip("Can only test deduplicate on google drive")
-	}
 	r := NewRun(t)
 	defer r.Finalise()
+	skipIfCantDedupe(t, r.fremote)
 
 	file1 := r.WriteUncheckedObject("one", "This is one", t1)
 	file2 := r.WriteUncheckedObject("one", "This is one A", t1)
@@ -560,11 +610,9 @@ func TestDeduplicateFirst(t *testing.T) {
 }
 
 func TestDeduplicateNewest(t *testing.T) {
-	if *RemoteName != "TestDrive:" {
-		t.Skip("Can only test deduplicate on google drive")
-	}
 	r := NewRun(t)
 	defer r.Finalise()
+	skipIfCantDedupe(t, r.fremote)
 
 	file1 := r.WriteUncheckedObject("one", "This is one", t1)
 	file2 := r.WriteUncheckedObject("one", "This is one too", t2)
@@ -578,11 +626,9 @@ func TestDeduplicateNewest(t *testing.T) {
 }
 
 func TestDeduplicateOldest(t *testing.T) {
-	if *RemoteName != "TestDrive:" {
-		t.Skip("Can only test deduplicate on google drive")
-	}
 	r := NewRun(t)
 	defer r.Finalise()
+	skipIfCantDedupe(t, r.fremote)
 
 	file1 := r.WriteUncheckedObject("one", "This is one", t1)
 	file2 := r.WriteUncheckedObject("one", "This is one too", t2)
@@ -596,11 +642,9 @@ func TestDeduplicateOldest(t *testing.T) {
 }
 
 func TestDeduplicateRename(t *testing.T) {
-	if *RemoteName != "TestDrive:" {
-		t.Skip("Can only test deduplicate on google drive")
-	}
 	r := NewRun(t)
 	defer r.Finalise()
+	skipIfCantDedupe(t, r.fremote)
 
 	file1 := r.WriteUncheckedObject("one.txt", "This is one", t1)
 	file2 := r.WriteUncheckedObject("one.txt", "This is one too", t2)
@@ -634,18 +678,30 @@ func TestDeduplicateRename(t *testing.T) {
 func TestCat(t *testing.T) {
 	r := NewRun(t)
 	defer r.Finalise()
-	file1 := r.WriteBoth("file1", "aaa", t1)
-	file2 := r.WriteBoth("file2", "bbb", t2)
+	file1 := r.WriteBoth("file1", "ABCDEFGHIJ", t1)
+	file2 := r.WriteBoth("file2", "012345678", t2)
 
 	fstest.CheckItems(t, r.fremote, file1, file2)
 
-	var buf bytes.Buffer
-	err := fs.Cat(r.fremote, &buf)
-	require.NoError(t, err)
-	res := buf.String()
+	for _, test := range []struct {
+		offset int64
+		count  int64
+		a      string
+		b      string
+	}{
+		{0, -1, "ABCDEFGHIJ", "012345678"},
+		{0, 5, "ABCDE", "01234"},
+		{-3, -1, "HIJ", "678"},
+		{2, 3, "CDE", "234"},
+	} {
+		var buf bytes.Buffer
+		err := fs.Cat(r.fremote, &buf, test.offset, test.count)
+		require.NoError(t, err)
+		res := buf.String()
 
-	if res != "aaabbb" && res != "bbbaaa" {
-		t.Errorf("Incorrect output from Cat: %q", res)
+		if res != test.a+test.b && res != test.b+test.a {
+			t.Errorf("Incorrect output from Cat(%d,%d): %q", test.offset, test.count, res)
+		}
 	}
 }
 
@@ -656,7 +712,7 @@ func TestRmdirs(t *testing.T) {
 
 	// Clean any directories that have crept in so far
 	// FIXME make the Finalise method do this?
-	require.NoError(t, fs.Rmdirs(r.fremote))
+	require.NoError(t, fs.Rmdirs(r.fremote, ""))
 
 	// Make some files and dirs we expect to keep
 	r.ForceMkdir(r.fremote)
@@ -692,7 +748,7 @@ func TestRmdirs(t *testing.T) {
 		fs.Config.ModifyWindow,
 	)
 
-	require.NoError(t, fs.Rmdirs(r.fremote))
+	require.NoError(t, fs.Rmdirs(r.fremote, ""))
 
 	fstest.CheckListingWithPrecision(
 		t,
@@ -714,7 +770,6 @@ func TestMoveFile(t *testing.T) {
 	r := NewRun(t)
 	defer r.Finalise()
 
-	r.Mkdir(r.fremote)
 	file1 := r.WriteFile("file1", "file1 contents", t1)
 	fstest.CheckItems(t, r.flocal, file1)
 
@@ -745,7 +800,6 @@ func TestCopyFile(t *testing.T) {
 	file2 := file1
 	file2.Path = "sub/file2"
 
-	r.Mkdir(r.fremote)
 	err := fs.CopyFile(r.fremote, r.flocal, file2.Path, file1.Path)
 	require.NoError(t, err)
 	fstest.CheckItems(t, r.flocal, file1)
@@ -756,3 +810,262 @@ func TestCopyFile(t *testing.T) {
 	fstest.CheckItems(t, r.flocal, file1)
 	fstest.CheckItems(t, r.fremote, file2)
 }
+
+// testFsInfo is for unit testing fs.Info
+type testFsInfo struct {
+	name      string
+	root      string
+	stringVal string
+	precision time.Duration
+	hashes    fs.HashSet
+	features  fs.Features
+}
+
+// Name of the remote (as passed into NewFs)
+func (i *testFsInfo) Name() string { return i.name }
+
+// Root of the remote (as passed into NewFs)
+func (i *testFsInfo) Root() string { return i.root }
+
+// String returns a description of the FS
+func (i *testFsInfo) String() string { return i.stringVal }
+
+// Precision of the ModTimes in this Fs
+func (i *testFsInfo) Precision() time.Duration { return i.precision }
+
+// Returns the supported hash types of the filesystem
+func (i *testFsInfo) Hashes() fs.HashSet { return i.hashes }
+
+// Returns the supported hash types of the filesystem
+func (i *testFsInfo) Features() *fs.Features { return &i.features }
+
+func TestSameConfig(t *testing.T) {
+	a := &testFsInfo{name: "name", root: "root"}
+	for _, test := range []struct {
+		name     string
+		root     string
+		expected bool
+	}{
+		{"name", "root", true},
+		{"name", "rooty", true},
+		{"namey", "root", false},
+		{"namey", "roott", false},
+	} {
+		b := &testFsInfo{name: test.name, root: test.root}
+		actual := fs.SameConfig(a, b)
+		assert.Equal(t, test.expected, actual)
+		actual = fs.SameConfig(b, a)
+		assert.Equal(t, test.expected, actual)
+	}
+}
+
+func TestSame(t *testing.T) {
+	a := &testFsInfo{name: "name", root: "root"}
+	for _, test := range []struct {
+		name     string
+		root     string
+		expected bool
+	}{
+		{"name", "root", true},
+		{"name", "rooty", false},
+		{"namey", "root", false},
+		{"namey", "roott", false},
+	} {
+		b := &testFsInfo{name: test.name, root: test.root}
+		actual := fs.Same(a, b)
+		assert.Equal(t, test.expected, actual)
+		actual = fs.Same(b, a)
+		assert.Equal(t, test.expected, actual)
+	}
+}
+
+func TestOverlapping(t *testing.T) {
+	a := &testFsInfo{name: "name", root: "root"}
+	for _, test := range []struct {
+		name     string
+		root     string
+		expected bool
+	}{
+		{"name", "root", true},
+		{"namey", "root", false},
+		{"name", "rooty", false},
+		{"namey", "rooty", false},
+		{"name", "roo", false},
+		{"name", "root/toot", true},
+		{"name", "root/toot/", true},
+		{"name", "", true},
+		{"name", "/", true},
+	} {
+		b := &testFsInfo{name: test.name, root: test.root}
+		what := fmt.Sprintf("(%q,%q) vs (%q,%q)", a.name, a.root, b.name, b.root)
+		actual := fs.Overlapping(a, b)
+		assert.Equal(t, test.expected, actual, what)
+		actual = fs.Overlapping(b, a)
+		assert.Equal(t, test.expected, actual, what)
+	}
+}
+
+func TestListDirSorted(t *testing.T) {
+	r := NewRun(t)
+	defer r.Finalise()
+
+	fs.Config.Filter.MaxSize = 10
+	defer func() {
+		fs.Config.Filter.MaxSize = -1
+	}()
+
+	files := []fstest.Item{
+		r.WriteObject("a.txt", "hello world", t1),
+		r.WriteObject("zend.txt", "hello", t1),
+		r.WriteObject("sub dir/hello world", "hello world", t1),
+		r.WriteObject("sub dir/hello world2", "hello world", t1),
+		r.WriteObject("sub dir/sub sub dir/hello world3", "hello world", t1),
+	}
+	fstest.CheckItems(t, r.fremote, files...)
+	var items fs.DirEntries
+	var err error
+
+	// Turn the BasicInfo into a name, ending with a / if it is a
+	// dir
+	str := func(i int) string {
+		item := items[i]
+		name := item.Remote()
+		switch item.(type) {
+		case fs.Object:
+		case *fs.Dir:
+			name += "/"
+		default:
+			t.Fatalf("Unknown type %+v", item)
+		}
+		return name
+	}
+
+	items, err = fs.ListDirSorted(r.fremote, true, "")
+	require.NoError(t, err)
+	require.Len(t, items, 3)
+	assert.Equal(t, "a.txt", str(0))
+	assert.Equal(t, "sub dir/", str(1))
+	assert.Equal(t, "zend.txt", str(2))
+
+	items, err = fs.ListDirSorted(r.fremote, false, "")
+	require.NoError(t, err)
+	require.Len(t, items, 2)
+	assert.Equal(t, "sub dir/", str(0))
+	assert.Equal(t, "zend.txt", str(1))
+
+	items, err = fs.ListDirSorted(r.fremote, true, "sub dir")
+	require.NoError(t, err)
+	require.Len(t, items, 3)
+	assert.Equal(t, "sub dir/hello world", str(0))
+	assert.Equal(t, "sub dir/hello world2", str(1))
+	assert.Equal(t, "sub dir/sub sub dir/", str(2))
+
+	items, err = fs.ListDirSorted(r.fremote, false, "sub dir")
+	require.NoError(t, err)
+	require.Len(t, items, 1)
+	assert.Equal(t, "sub dir/sub sub dir/", str(0))
+}
+
+type byteReader struct {
+	c byte
+}
+
+func (br *byteReader) Read(p []byte) (n int, err error) {
+	if br.c == 0 {
+		err = io.EOF
+	} else if len(p) >= 1 {
+		p[0] = br.c
+		n = 1
+		br.c--
+	}
+	return
+}
+
+func TestReadFill(t *testing.T) {
+	buf := []byte{9, 9, 9, 9, 9}
+
+	n, err := fs.ReadFill(&byteReader{0}, buf)
+	assert.Equal(t, io.EOF, err)
+	assert.Equal(t, 0, n)
+	assert.Equal(t, []byte{9, 9, 9, 9, 9}, buf)
+
+	n, err = fs.ReadFill(&byteReader{3}, buf)
+	assert.Equal(t, io.EOF, err)
+	assert.Equal(t, 3, n)
+	assert.Equal(t, []byte{3, 2, 1, 9, 9}, buf)
+
+	n, err = fs.ReadFill(&byteReader{8}, buf)
+	assert.Equal(t, nil, err)
+	assert.Equal(t, 5, n)
+	assert.Equal(t, []byte{8, 7, 6, 5, 4}, buf)
+}
+
+type errorReader struct {
+	err error
+}
+
+func (er errorReader) Read(p []byte) (n int, err error) {
+	return 0, er.err
+}
+
+func TestCheckEqualReaders(t *testing.T) {
+	b65a := make([]byte, 65*1024)
+	b65b := make([]byte, 65*1024)
+	b65b[len(b65b)-1] = 1
+	b66 := make([]byte, 66*1024)
+
+	differ, err := fs.CheckEqualReaders(bytes.NewBuffer(b65a), bytes.NewBuffer(b65a))
+	assert.NoError(t, err)
+	assert.Equal(t, differ, false)
+
+	differ, err = fs.CheckEqualReaders(bytes.NewBuffer(b65a), bytes.NewBuffer(b65b))
+	assert.NoError(t, err)
+	assert.Equal(t, differ, true)
+
+	differ, err = fs.CheckEqualReaders(bytes.NewBuffer(b65a), bytes.NewBuffer(b66))
+	assert.NoError(t, err)
+	assert.Equal(t, differ, true)
+
+	differ, err = fs.CheckEqualReaders(bytes.NewBuffer(b66), bytes.NewBuffer(b65a))
+	assert.NoError(t, err)
+	assert.Equal(t, differ, true)
+
+	myErr := errors.New("sentinel")
+	wrap := func(b []byte) io.Reader {
+		r := bytes.NewBuffer(b)
+		e := errorReader{myErr}
+		return io.MultiReader(r, e)
+	}
+
+	differ, err = fs.CheckEqualReaders(wrap(b65a), bytes.NewBuffer(b65a))
+	assert.Equal(t, myErr, err)
+	assert.Equal(t, differ, true)
+
+	differ, err = fs.CheckEqualReaders(wrap(b65a), bytes.NewBuffer(b65b))
+	assert.Equal(t, myErr, err)
+	assert.Equal(t, differ, true)
+
+	differ, err = fs.CheckEqualReaders(wrap(b65a), bytes.NewBuffer(b66))
+	assert.Equal(t, myErr, err)
+	assert.Equal(t, differ, true)
+
+	differ, err = fs.CheckEqualReaders(wrap(b66), bytes.NewBuffer(b65a))
+	assert.Equal(t, myErr, err)
+	assert.Equal(t, differ, true)
+
+	differ, err = fs.CheckEqualReaders(bytes.NewBuffer(b65a), wrap(b65a))
+	assert.Equal(t, myErr, err)
+	assert.Equal(t, differ, true)
+
+	differ, err = fs.CheckEqualReaders(bytes.NewBuffer(b65a), wrap(b65b))
+	assert.Equal(t, myErr, err)
+	assert.Equal(t, differ, true)
+
+	differ, err = fs.CheckEqualReaders(bytes.NewBuffer(b65a), wrap(b66))
+	assert.Equal(t, myErr, err)
+	assert.Equal(t, differ, true)
+
+	differ, err = fs.CheckEqualReaders(bytes.NewBuffer(b66), wrap(b65a))
+	assert.Equal(t, myErr, err)
+	assert.Equal(t, differ, true)
+}
