@@ -308,10 +308,22 @@ func TestSyncAfterChangingModtimeOnly(t *testing.T) {
 	fstest.CheckItems(t, r.flocal, file1)
 	fstest.CheckItems(t, r.fremote, file2)
 
+	fs.Config.DryRun = true
+	defer func() { fs.Config.DryRun = false }()
+
 	fs.Stats.ResetCounters()
 	err := fs.Sync(r.fremote, r.flocal)
 	require.NoError(t, err)
 
+	fstest.CheckItems(t, r.flocal, file1)
+	fstest.CheckItems(t, r.fremote, file2)
+
+	fs.Config.DryRun = false
+
+	fs.Stats.ResetCounters()
+	err = fs.Sync(r.fremote, r.flocal)
+	require.NoError(t, err)
+
 	fstest.CheckItems(t, r.flocal, file1)
 	fstest.CheckItems(t, r.fremote, file1)
 }
@@ -321,7 +333,7 @@ func TestSyncAfterChangingModtimeOnlyWithNoUpdateModTime(t *testing.T) {
 	defer r.Finalise()
 
 	if r.fremote.Hashes().Count() == 0 {
-		t.Log("Can't check this if no hashes supported")
+		t.Logf("Can't check this if no hashes supported")
 		return
 	}
 
@@ -457,6 +469,23 @@ func TestSyncAfterRemovingAFileAndAddingAFile(t *testing.T) {
 	fstest.CheckItems(t, r.fremote, file1, file3)
 }
 
+// Sync after removing a file and adding a file
+func TestSyncAfterRemovingAFileAndAddingAFileSubDir(t *testing.T) {
+	r := NewRun(t)
+	defer r.Finalise()
+	file1 := r.WriteFile("a/potato2", "------------------------------------------------------------", t1)
+	file2 := r.WriteObject("b/potato", "SMALLER BUT SAME DATE", t2)
+	file3 := r.WriteBoth("c/non empty space", "AhHa!", t2)
+	fstest.CheckItems(t, r.fremote, file2, file3)
+	fstest.CheckItems(t, r.flocal, file1, file3)
+
+	fs.Stats.ResetCounters()
+	err := fs.Sync(r.fremote, r.flocal)
+	require.NoError(t, err)
+	fstest.CheckItems(t, r.flocal, file1, file3)
+	fstest.CheckItems(t, r.fremote, file1, file3)
+}
+
 // Sync after removing a file and adding a file with IO Errors
 func TestSyncAfterRemovingAFileAndAddingAFileWithErrors(t *testing.T) {
 	r := NewRun(t)
@@ -475,41 +504,54 @@ func TestSyncAfterRemovingAFileAndAddingAFileWithErrors(t *testing.T) {
 	fstest.CheckItems(t, r.fremote, file1, file2, file3)
 }
 
-// Sync test delete during
-func TestSyncDeleteDuring(t *testing.T) {
+// Sync test delete after
+func TestSyncDeleteAfter(t *testing.T) {
 	// This is the default so we've checked this already
 	// check it is the default
-	if !(!fs.Config.DeleteBefore && fs.Config.DeleteDuring && !fs.Config.DeleteAfter) {
-		t.Fatalf("Didn't default to --delete-during")
-	}
+	require.Equal(t, fs.Config.DeleteMode, fs.DeleteModeAfter, "Didn't default to --delete-after")
+}
+
+// Sync test delete during
+func TestSyncDeleteDuring(t *testing.T) {
+	fs.Config.DeleteMode = fs.DeleteModeDuring
+	defer func() {
+		fs.Config.DeleteMode = fs.DeleteModeDefault
+	}()
+
+	TestSyncAfterRemovingAFileAndAddingAFile(t)
 }
 
 // Sync test delete before
 func TestSyncDeleteBefore(t *testing.T) {
-	fs.Config.DeleteBefore = true
-	fs.Config.DeleteDuring = false
-	fs.Config.DeleteAfter = false
+	fs.Config.DeleteMode = fs.DeleteModeBefore
 	defer func() {
-		fs.Config.DeleteBefore = false
-		fs.Config.DeleteDuring = true
-		fs.Config.DeleteAfter = false
+		fs.Config.DeleteMode = fs.DeleteModeDefault
 	}()
 
 	TestSyncAfterRemovingAFileAndAddingAFile(t)
 }
 
-// Sync test delete after
-func TestSyncDeleteAfter(t *testing.T) {
-	fs.Config.DeleteBefore = false
-	fs.Config.DeleteDuring = false
-	fs.Config.DeleteAfter = true
+// Copy test delete before - shouldn't delete anything
+func TestCopyDeleteBefore(t *testing.T) {
+	r := NewRun(t)
+	defer r.Finalise()
+
+	fs.Config.DeleteMode = fs.DeleteModeBefore
 	defer func() {
-		fs.Config.DeleteBefore = false
-		fs.Config.DeleteDuring = true
-		fs.Config.DeleteAfter = false
+		fs.Config.DeleteMode = fs.DeleteModeDefault
 	}()
 
-	TestSyncAfterRemovingAFileAndAddingAFile(t)
+	file1 := r.WriteObject("potato", "hopefully not deleted", t1)
+	file2 := r.WriteFile("potato2", "hopefully copied in", t1)
+	fstest.CheckItems(t, r.fremote, file1)
+	fstest.CheckItems(t, r.flocal, file2)
+
+	fs.Stats.ResetCounters()
+	err := fs.CopyDir(r.fremote, r.flocal)
+	require.NoError(t, err)
+
+	fstest.CheckItems(t, r.fremote, file1, file2)
+	fstest.CheckItems(t, r.flocal, file2)
 }
 
 // Test with exclude
@@ -605,6 +647,45 @@ func TestSyncWithUpdateOlder(t *testing.T) {
 	fstest.CheckItems(t, r.fremote, oneO, twoF, threeO, fourF, fiveF)
 }
 
+// Test with TrackRenames set
+func TestSyncWithTrackRenames(t *testing.T) {
+	r := NewRun(t)
+	defer r.Finalise()
+
+	fs.Config.TrackRenames = true
+	defer func() {
+		fs.Config.TrackRenames = false
+
+	}()
+
+	haveHash := r.fremote.Hashes().Overlap(r.flocal.Hashes()).GetOne() != fs.HashNone
+	canTrackRenames := haveHash && fs.CanServerSideMove(r.fremote)
+	t.Logf("Can track renames: %v", canTrackRenames)
+
+	f1 := r.WriteFile("potato", "Potato Content", t1)
+	f2 := r.WriteFile("yam", "Yam Content", t2)
+
+	fs.Stats.ResetCounters()
+	require.NoError(t, fs.Sync(r.fremote, r.flocal))
+
+	fstest.CheckItems(t, r.fremote, f1, f2)
+	fstest.CheckItems(t, r.flocal, f1, f2)
+
+	// Now rename locally.
+	f2 = r.RenameFile(f2, "yaml")
+
+	fs.Stats.ResetCounters()
+	require.NoError(t, fs.Sync(r.fremote, r.flocal))
+
+	fstest.CheckItems(t, r.fremote, f1, f2)
+
+	if canTrackRenames {
+		assert.Equal(t, fs.Stats.GetTransfers(), int64(0))
+	} else {
+		assert.Equal(t, fs.Stats.GetTransfers(), int64(1))
+	}
+}
+
 // Test a server side move if possible, or the backup path if not
 func testServerSideMove(t *testing.T, r *Run, fremoteMove fs.Fs, withFilter bool) {
 	file1 := r.WriteBoth("potato2", "------------------------------------------------------------", t1)
@@ -680,7 +761,7 @@ func TestServerSideMoveOverlap(t *testing.T) {
 	r := NewRun(t)
 	defer r.Finalise()
 
-	if _, ok := r.fremote.(fs.DirMover); ok {
+	if r.fremote.Features().DirMove != nil {
 		t.Skip("Skipping test as remote supports DirMove")
 	}
 
@@ -703,3 +784,71 @@ func TestServerSideMoveOverlap(t *testing.T) {
 	err = fs.MoveDir(fremoteMove, r.fremote)
 	assert.EqualError(t, err, fs.ErrorCantMoveOverlapping.Error())
 }
+
+// Test with BackupDir set
+func testSyncBackupDir(t *testing.T, suffix string) {
+	r := NewRun(t)
+	defer r.Finalise()
+
+	if !fs.CanServerSideMove(r.fremote) {
+		t.Skip("Skipping test as remote does not support server side move")
+	}
+	r.Mkdir(r.fremote)
+
+	fs.Config.BackupDir = r.fremoteName + "/backup"
+	fs.Config.Suffix = suffix
+	defer func() {
+		fs.Config.BackupDir = ""
+		fs.Config.Suffix = ""
+	}()
+
+	// Make the setup so we have one, two, three in the dest
+	// and one (different), two (same) in the source
+	file1 := r.WriteObject("dst/one", "one", t1)
+	file2 := r.WriteObject("dst/two", "two", t1)
+	file3 := r.WriteObject("dst/three", "three", t1)
+	file2a := r.WriteFile("two", "two", t1)
+	file1a := r.WriteFile("one", "oneA", t2)
+
+	fstest.CheckItems(t, r.fremote, file1, file2, file3)
+	fstest.CheckItems(t, r.flocal, file1a, file2a)
+
+	fdst, err := fs.NewFs(r.fremoteName + "/dst")
+	require.NoError(t, err)
+
+	fs.Stats.ResetCounters()
+	err = fs.Sync(fdst, r.flocal)
+	require.NoError(t, err)
+
+	// one should be moved to the backup dir and the new one installed
+	file1.Path = "backup/one" + suffix
+	file1a.Path = "dst/one"
+	// two should be unchanged
+	// three should be moved to the backup dir
+	file3.Path = "backup/three" + suffix
+
+	fstest.CheckItems(t, r.fremote, file1, file2, file3, file1a)
+
+	// Now check what happens if we do it again
+	// Restore a different three and update one in the source
+	file3a := r.WriteObject("dst/three", "threeA", t2)
+	file1b := r.WriteFile("one", "oneBB", t3)
+	fstest.CheckItems(t, r.fremote, file1, file2, file3, file1a, file3a)
+
+	// This should delete three and overwrite one again, checking
+	// the files got overwritten correctly in backup-dir
+	fs.Stats.ResetCounters()
+	err = fs.Sync(fdst, r.flocal)
+	require.NoError(t, err)
+
+	// one should be moved to the backup dir and the new one installed
+	file1a.Path = "backup/one" + suffix
+	file1b.Path = "dst/one"
+	// two should be unchanged
+	// three should be moved to the backup dir
+	file3a.Path = "backup/three" + suffix
+
+	fstest.CheckItems(t, r.fremote, file1b, file2, file3a, file1a)
+}
+func TestSyncBackupDir(t *testing.T)           { testSyncBackupDir(t, "") }
+func TestSyncBackupDirWithSuffix(t *testing.T) { testSyncBackupDir(t, ".bak") }
