@@ -27,7 +27,8 @@ import (
 var (
 	// MatchTestRemote matches the remote names used for testing
 	MatchTestRemote = regexp.MustCompile(`^rclone-test-[abcdefghijklmnopqrstuvwxyz0123456789]{24}$`)
-	listRetries     = flag.Int("list-retries", 6, "Number or times to retry listing")
+	// ListRetries is the number of times to retry a listing to overcome eventual consistency
+	ListRetries = flag.Int("list-retries", 6, "Number or times to retry listing")
 )
 
 // Seed the random number generator
@@ -93,7 +94,7 @@ func (i *Item) CheckHashes(t *testing.T, obj fs.Object) {
 // Check checks all the attributes of the object are correct
 func (i *Item) Check(t *testing.T, obj fs.Object, precision time.Duration) {
 	i.CheckHashes(t, obj)
-	assert.Equal(t, i.Size, obj.Size())
+	assert.Equal(t, i.Size, obj.Size(), fmt.Sprintf("%s: size incorrect", i.Path))
 	i.CheckModTime(t, obj, obj.ModTime(), precision)
 }
 
@@ -143,6 +144,36 @@ func (is *Items) Done(t *testing.T) {
 	assert.Equal(t, 0, len(is.byName), fmt.Sprintf("%d objects not found", len(is.byName)))
 }
 
+// makeListingFromItems returns a string representation of the items
+//
+// it returns two possible strings, one normal and one for windows
+func makeListingFromItems(items []Item) (string, string) {
+	nameLengths1 := make([]string, len(items))
+	nameLengths2 := make([]string, len(items))
+	for i, item := range items {
+		remote1 := item.Path
+		remote2 := item.Path
+		if item.WinPath != "" {
+			remote2 = item.WinPath
+		}
+		nameLengths1[i] = fmt.Sprintf("%s (%d)", remote1, item.Size)
+		nameLengths2[i] = fmt.Sprintf("%s (%d)", remote2, item.Size)
+	}
+	sort.Strings(nameLengths1)
+	sort.Strings(nameLengths2)
+	return strings.Join(nameLengths1, ", "), strings.Join(nameLengths2, ", ")
+}
+
+// makeListingFromObjects returns a string representation of the objects
+func makeListingFromObjects(objs []fs.Object) string {
+	nameLengths := make([]string, len(objs))
+	for i, obj := range objs {
+		nameLengths[i] = fmt.Sprintf("%s (%d)", obj.Remote(), obj.Size())
+	}
+	sort.Strings(nameLengths)
+	return strings.Join(nameLengths, ", ")
+}
+
 // CheckListingWithPrecision checks the fs to see if it has the
 // expected contents with the given precision.
 //
@@ -155,14 +186,19 @@ func CheckListingWithPrecision(t *testing.T, f fs.Fs, items []Item, expectedDirs
 	var objs []fs.Object
 	var dirs []*fs.Dir
 	var err error
-	var retries = *listRetries
+	var retries = *ListRetries
 	sleep := time.Second / 2
+	wantListing1, wantListing2 := makeListingFromItems(items)
+	gotListing := "<unset>"
+	listingOK := false
 	for i := 1; i <= retries; i++ {
 		objs, dirs, err = fs.NewLister().Start(f, "").GetAll()
 		if err != nil && err != fs.ErrorDirNotFound {
 			t.Fatalf("Error listing: %v", err)
 		}
-		if len(objs) == len(items) && (expectedDirs == nil || len(dirs) == 0 || len(dirs) == len(expectedDirs)) {
+		gotListing = makeListingFromObjects(objs)
+		listingOK = wantListing1 == gotListing || wantListing2 == gotListing
+		if listingOK && (expectedDirs == nil || len(dirs) == 0 || len(dirs) == len(expectedDirs)) {
 			// Put an extra sleep in if we did any retries just to make sure it really
 			// is consistent (here is looking at you Amazon Drive!)
 			if i != 1 {
@@ -175,11 +211,12 @@ func CheckListingWithPrecision(t *testing.T, f fs.Fs, items []Item, expectedDirs
 		sleep *= 2
 		t.Logf("Sleeping for %v for list eventual consistency: %d/%d", sleep, i, retries)
 		time.Sleep(sleep)
-		if do, ok := f.(fs.DirCacheFlusher); ok {
+		if doDirCacheFlush := f.Features().DirCacheFlush; doDirCacheFlush != nil {
 			t.Logf("Flushing the directory cache")
-			do.DirCacheFlush()
+			doDirCacheFlush()
 		}
 	}
+	assert.True(t, listingOK, fmt.Sprintf("listing wrong, want\n  %s or\n  %s got\n  %s", wantListing1, wantListing2, gotListing))
 	for _, obj := range objs {
 		require.NotNil(t, obj)
 		is.Find(t, obj, precision)
