@@ -7,7 +7,6 @@ package fstests
 
 import (
 	"bytes"
-	"errors"
 	"flag"
 	"fmt"
 	"io"
@@ -21,6 +20,7 @@ import (
 
 	"github.com/ncw/rclone/fs"
 	"github.com/ncw/rclone/fstest"
+	"github.com/pkg/errors"
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
 )
@@ -54,8 +54,6 @@ var (
 // ExtraConfigItem describes a config item added on the fly while testing
 type ExtraConfigItem struct{ Name, Key, Value string }
 
-const eventualConsistencyRetries = 10
-
 func init() {
 	flag.StringVar(&RemoteName, "remote", "", "Set this to override the default remote name (eg s3:)")
 }
@@ -71,10 +69,11 @@ func TestInit(t *testing.T) {
 	fs.LoadConfig()
 	// Set extra config if supplied
 	for _, item := range ExtraConfig {
-		fs.ConfigFile.SetValue(item.Name, item.Key, item.Value)
+		fs.ConfigFileSet(item.Name, item.Key, item.Value)
+	}
+	if *verbose {
+		fs.Config.LogLevel = fs.LogLevelDebug
 	}
-	fs.Config.Verbose = *verbose
-	fs.Config.Quiet = !*verbose
 	fs.Config.DumpHeaders = *dumpHeaders
 	fs.Config.DumpBodies = *dumpBodies
 	t.Logf("Using remote %q", RemoteName)
@@ -204,12 +203,12 @@ func TestFsNewObjectNotFound(t *testing.T) {
 func findObject(t *testing.T, Name string) fs.Object {
 	var obj fs.Object
 	var err error
-	for i := 1; i <= eventualConsistencyRetries; i++ {
+	for i := 1; i <= *fstest.ListRetries; i++ {
 		obj, err = remote.NewObject(Name)
 		if err == nil {
 			break
 		}
-		t.Logf("Sleeping for 1 second for findObject eventual consistency: %d/%d (%v)", i, eventualConsistencyRetries, err)
+		t.Logf("Sleeping for 1 second for findObject eventual consistency: %d/%d (%v)", i, *fstest.ListRetries, err)
 		time.Sleep(1 * time.Second)
 	}
 	require.NoError(t, err)
@@ -297,23 +296,48 @@ func TestFsUpdateFile1(t *testing.T) {
 	// Note that the next test will check there are no duplicated file names
 }
 
-// TestFsListDirFile2 tests the files are correctly uploaded
+// TestFsListDirFile2 tests the files are correctly uploaded by doing
+// Depth 1 directory listings
 func TestFsListDirFile2(t *testing.T) {
 	skipIfNotOk(t)
-	var objNames, dirNames []string
-	for i := 1; i <= eventualConsistencyRetries; i++ {
-		objs, dirs, err := fs.NewLister().SetLevel(1).Start(remote, "").GetAll()
-		require.NoError(t, err)
-		objNames = objsToNames(objs)
-		dirNames = dirsToNames(dirs)
-		if len(objNames) >= 1 && len(dirNames) >= 1 {
-			break
+	list := func(dir string, expectedDirNames, expectedObjNames []string) {
+		var objNames, dirNames []string
+		for i := 1; i <= *fstest.ListRetries; i++ {
+			objs, dirs, err := fs.NewLister().SetLevel(1).Start(remote, dir).GetAll()
+			if err == fs.ErrorDirNotFound {
+				objs, dirs, err = fs.NewLister().SetLevel(1).Start(remote, winPath(dir)).GetAll()
+			}
+			require.NoError(t, err)
+			objNames = objsToNames(objs)
+			dirNames = dirsToNames(dirs)
+			if len(objNames) >= len(expectedObjNames) && len(dirNames) >= len(expectedDirNames) {
+				break
+			}
+			t.Logf("Sleeping for 1 second for TestFsListDirFile2 eventual consistency: %d/%d", i, *fstest.ListRetries)
+			time.Sleep(1 * time.Second)
 		}
-		t.Logf("Sleeping for 1 second for TestFsListDirFile2 eventual consistency: %d/%d", i, eventualConsistencyRetries)
-		time.Sleep(1 * time.Second)
+		assert.Equal(t, expectedDirNames, dirNames)
+		assert.Equal(t, expectedObjNames, objNames)
+	}
+	dir := file2.Path
+	deepest := true
+	for dir != "" {
+		expectedObjNames := []string{}
+		expectedDirNames := []string{}
+		child := dir
+		dir = path.Dir(dir)
+		if dir == "." {
+			dir = ""
+			expectedObjNames = append(expectedObjNames, winPath(file1.Path))
+		}
+		if deepest {
+			expectedObjNames = append(expectedObjNames, winPath(file2.Path))
+			deepest = false
+		} else {
+			expectedDirNames = append(expectedDirNames, winPath(child))
+		}
+		list(dir, expectedDirNames, expectedObjNames)
 	}
-	assert.Equal(t, []string{`hello_ sausage`}, dirNames)
-	assert.Equal(t, []string{file1.Path}, objNames)
 }
 
 // TestFsListDirRoot tests that DirList works in the root
@@ -379,13 +403,22 @@ func TestFsListFile1and2(t *testing.T) {
 	fstest.CheckListing(t, remote, []fstest.Item{file1, file2})
 }
 
+// TestFsNewObjectDir tests NewObject on a directory which should produce an error
+func TestFsNewObjectDir(t *testing.T) {
+	skipIfNotOk(t)
+	dir := path.Dir(file2.Path)
+	obj, err := remote.NewObject(dir)
+	assert.Nil(t, obj)
+	assert.NotNil(t, err)
+}
+
 // TestFsCopy tests Copy
 func TestFsCopy(t *testing.T) {
 	skipIfNotOk(t)
 
 	// Check have Copy
-	_, ok := remote.(fs.Copier)
-	if !ok {
+	doCopy := remote.Features().Copy
+	if doCopy == nil {
 		t.Skip("FS has no Copier interface")
 	}
 
@@ -394,7 +427,7 @@ func TestFsCopy(t *testing.T) {
 
 	// do the copy
 	src := findObject(t, file1.Path)
-	dst, err := remote.(fs.Copier).Copy(src, file1Copy.Path)
+	dst, err := doCopy(src, file1Copy.Path)
 	if err == fs.ErrorCantCopy {
 		t.Skip("FS can't copy")
 	}
@@ -417,8 +450,8 @@ func TestFsMove(t *testing.T) {
 	skipIfNotOk(t)
 
 	// Check have Move
-	_, ok := remote.(fs.Mover)
-	if !ok {
+	doMove := remote.Features().Move
+	if doMove == nil {
 		t.Skip("FS has no Mover interface")
 	}
 
@@ -432,8 +465,9 @@ func TestFsMove(t *testing.T) {
 	// check happy path, i.e. no naming conflicts when rename and move are two
 	// separate operations
 	file2Move.Path = "other.txt"
+	file2Move.WinPath = ""
 	src := findObject(t, file2.Path)
-	dst, err := remote.(fs.Mover).Move(src, file2Move.Path)
+	dst, err := doMove(src, file2Move.Path)
 	if err == fs.ErrorCantMove {
 		t.Skip("FS can't move")
 	}
@@ -448,7 +482,7 @@ func TestFsMove(t *testing.T) {
 	// Check conflict on "rename, then move"
 	file1Move.Path = "moveTest/other.txt"
 	src = findObject(t, file1.Path)
-	_, err = remote.(fs.Mover).Move(src, file1Move.Path)
+	_, err = doMove(src, file1Move.Path)
 	require.NoError(t, err)
 	fstest.CheckListing(t, remote, []fstest.Item{file1Move, file2Move})
 	// 1: moveTest/other.txt
@@ -456,14 +490,14 @@ func TestFsMove(t *testing.T) {
 
 	// Check conflict on "move, then rename"
 	src = findObject(t, file1Move.Path)
-	_, err = remote.(fs.Mover).Move(src, file1.Path)
+	_, err = doMove(src, file1.Path)
 	require.NoError(t, err)
 	fstest.CheckListing(t, remote, []fstest.Item{file1, file2Move})
 	// 1: file name.txt
 	// 2: other.txt
 
 	src = findObject(t, file2Move.Path)
-	_, err = remote.(fs.Mover).Move(src, file2.Path)
+	_, err = doMove(src, file2.Path)
 	require.NoError(t, err)
 	fstest.CheckListing(t, remote, []fstest.Item{file1, file2})
 	// 1: file name.txt
@@ -479,17 +513,19 @@ func TestFsMove(t *testing.T) {
 // If destination exists then return fs.ErrorDirExists
 
 // TestFsDirMove tests DirMove
+//
+// go test -v -run '^Test(Setup|Init|FsMkdir|FsPutFile1|FsPutFile2|FsUpdateFile1|FsDirMove)$
 func TestFsDirMove(t *testing.T) {
 	skipIfNotOk(t)
 
 	// Check have DirMove
-	_, ok := remote.(fs.DirMover)
-	if !ok {
+	doDirMove := remote.Features().DirMove
+	if doDirMove == nil {
 		t.Skip("FS has no DirMover interface")
 	}
 
 	// Check it can't move onto itself
-	err := remote.(fs.DirMover).DirMove(remote)
+	err := doDirMove(remote, "", "")
 	require.Equal(t, fs.ErrorDirExists, err)
 
 	// new remote
@@ -497,17 +533,23 @@ func TestFsDirMove(t *testing.T) {
 	require.NoError(t, err)
 	defer removeNewRemote()
 
+	const newName = "new_name/sub_new_name"
 	// try the move
-	err = newRemote.(fs.DirMover).DirMove(remote)
+	err = newRemote.Features().DirMove(remote, "", newName)
 	require.NoError(t, err)
 
 	// check remotes
 	// FIXME: Prints errors.
 	fstest.CheckListing(t, remote, []fstest.Item{})
-	fstest.CheckListing(t, newRemote, []fstest.Item{file2, file1})
+	file1Copy := file1
+	file1Copy.Path = path.Join(newName, file1.Path)
+	file2Copy := file2
+	file2Copy.Path = path.Join(newName, file2.Path)
+	file2Copy.WinPath = path.Join(newName, file2.WinPath)
+	fstest.CheckListing(t, newRemote, []fstest.Item{file2Copy, file1Copy})
 
 	// move it back
-	err = remote.(fs.DirMover).DirMove(newRemote)
+	err = doDirMove(newRemote, newName, "")
 	require.NoError(t, err)
 
 	// check remotes
@@ -550,8 +592,8 @@ func TestObjectFs(t *testing.T) {
 	testRemote := remote
 	if obj.Fs() != testRemote {
 		// Check to see if this wraps something else
-		if unwrap, ok := testRemote.(fs.UnWrapper); ok {
-			testRemote = unwrap.UnWrap()
+		if doUnWrap := testRemote.Features().UnWrap; doUnWrap != nil {
+			testRemote = doUnWrap()
 		}
 	}
 	assert.Equal(t, obj.Fs(), testRemote)
@@ -684,6 +726,7 @@ func TestFsIsFile(t *testing.T) {
 	remoteName := subRemoteName + "/" + file2.Path
 	file2Copy := file2
 	file2Copy.Path = "z.txt"
+	file2Copy.WinPath = ""
 	fileRemote, err := fs.NewFs(remoteName)
 	assert.Equal(t, fs.ErrorIsFile, err)
 	fstest.CheckListing(t, fileRemote, []fstest.Item{file2Copy})
