@@ -62,7 +62,7 @@ func init() {
 		Description: "Google Cloud Storage (this is not Google Drive)",
 		NewFs:       NewFs,
 		Config: func(name string) {
-			if fs.ConfigFile.MustValue(name, "service_account_file") != "" {
+			if fs.ConfigFileGet(name, "service_account_file") != "" {
 				return
 			}
 			err := oauthutil.Config("google cloud storage", name, storageConfig)
@@ -130,10 +130,11 @@ func init() {
 // Fs represents a remote storage server
 type Fs struct {
 	name          string           // name of this remote
+	root          string           // the path we are working on if any
+	features      *fs.Features     // optional features
 	svc           *storage.Service // the connection to the storage server
 	client        *http.Client     // authorized client
 	bucket        string           // the bucket we are working on
-	root          string           // the path we are working on if any
 	projectNumber string           // used for finding buckets
 	objectACL     string           // used when creating new objects
 	bucketACL     string           // used when creating new buckets
@@ -175,6 +176,11 @@ func (f *Fs) String() string {
 	return fmt.Sprintf("Storage bucket %s path %s", f.bucket, f.root)
 }
 
+// Features returns the optional features of this Fs
+func (f *Fs) Features() *fs.Features {
+	return f.features
+}
+
 // Pattern to match a storage path
 var matcher = regexp.MustCompile(`^([^/]*)(.*)$`)
 
@@ -208,7 +214,7 @@ func NewFs(name, root string) (fs.Fs, error) {
 	var oAuthClient *http.Client
 	var err error
 
-	serviceAccountPath := fs.ConfigFile.MustValue(name, "service_account_file")
+	serviceAccountPath := fs.ConfigFileGet(name, "service_account_file")
 	if serviceAccountPath != "" {
 		oAuthClient, err = getServiceAccountClient(serviceAccountPath)
 		if err != nil {
@@ -230,10 +236,11 @@ func NewFs(name, root string) (fs.Fs, error) {
 		name:          name,
 		bucket:        bucket,
 		root:          directory,
-		projectNumber: fs.ConfigFile.MustValue(name, "project_number"),
-		objectACL:     fs.ConfigFile.MustValue(name, "object_acl"),
-		bucketACL:     fs.ConfigFile.MustValue(name, "bucket_acl"),
+		projectNumber: fs.ConfigFileGet(name, "project_number"),
+		objectACL:     fs.ConfigFileGet(name, "object_acl"),
+		bucketACL:     fs.ConfigFileGet(name, "bucket_acl"),
 	}
+	f.features = (&fs.Features{ReadMimeType: true, WriteMimeType: true}).Fill(f)
 	if f.objectACL == "" {
 		f.objectACL = "private"
 	}
@@ -324,7 +331,7 @@ func (f *Fs) list(dir string, level int, fn listFn) error {
 				if !strings.HasSuffix(prefix, "/") {
 					continue
 				}
-				err = fn(prefix[:len(prefix)-1], &object, true)
+				err = fn(prefix[rootLength:len(prefix)-1], &object, true)
 				if err != nil {
 					return err
 				}
@@ -332,7 +339,7 @@ func (f *Fs) list(dir string, level int, fn listFn) error {
 		}
 		for _, object := range objects.Items {
 			if !strings.HasPrefix(object.Name, root) {
-				fs.Log(f, "Odd name received %q", object.Name)
+				fs.Logf(f, "Odd name received %q", object.Name)
 				continue
 			}
 			remote := object.Name[rootLength:]
@@ -498,7 +505,7 @@ func (f *Fs) Precision() time.Duration {
 func (f *Fs) Copy(src fs.Object, remote string) (fs.Object, error) {
 	srcObj, ok := src.(*Object)
 	if !ok {
-		fs.Debug(src, "Can't copy - not same remote type")
+		fs.Debugf(src, "Can't copy - not same remote type")
 		return nil, fs.ErrorCantCopy
 	}
 
@@ -568,7 +575,7 @@ func (o *Object) setMetaData(info *storage.Object) {
 	// Read md5sum
 	md5sumData, err := base64.StdEncoding.DecodeString(info.Md5Hash)
 	if err != nil {
-		fs.Log(o, "Bad MD5 decode: %v", err)
+		fs.Logf(o, "Bad MD5 decode: %v", err)
 	} else {
 		o.md5sum = hex.EncodeToString(md5sumData)
 	}
@@ -581,13 +588,13 @@ func (o *Object) setMetaData(info *storage.Object) {
 			o.modTime = modTime
 			return
 		}
-		fs.Debug(o, "Failed to read mtime from metadata: %s", err)
+		fs.Debugf(o, "Failed to read mtime from metadata: %s", err)
 	}
 
 	// Fallback to the Updated time
 	modTime, err := time.Parse(timeFormatIn, info.Updated)
 	if err != nil {
-		fs.Log(o, "Bad time decode: %v", err)
+		fs.Logf(o, "Bad time decode: %v", err)
 	} else {
 		o.modTime = modTime
 	}
@@ -620,7 +627,7 @@ func (o *Object) readMetaData() (err error) {
 func (o *Object) ModTime() time.Time {
 	err := o.readMetaData()
 	if err != nil {
-		// fs.Log(o, "Failed to read metadata: %v", err)
+		// fs.Logf(o, "Failed to read metadata: %v", err)
 		return time.Now()
 	}
 	return o.modTime
