@@ -21,6 +21,10 @@ import (
 	"github.com/pkg/errors"
 )
 
+var (
+	followSymlinks = fs.BoolP("copy-links", "L", false, "Follow symlinks and copy the pointed to item.")
+)
+
 // Constants
 const devUnset = 0xdeadbeefcafebabe // a device id meaning it is unset
 
@@ -47,12 +51,16 @@ func init() {
 type Fs struct {
 	name        string              // the name of the remote
 	root        string              // The root directory (OS path)
+	features    *fs.Features        // optional features
 	dev         uint64              // device number of root node
 	precisionOk sync.Once           // Whether we need to read the precision
 	precision   time.Duration       // precision of local filesystem
 	wmu         sync.Mutex          // used for locking access to 'warned'.
 	warned      map[string]struct{} // whether we have warned about this string
 	nounc       bool                // Skip UNC conversion on Windows
+	// do os.Lstat or os.Stat
+	lstat    func(name string) (os.FileInfo, error)
+	dirNames *mapper // directory name mapping
 }
 
 // Object represents a local filesystem object
@@ -70,17 +78,23 @@ type Object struct {
 func NewFs(name, root string) (fs.Fs, error) {
 	var err error
 
-	nounc, _ := fs.ConfigFile.GetValue(name, "nounc")
+	nounc := fs.ConfigFileGet(name, "nounc")
 	f := &Fs{
-		name:   name,
-		warned: make(map[string]struct{}),
-		nounc:  nounc == "true",
-		dev:    devUnset,
+		name:     name,
+		warned:   make(map[string]struct{}),
+		nounc:    nounc == "true",
+		dev:      devUnset,
+		lstat:    os.Lstat,
+		dirNames: newMapper(),
 	}
 	f.root = f.cleanPath(root)
+	f.features = (&fs.Features{CaseInsensitive: f.caseInsensitive()}).Fill(f)
+	if *followSymlinks {
+		f.lstat = os.Stat
+	}
 
 	// Check to see if this points to a file
-	fi, err := os.Lstat(f.root)
+	fi, err := f.lstat(f.root)
 	if err == nil {
 		f.dev = readDevice(fi)
 	}
@@ -108,9 +122,28 @@ func (f *Fs) String() string {
 	return fmt.Sprintf("Local file system at %s", f.root)
 }
 
+// Features returns the optional features of this Fs
+func (f *Fs) Features() *fs.Features {
+	return f.features
+}
+
+// caseInsenstive returns whether the remote is case insensitive or not
+func (f *Fs) caseInsensitive() bool {
+	// FIXME not entirely accurate since you can have case
+	// sensitive Fses on darwin and case insenstive Fses on linux.
+	// Should probably check but that would involve creating a
+	// file in the remote to be most accurate which probably isn't
+	// desirable.
+	return runtime.GOOS == "windows" || runtime.GOOS == "darwin"
+}
+
 // newObject makes a half completed Object
-func (f *Fs) newObject(remote string) *Object {
-	dstPath := f.cleanPath(filepath.Join(f.root, remote))
+//
+// if dstPath is empty then it is made from remote
+func (f *Fs) newObject(remote, dstPath string) *Object {
+	if dstPath == "" {
+		dstPath = f.cleanPath(filepath.Join(f.root, remote))
+	}
 	remote = f.cleanRemote(remote)
 	return &Object{
 		fs:     f,
@@ -122,8 +155,8 @@ func (f *Fs) newObject(remote string) *Object {
 // Return an Object from a path
 //
 // May return nil if an error occurred
-func (f *Fs) newObjectWithInfo(remote string, info os.FileInfo) (fs.Object, error) {
-	o := f.newObject(remote)
+func (f *Fs) newObjectWithInfo(remote, dstPath string, info os.FileInfo) (fs.Object, error) {
+	o := f.newObject(remote, dstPath)
 	if info != nil {
 		o.info = info
 	} else {
@@ -135,13 +168,16 @@ func (f *Fs) newObjectWithInfo(remote string, info os.FileInfo) (fs.Object, erro
 			return nil, err
 		}
 	}
+	if o.info.Mode().IsDir() {
+		return nil, errors.Wrapf(fs.ErrorNotAFile, "%q", remote)
+	}
 	return o, nil
 }
 
 // NewObject finds the Object at remote.  If it can't be found
 // it returns the error ErrorObjectNotFound.
 func (f *Fs) NewObject(remote string) (fs.Object, error) {
-	return f.newObjectWithInfo(remote, nil)
+	return f.newObjectWithInfo(remote, "", nil)
 }
 
 // listArgs is the arguments that a new list takes
@@ -180,14 +216,24 @@ func (f *Fs) list(out fs.ListOpts, remote string, dirpath string, level int) (su
 
 		for _, fi := range fis {
 			name := fi.Name()
+			mode := fi.Mode()
 			newRemote := path.Join(remote, name)
 			newPath := filepath.Join(dirpath, name)
+			// Follow symlinks if required
+			if *followSymlinks && (mode&os.ModeSymlink) != 0 {
+				fi, err = os.Stat(newPath)
+				if err != nil {
+					out.SetError(err)
+					return nil
+				}
+				mode = fi.Mode()
+			}
 			if fi.IsDir() {
 				// Ignore directories which are symlinks.  These are junction points under windows which
 				// are kind of a souped up symlink. Unix doesn't have directories which are symlinks.
-				if (fi.Mode()&os.ModeSymlink) == 0 && out.IncludeDirectory(newRemote) {
+				if (mode&os.ModeSymlink) == 0 && out.IncludeDirectory(newRemote) && f.dev == readDevice(fi) {
 					dir := &fs.Dir{
-						Name:  f.cleanRemote(newRemote),
+						Name:  f.dirNames.Save(newRemote, f.cleanRemote(newRemote)),
 						When:  fi.ModTime(),
 						Bytes: 0,
 						Count: 0,
@@ -195,12 +241,12 @@ func (f *Fs) list(out fs.ListOpts, remote string, dirpath string, level int) (su
 					if out.AddDir(dir) {
 						return nil
 					}
-					if level > 0 && f.dev == readDevice(fi) {
+					if level > 0 {
 						subdirs = append(subdirs, listArgs{remote: newRemote, dirpath: newPath, level: level - 1})
 					}
 				}
 			} else {
-				fso, err := f.newObjectWithInfo(newRemote, fi)
+				fso, err := f.newObjectWithInfo(newRemote, newPath, fi)
 				if err != nil {
 					out.SetError(err)
 					return nil
@@ -219,6 +265,7 @@ func (f *Fs) list(out fs.ListOpts, remote string, dirpath string, level int) (su
 // Ignores everything which isn't Storable, eg links etc
 func (f *Fs) List(out fs.ListOpts, dir string) {
 	defer out.Finished()
+	dir = f.dirNames.Load(dir)
 	root := f.cleanPath(filepath.Join(f.root, dir))
 	dir = f.cleanRemote(dir)
 	_, err := os.Stat(root)
@@ -272,7 +319,7 @@ func (f *Fs) cleanRemote(name string) string {
 	if !utf8.ValidString(name) {
 		f.wmu.Lock()
 		if _, ok := f.warned[name]; !ok {
-			fs.Debug(f, "Replacing invalid UTF-8 characters in %q", name)
+			fs.Logf(f, "Replacing invalid UTF-8 characters in %q", name)
 			f.warned[name] = struct{}{}
 		}
 		f.wmu.Unlock()
@@ -283,11 +330,49 @@ func (f *Fs) cleanRemote(name string) string {
 	return name
 }
 
+// mapper maps raw to cleaned directory names
+type mapper struct {
+	mu sync.RWMutex      // mutex to protect the below
+	m  map[string]string // map of un-normalised directory names
+}
+
+func newMapper() *mapper {
+	return &mapper{
+		m: make(map[string]string),
+	}
+}
+
+// Lookup a directory name to make a local name (reverses
+// cleanDirName)
+//
+// FIXME this is temporary before we make a proper Directory object
+func (m *mapper) Load(in string) string {
+	m.mu.RLock()
+	out, ok := m.m[in]
+	m.mu.RUnlock()
+	if ok {
+		return out
+	}
+	return in
+}
+
+// Cleans a directory name recording if it needed to be altered
+//
+// FIXME this is temporary before we make a proper Directory object
+func (m *mapper) Save(in, out string) string {
+	if in != out {
+		m.mu.Lock()
+		m.m[out] = in
+		m.mu.Unlock()
+	}
+	return out
+}
+
 // Put the Object to the local filesystem
 func (f *Fs) Put(in io.Reader, src fs.ObjectInfo) (fs.Object, error) {
 	remote := src.Remote()
 	// Temporary Object under construction - info filled in by Update()
-	o := f.newObject(remote)
+	o := f.newObject(remote, "")
 	err := o.Update(in, src)
 	if err != nil {
 		return nil, err
@@ -304,7 +389,7 @@ func (f *Fs) Mkdir(dir string) error {
 		return err
 	}
 	if dir == "" {
-		fi, err := os.Lstat(root)
+		fi, err := f.lstat(root)
 		if err != nil {
 			return err
 		}
@@ -387,7 +472,7 @@ func (f *Fs) readPrecision() (precision time.Duration) {
 // deleting all the files quicker than just running Remove() on the
 // result of List()
 func (f *Fs) Purge() error {
-	fi, err := os.Lstat(f.root)
+	fi, err := f.lstat(f.root)
 	if err != nil {
 		return err
 	}
@@ -409,12 +494,12 @@ func (f *Fs) Purge() error {
 func (f *Fs) Move(src fs.Object, remote string) (fs.Object, error) {
 	srcObj, ok := src.(*Object)
 	if !ok {
-		fs.Debug(src, "Can't move - not same remote type")
+		fs.Debugf(src, "Can't move - not same remote type")
 		return nil, fs.ErrorCantMove
 	}
 
 	// Temporary Object under construction
-	dstObj := f.newObject(remote)
+	dstObj := f.newObject(remote, "")
 
 	// Check it is a file if it exists
 	err := dstObj.lstat()
@@ -448,38 +533,38 @@ func (f *Fs) Move(src fs.Object, remote string) (fs.Object, error) {
 	return dstObj, nil
 }
 
-// DirMove moves src directory to this remote using server side move
-// operations.
+// DirMove moves src, srcRemote to this remote at dstRemote
+// using server side move operations.
 //
 // Will only be called if src.Fs().Name() == f.Name()
 //
 // If it isn't possible then return fs.ErrorCantDirMove
 //
 // If destination exists then return fs.ErrorDirExists
-func (f *Fs) DirMove(src fs.Fs) error {
+func (f *Fs) DirMove(src fs.Fs, srcRemote, dstRemote string) error {
 	srcFs, ok := src.(*Fs)
 	if !ok {
-		fs.Debug(srcFs, "Can't move directory - not same remote type")
-		return fs.ErrorCantDirMove
-	}
-	// Check if source exists
-	sstat, err := os.Lstat(srcFs.root)
-	if err != nil {
-		return err
-	}
-	// And is a directory
-	if !sstat.IsDir() {
+		fs.Debugf(srcFs, "Can't move directory - not same remote type")
 		return fs.ErrorCantDirMove
 	}
+	srcPath := f.cleanPath(filepath.Join(srcFs.root, srcRemote))
+	dstPath := f.cleanPath(filepath.Join(f.root, dstRemote))
 
 	// Check if destination exists
-	_, err = os.Lstat(f.root)
+	_, err := os.Lstat(dstPath)
 	if !os.IsNotExist(err) {
 		return fs.ErrorDirExists
 	}
 
+	// Create parent of destination
+	dstParentPath, _ := getDirFile(dstPath)
+	err = os.MkdirAll(dstParentPath, 0777)
+	if err != nil {
+		return err
+	}
+
 	// Do the move
-	return os.Rename(srcFs.root, f.root)
+	return os.Rename(srcPath, dstPath)
 }
 
 // Hashes returns the supported hash sets.
@@ -564,21 +649,24 @@ func (o *Object) Storable() bool {
 	// Check for control characters in the remote name and show non storable
 	for _, c := range o.Remote() {
 		if c >= 0x00 && c < 0x20 || c == 0x7F {
-			fs.Debug(o.fs, "Can't store file with control characters: %q", o.Remote())
+			fs.Logf(o.fs, "Can't store file with control characters: %q", o.Remote())
 			return false
 		}
 	}
 	mode := o.info.Mode()
 	// On windows a file with os.ModeSymlink represents a file with reparse points
 	if runtime.GOOS == "windows" && (mode&os.ModeSymlink) != 0 {
-		fs.Debug(o, "Clearing symlink bit to allow a file with reparse points to be copied")
+		fs.Debugf(o, "Clearing symlink bit to allow a file with reparse points to be copied")
 		mode &^= os.ModeSymlink
 	}
-	if mode&(os.ModeSymlink|os.ModeNamedPipe|os.ModeSocket|os.ModeDevice) != 0 {
-		fs.Debug(o, "Can't transfer non file/directory")
+	if mode&os.ModeSymlink != 0 {
+		fs.Logf(o, "Can't follow symlink without -L/--copy-links")
+		return false
+	} else if mode&(os.ModeNamedPipe|os.ModeSocket|os.ModeDevice) != 0 {
+		fs.Logf(o, "Can't transfer non file/directory")
 		return false
 	} else if mode&os.ModeDir != 0 {
-		// fs.Debug(o, "Skipping directory")
+		// fs.Debugf(o, "Skipping directory")
 		return false
 	}
 	return true
@@ -622,7 +710,7 @@ func (o *Object) Open(options ...fs.OpenOption) (in io.ReadCloser, err error) {
 			offset = x.Offset
 		default:
 			if option.Mandatory() {
-				fs.Log(o, "Unsupported mandatory option: %v", option)
+				fs.Logf(o, "Unsupported mandatory option: %v", option)
 			}
 		}
 	}
@@ -659,7 +747,7 @@ func (o *Object) Update(in io.Reader, src fs.ObjectInfo) error {
 		return err
 	}
 
-	out, err := os.Create(o.path)
+	out, err := os.OpenFile(o.path, os.O_WRONLY|os.O_CREATE|os.O_TRUNC, 0666)
 	if err != nil {
 		return err
 	}
@@ -674,9 +762,9 @@ func (o *Object) Update(in io.Reader, src fs.ObjectInfo) error {
 		err = closeErr
 	}
 	if err != nil {
-		fs.Debug(o, "Removing partially written file on error: %v", err)
+		fs.Logf(o, "Removing partially written file on error: %v", err)
 		if removeErr := os.Remove(o.path); removeErr != nil {
-			fs.ErrorLog(o, "Failed to remove partially written file: %v", removeErr)
+			fs.Errorf(o, "Failed to remove partially written file: %v", removeErr)
 		}
 		return err
 	}
@@ -696,7 +784,7 @@ func (o *Object) Update(in io.Reader, src fs.ObjectInfo) error {
 
 // Stat a Object into info
 func (o *Object) lstat() error {
-	info, err := os.Lstat(o.path)
+	info, err := o.fs.lstat(o.path)
 	o.info = info
 	return err
 }
@@ -819,7 +907,7 @@ func cleanWindowsName(f *Fs, name string) string {
 	if name2 != original && f != nil {
 		f.wmu.Lock()
 		if _, ok := f.warned[name]; !ok {
-			fs.Debug(f, "Replacing invalid characters in %q to %q", name, name2)
+			fs.Logf(f, "Replacing invalid characters in %q to %q", name, name2)
 			f.warned[name] = struct{}{}
 		}
 		f.wmu.Unlock()
