@@ -9,11 +9,10 @@ import (
 	"syscall"
 
 	"github.com/ncw/rclone/fs"
-	"github.com/spf13/pflag"
 )
 
 var (
-	oneFileSystem = pflag.BoolP("one-file-system", "x", false, "Don't cross filesystem boundaries.")
+	oneFileSystem = fs.BoolP("one-file-system", "x", false, "Don't cross filesystem boundaries.")
 )
 
 // readDevice turns a valid os.FileInfo into a device number,
@@ -24,7 +23,7 @@ func readDevice(fi os.FileInfo) uint64 {
 	}
 	statT, ok := fi.Sys().(*syscall.Stat_t)
 	if !ok {
-		fs.Debug(fi.Name(), "Type assertion fi.Sys().(*syscall.Stat_t) failed from: %#v", fi.Sys())
+		fs.Debugf(fi.Name(), "Type assertion fi.Sys().(*syscall.Stat_t) failed from: %#v", fi.Sys())
 		return devUnset
 	}
 	return uint64(statT.Dev)
