@@ -4,6 +4,7 @@ package onedrive
 
 import (
 	"bytes"
+	"encoding/json"
 	"fmt"
 	"io"
 	"log"
@@ -21,7 +22,6 @@ import (
 	"github.com/ncw/rclone/pacer"
 	"github.com/ncw/rclone/rest"
 	"github.com/pkg/errors"
-	"github.com/spf13/pflag"
 	"golang.org/x/oauth2"
 )
 
@@ -75,17 +75,19 @@ func init() {
 			Help: "Microsoft App Client Secret - leave blank normally.",
 		}},
 	})
-	pflag.VarP(&chunkSize, "onedrive-chunk-size", "", "Above this size files will be chunked - must be multiple of 320k.")
-	pflag.VarP(&uploadCutoff, "onedrive-upload-cutoff", "", "Cutoff for switching to chunked upload - must be <= 100MB")
+	fs.VarP(&chunkSize, "onedrive-chunk-size", "", "Above this size files will be chunked - must be multiple of 320k.")
+	fs.VarP(&uploadCutoff, "onedrive-upload-cutoff", "", "Cutoff for switching to chunked upload - must be <= 100MB")
 }
 
 // Fs represents a remote one drive
 type Fs struct {
-	name     string             // name of this remote
-	srv      *rest.Client       // the connection to the one drive server
-	root     string             // the path we are working on
-	dirCache *dircache.DirCache // Map of directory path to directory id
-	pacer    *pacer.Pacer       // pacer for API calls
+	name         string             // name of this remote
+	root         string             // the path we are working on
+	features     *fs.Features       // optional features
+	srv          *rest.Client       // the connection to the one drive server
+	dirCache     *dircache.DirCache // Map of directory path to directory id
+	pacer        *pacer.Pacer       // pacer for API calls
+	tokenRenewer *oauthutil.Renew   // renew the token on expiry
 }
 
 // Object describes a one drive object
@@ -119,6 +121,11 @@ func (f *Fs) String() string {
 	return fmt.Sprintf("One drive root '%s'", f.root)
 }
 
+// Features returns the optional features of this Fs
+func (f *Fs) Features() *fs.Features {
+	return f.features
+}
+
 // Pattern to match a one drive path
 var matcher = regexp.MustCompile(`^([^/]*)(.*)$`)
 
@@ -163,7 +170,7 @@ func errorHandler(resp *http.Response) error {
 	errResponse := new(api.Error)
 	err := rest.DecodeJSON(resp, &errResponse)
 	if err != nil {
-		fs.Debug(nil, "Couldn't decode error response: %v", err)
+		fs.Debugf(nil, "Couldn't decode error response: %v", err)
 	}
 	if errResponse.ErrorInfo.Code == "" {
 		errResponse.ErrorInfo.Code = resp.Status
@@ -174,7 +181,7 @@ func errorHandler(resp *http.Response) error {
 // NewFs constructs an Fs from the path, container:path
 func NewFs(name, root string) (fs.Fs, error) {
 	root = parsePath(root)
-	oAuthClient, _, err := oauthutil.NewClient(name, oauthConfig)
+	oAuthClient, ts, err := oauthutil.NewClient(name, oauthConfig)
 	if err != nil {
 		log.Fatalf("Failed to configure One Drive: %v", err)
 	}
@@ -185,6 +192,7 @@ func NewFs(name, root string) (fs.Fs, error) {
 		srv:   rest.NewClient(oAuthClient).SetRoot(rootURL),
 		pacer: pacer.New().SetMinSleep(minSleep).SetMaxSleep(maxSleep).SetDecayConstant(decayConstant),
 	}
+	f.features = (&fs.Features{CaseInsensitive: true, ReadMimeType: true}).Fill(f)
 	f.srv.SetErrorHandler(errorHandler)
 
 	// Get rootID
@@ -193,6 +201,12 @@ func NewFs(name, root string) (fs.Fs, error) {
 		return nil, errors.Wrap(err, "failed to get root")
 	}
 
+	// Renew the token in the background
+	f.tokenRenewer = oauthutil.NewRenew(f.String(), ts, func() error {
+		_, _, err := f.readMetaDataForPath("")
+		return err
+	})
+
 	f.dirCache = dircache.New(root, rootInfo.ID, f)
 
 	// Find the current root
@@ -239,14 +253,15 @@ func (f *Fs) newObjectWithInfo(remote string, info *api.Item) (fs.Object, error)
 		fs:     f,
 		remote: remote,
 	}
+	var err error
 	if info != nil {
 		// Set info
-		o.setMetaData(info)
+		err = o.setMetaData(info)
 	} else {
-		err := o.readMetaData() // reads info and meta, returning an error
-		if err != nil {
-			return nil, err
-		}
+		err = o.readMetaData() // reads info and meta, returning an error
+	}
+	if err != nil {
+		return nil, err
 	}
 	return o, nil
 }
@@ -259,7 +274,7 @@ func (f *Fs) NewObject(remote string) (fs.Object, error) {
 
 // FindLeaf finds a directory of name leaf in the folder with ID pathID
 func (f *Fs) FindLeaf(pathID, leaf string) (pathIDOut string, found bool, err error) {
-	// fs.Debug(f, "FindLeaf(%q, %q)", pathID, leaf)
+	// fs.Debugf(f, "FindLeaf(%q, %q)", pathID, leaf)
 	parent, ok := f.dirCache.GetInv(pathID)
 	if !ok {
 		return "", false, errors.New("couldn't find parent ID")
@@ -286,7 +301,7 @@ func (f *Fs) FindLeaf(pathID, leaf string) (pathIDOut string, found bool, err er
 
 // CreateDir makes a directory with pathID as parent and name leaf
 func (f *Fs) CreateDir(pathID, leaf string) (newID string, err error) {
-	// fs.Debug(f, "CreateDir(%q, %q)\n", pathID, leaf)
+	// fs.Debugf(f, "CreateDir(%q, %q)\n", pathID, leaf)
 	var resp *http.Response
 	var info *api.Item
 	opts := rest.Opts{
@@ -373,7 +388,7 @@ OUTER:
 
 // ListDir reads the directory specified by the job into out, returning any more jobs
 func (f *Fs) ListDir(out fs.ListOpts, job dircache.ListDirJob) (jobs []dircache.ListDirJob, err error) {
-	fs.Debug(f, "Reading %q", job.Path)
+	fs.Debugf(f, "Reading %q", job.Path)
 	_, err = f.listAll(job.DirID, false, false, func(info *api.Item) bool {
 		remote := job.Path + info.Name
 		if info.Folder != nil {
@@ -406,7 +421,7 @@ func (f *Fs) ListDir(out fs.ListOpts, job dircache.ListDirJob) (jobs []dircache.
 		}
 		return false
 	})
-	fs.Debug(f, "Finished reading %q", job.Path)
+	fs.Debugf(f, "Finished reading %q", job.Path)
 	return jobs, err
 }
 
@@ -423,7 +438,7 @@ func (f *Fs) List(out fs.ListOpts, dir string) {
 // Used to create new objects
 func (f *Fs) createObject(remote string, modTime time.Time, size int64) (o *Object, leaf string, directoryID string, err error) {
 	// Create the directory for the object if it doesn't exist
-	leaf, directoryID, err = f.dirCache.FindPath(remote, true)
+	leaf, directoryID, err = f.dirCache.FindRootAndPath(remote, true)
 	if err != nil {
 		return nil, leaf, directoryID, err
 	}
@@ -485,6 +500,10 @@ func (f *Fs) purgeCheck(dir string, check bool) error {
 		return errors.New("can't purge root directory")
 	}
 	dc := f.dirCache
+	err := dc.FindRoot(false)
+	if err != nil {
+		return err
+	}
 	rootID, err := dc.FindDir(dir, false)
 	if err != nil {
 		return err
@@ -527,36 +546,43 @@ func (f *Fs) waitForJob(location string, o *Object) error {
 	deadline := time.Now().Add(fs.Config.Timeout)
 	for time.Now().Before(deadline) {
 		opts := rest.Opts{
-			Method:   "GET",
-			Path:     location,
-			Absolute: true,
+			Method:       "GET",
+			Path:         location,
+			Absolute:     true,
+			IgnoreStatus: true, // Ignore the http status response since it seems to return valid info on 500 errors
 		}
 		var resp *http.Response
 		var err error
+		var body []byte
 		err = f.pacer.Call(func() (bool, error) {
 			resp, err = f.srv.Call(&opts)
-			return shouldRetry(resp, err)
+			if err != nil {
+				return fs.ShouldRetry(err), err
+			}
+			body, err = rest.ReadBody(resp)
+			return fs.ShouldRetry(err), err
 		})
 		if err != nil {
 			return err
 		}
-		if resp.StatusCode == 202 {
-			var status api.AsyncOperationStatus
-			err = rest.DecodeJSON(resp, &status)
-			if err != nil {
-				return err
-			}
+		// Try to decode the body first as an api.AsyncOperationStatus
+		var status api.AsyncOperationStatus
+		err = json.Unmarshal(body, &status)
+		if err != nil {
+			return errors.Wrapf(err, "async status result not JSON: %q", body)
+		}
+		// See if we decoded anything...
+		if !(status.Operation == "" && status.PercentageComplete == 0 && status.Status == "") {
 			if status.Status == "failed" || status.Status == "deleteFailed" {
-				return errors.Errorf("async operation %q returned %q", status.Operation, status.Status)
+				return errors.Errorf("%s: async operation %q returned %q", o.remote, status.Operation, status.Status)
 			}
-		} else {
+		} else if resp.StatusCode == 200 {
 			var info api.Item
-			err = rest.DecodeJSON(resp, &info)
+			err = json.Unmarshal(body, &info)
 			if err != nil {
-				return err
+				return errors.Wrapf(err, "async item result not JSON: %q", body)
 			}
-			o.setMetaData(&info)
-			return nil
+			return o.setMetaData(&info)
 		}
 		time.Sleep(1 * time.Second)
 	}
@@ -575,7 +601,7 @@ func (f *Fs) waitForJob(location string, o *Object) error {
 func (f *Fs) Copy(src fs.Object, remote string) (fs.Object, error) {
 	srcObj, ok := src.(*Object)
 	if !ok {
-		fs.Debug(src, "Can't copy - not same remote type")
+		fs.Debugf(src, "Can't copy - not same remote type")
 		return nil, fs.ErrorCantCopy
 	}
 	err := srcObj.readMetaData()
@@ -583,6 +609,12 @@ func (f *Fs) Copy(src fs.Object, remote string) (fs.Object, error) {
 		return nil, err
 	}
 
+	srcPath := srcObj.fs.rootSlash() + srcObj.remote
+	dstPath := f.rootSlash() + remote
+	if strings.ToLower(srcPath) == strings.ToLower(dstPath) {
+		return nil, errors.Errorf("can't copy %q -> %q as are same name when lowercase", srcPath, dstPath)
+	}
+
 	// Create temporary object
 	dstObj, leaf, directoryID, err := f.createObject(remote, srcObj.modTime, srcObj.size)
 	if err != nil {
@@ -635,6 +667,61 @@ func (f *Fs) Purge() error {
 	return f.purgeCheck("", false)
 }
 
+// Move src to this remote using server side move operations.
+//
+// This is stored with the remote path given
+//
+// It returns the destination Object and a possible error
+//
+// Will only be called if src.Fs().Name() == f.Name()
+//
+// If it isn't possible then return fs.ErrorCantMove
+func (f *Fs) Move(src fs.Object, remote string) (fs.Object, error) {
+	srcObj, ok := src.(*Object)
+	if !ok {
+		fs.Debugf(src, "Can't move - not same remote type")
+		return nil, fs.ErrorCantMove
+	}
+
+	// Create temporary object
+	dstObj, leaf, directoryID, err := f.createObject(remote, srcObj.modTime, srcObj.size)
+	if err != nil {
+		return nil, err
+	}
+
+	// Move the object
+	opts := rest.Opts{
+		Method: "PATCH",
+		Path:   "/drive/items/" + srcObj.id,
+	}
+	move := api.MoveItemRequest{
+		Name: replaceReservedChars(leaf),
+		ParentReference: &api.ItemReference{
+			ID: directoryID,
+		},
+		// We set the mod time too as it gets reset otherwise
+		FileSystemInfo: &api.FileSystemInfoFacet{
+			CreatedDateTime:      api.Timestamp(srcObj.modTime),
+			LastModifiedDateTime: api.Timestamp(srcObj.modTime),
+		},
+	}
+	var resp *http.Response
+	var info api.Item
+	err = f.pacer.Call(func() (bool, error) {
+		resp, err = f.srv.CallJSON(&opts, &move, &info)
+		return shouldRetry(resp, err)
+	})
+	if err != nil {
+		return nil, err
+	}
+
+	err = dstObj.setMetaData(&info)
+	if err != nil {
+		return nil, err
+	}
+	return dstObj, nil
+}
+
 // DirCacheFlush resets the directory cache - used in testing as an
 // optional interface
 func (f *Fs) DirCacheFlush() {
@@ -683,14 +770,17 @@ func (o *Object) Hash(t fs.HashType) (string, error) {
 func (o *Object) Size() int64 {
 	err := o.readMetaData()
 	if err != nil {
-		fs.Log(o, "Failed to read metadata: %v", err)
+		fs.Logf(o, "Failed to read metadata: %v", err)
 		return 0
 	}
 	return o.size
 }
 
 // setMetaData sets the metadata from info
-func (o *Object) setMetaData(info *api.Item) {
+func (o *Object) setMetaData(info *api.Item) (err error) {
+	if info.Folder != nil {
+		return errors.Wrapf(fs.ErrorNotAFile, "%q", o.remote)
+	}
 	o.hasMetaData = true
 	o.size = info.Size
 
@@ -713,6 +803,7 @@ func (o *Object) setMetaData(info *api.Item) {
 		o.modTime = time.Time(info.LastModifiedDateTime)
 	}
 	o.id = info.ID
+	return nil
 }
 
 // readMetaData gets the metadata if it hasn't already been fetched
@@ -722,10 +813,6 @@ func (o *Object) readMetaData() (err error) {
 	if o.hasMetaData {
 		return nil
 	}
-	// leaf, directoryID, err := o.fs.dirCache.FindPath(o.remote, false)
-	// if err != nil {
-	// 	return err
-	// }
 	info, _, err := o.fs.readMetaDataForPath(o.srvPath())
 	if err != nil {
 		if apiErr, ok := err.(*api.Error); ok {
@@ -735,8 +822,7 @@ func (o *Object) readMetaData() (err error) {
 		}
 		return err
 	}
-	o.setMetaData(info)
-	return nil
+	return o.setMetaData(info)
 }
 
 // ModTime returns the modification time of the object
@@ -747,7 +833,7 @@ func (o *Object) readMetaData() (err error) {
 func (o *Object) ModTime() time.Time {
 	err := o.readMetaData()
 	if err != nil {
-		fs.Log(o, "Failed to read metadata: %v", err)
+		fs.Logf(o, "Failed to read metadata: %v", err)
 		return time.Now()
 	}
 	return o.modTime
@@ -779,8 +865,7 @@ func (o *Object) SetModTime(modTime time.Time) error {
 	if err != nil {
 		return err
 	}
-	o.setMetaData(info)
-	return nil
+	return o.setMetaData(info)
 }
 
 // Storable returns a boolean showing whether this object storable
@@ -866,7 +951,7 @@ func (o *Object) uploadMultipart(in io.Reader, size int64) (err error) {
 	}
 
 	// Create upload session
-	fs.Debug(o, "Starting multipart upload")
+	fs.Debugf(o, "Starting multipart upload")
 	session, err := o.createUploadSession()
 	if err != nil {
 		return err
@@ -876,10 +961,10 @@ func (o *Object) uploadMultipart(in io.Reader, size int64) (err error) {
 	// Cancel the session if something went wrong
 	defer func() {
 		if err != nil {
-			fs.Debug(o, "Cancelling multipart upload")
+			fs.Debugf(o, "Cancelling multipart upload")
 			cancelErr := o.cancelUploadSession(uploadURL)
 			if cancelErr != nil {
-				fs.Log(o, "Failed to cancel multipart upload: %v", err)
+				fs.Logf(o, "Failed to cancel multipart upload: %v", err)
 			}
 		}
 	}()
@@ -898,7 +983,7 @@ func (o *Object) uploadMultipart(in io.Reader, size int64) (err error) {
 		if err != nil {
 			return err
 		}
-		fs.Debug(o, "Uploading segment %d/%d size %d", position, size, n)
+		fs.Debugf(o, "Uploading segment %d/%d size %d", position, size, n)
 		err = o.uploadFragment(uploadURL, position, size, buf)
 		if err != nil {
 			return err
@@ -914,6 +999,9 @@ func (o *Object) uploadMultipart(in io.Reader, size int64) (err error) {
 //
 // The new object may have been created if an error is returned
 func (o *Object) Update(in io.Reader, src fs.ObjectInfo) (err error) {
+	o.fs.tokenRenewer.Start()
+	defer o.fs.tokenRenewer.Stop()
+
 	size := src.Size()
 	modTime := src.ModTime()
 
@@ -926,6 +1014,11 @@ func (o *Object) Update(in io.Reader, src fs.ObjectInfo) (err error) {
 			Path:   "/drive/root:/" + url.QueryEscape(o.srvPath()) + ":/content",
 			Body:   in,
 		}
+		// for go1.8 (see release notes) we must nil the Body if we want a
+		// "Content-Length: 0" header which onedrive requires for all files.
+		if size == 0 {
+			opts.Body = nil
+		}
 		err = o.fs.pacer.CallNoRetry(func() (bool, error) {
 			resp, err = o.fs.srv.CallJSON(&opts, nil, &info)
 			return shouldRetry(resp, err)
@@ -933,20 +1026,19 @@ func (o *Object) Update(in io.Reader, src fs.ObjectInfo) (err error) {
 		if err != nil {
 			return err
 		}
-		o.setMetaData(info)
+		err = o.setMetaData(info)
 	} else {
 		err = o.uploadMultipart(in, size)
-		if err != nil {
-			return err
-		}
+	}
+	if err != nil {
+		return err
 	}
 	// Set the mod time now and read metadata
 	info, err = o.setModTime(modTime)
 	if err != nil {
 		return err
 	}
-	o.setMetaData(info)
-	return nil
+	return o.setMetaData(info)
 }
 
 // Remove an object
@@ -964,7 +1056,7 @@ var (
 	_ fs.Fs     = (*Fs)(nil)
 	_ fs.Purger = (*Fs)(nil)
 	_ fs.Copier = (*Fs)(nil)
-	// _ fs.Mover    = (*Fs)(nil)
+	_ fs.Mover  = (*Fs)(nil)
 	// _ fs.DirMover = (*Fs)(nil)
 	_ fs.DirCacheFlusher = (*Fs)(nil)
 	_ fs.Object          = (*Object)(nil)
