@@ -36,7 +36,6 @@ import (
 	"github.com/ncw/rclone/fs"
 	"github.com/ncw/swift"
 	"github.com/pkg/errors"
-	"github.com/spf13/pflag"
 )
 
 // Register with Fs
@@ -76,6 +75,9 @@ func init() {
 			}, {
 				Value: "us-west-1",
 				Help:  "US West (Northern California) Region\nNeeds location constraint us-west-1.",
+			}, {
+				Value: "ca-central-1",
+				Help:  "Canada (Central) Region\nNeeds location constraint ca-central-1.",
 			}, {
 				Value: "eu-west-1",
 				Help:  "EU (Ireland) Region Region\nNeeds location constraint EU or eu-west-1.",
@@ -122,6 +124,9 @@ func init() {
 			}, {
 				Value: "us-west-1",
 				Help:  "US West (Northern California) Region.",
+			}, {
+				Value: "ca-central-1",
+				Help:  "Canada (Central) Region.",
 			}, {
 				Value: "eu-west-1",
 				Help:  "EU (Ireland) Region.",
@@ -210,18 +215,19 @@ const (
 // Globals
 var (
 	// Flags
-	s3ACL          = pflag.StringP("s3-acl", "", "", "Canned ACL used when creating buckets and/or storing objects in S3")
-	s3StorageClass = pflag.StringP("s3-storage-class", "", "", "Storage class to use when uploading S3 objects (STANDARD|REDUCED_REDUNDANCY|STANDARD_IA)")
+	s3ACL          = fs.StringP("s3-acl", "", "", "Canned ACL used when creating buckets and/or storing objects in S3")
+	s3StorageClass = fs.StringP("s3-storage-class", "", "", "Storage class to use when uploading S3 objects (STANDARD|REDUCED_REDUNDANCY|STANDARD_IA)")
 )
 
 // Fs represents a remote s3 server
 type Fs struct {
 	name               string           // the name of the remote
+	root               string           // root of the bucket - ignore all objects above this
+	features           *fs.Features     // optional features
 	c                  *s3.S3           // the connection to the s3 server
 	ses                *session.Session // the s3 session
 	bucket             string           // the bucket we are working on
 	acl                string           // ACL for new buckets / objects
-	root               string           // root of the bucket - ignore all objects above this
 	locationConstraint string           // location constraint of new buckets
 	sse                string           // the type of server-side encryption
 	storageClass       string           // storage class
@@ -265,6 +271,11 @@ func (f *Fs) String() string {
 	return fmt.Sprintf("S3 bucket %s path %s", f.bucket, f.root)
 }
 
+// Features returns the optional features of this Fs
+func (f *Fs) Features() *fs.Features {
+	return f.features
+}
+
 // Pattern to match a s3 path
 var matcher = regexp.MustCompile(`^([^/]*)(.*)$`)
 
@@ -284,8 +295,8 @@ func s3ParsePath(path string) (bucket, directory string, err error) {
 func s3Connection(name string) (*s3.S3, *session.Session, error) {
 	// Make the auth
 	v := credentials.Value{
-		AccessKeyID:     fs.ConfigFile.MustValue(name, "access_key_id"),
-		SecretAccessKey: fs.ConfigFile.MustValue(name, "secret_access_key"),
+		AccessKeyID:     fs.ConfigFileGet(name, "access_key_id"),
+		SecretAccessKey: fs.ConfigFileGet(name, "secret_access_key"),
 	}
 
 	// first provider to supply a credential set "wins"
@@ -308,7 +319,7 @@ func s3Connection(name string) (*s3.S3, *session.Session, error) {
 	cred := credentials.NewChainCredentials(providers)
 
 	switch {
-	case fs.ConfigFile.MustBool(name, "env_auth", false):
+	case fs.ConfigFileGetBool(name, "env_auth", false):
 		// No need for empty checks if "env_auth" is true
 	case v.AccessKeyID == "" && v.SecretAccessKey == "":
 		// if no access key/secret and iam is explicitly disabled then fall back to anon interaction
@@ -319,8 +330,8 @@ func s3Connection(name string) (*s3.S3, *session.Session, error) {
 		return nil, nil, errors.New("secret_access_key not found")
 	}
 
-	endpoint := fs.ConfigFile.MustValue(name, "endpoint")
-	region := fs.ConfigFile.MustValue(name, "region")
+	endpoint := fs.ConfigFileGet(name, "endpoint")
+	region := fs.ConfigFileGet(name, "region")
 	if region == "" && endpoint == "" {
 		endpoint = "https://s3.amazonaws.com/"
 	}
@@ -338,7 +349,7 @@ func s3Connection(name string) (*s3.S3, *session.Session, error) {
 	ses := session.New()
 	c := s3.New(ses, awsConfig)
 	if region == "other-v2-signature" {
-		fs.Debug(name, "Using v2 auth")
+		fs.Debugf(name, "Using v2 auth")
 		signer := func(req *request.Request) {
 			// Ignore AnonymousCredentials object
 			if req.Config.Credentials == credentials.AnonymousCredentials {
@@ -368,12 +379,13 @@ func NewFs(name, root string) (fs.Fs, error) {
 		c:                  c,
 		bucket:             bucket,
 		ses:                ses,
-		acl:                fs.ConfigFile.MustValue(name, "acl"),
+		acl:                fs.ConfigFileGet(name, "acl"),
 		root:               directory,
-		locationConstraint: fs.ConfigFile.MustValue(name, "location_constraint"),
-		sse:                fs.ConfigFile.MustValue(name, "server_side_encryption"),
-		storageClass:       fs.ConfigFile.MustValue(name, "storage_class"),
+		locationConstraint: fs.ConfigFileGet(name, "location_constraint"),
+		sse:                fs.ConfigFileGet(name, "server_side_encryption"),
+		storageClass:       fs.ConfigFileGet(name, "storage_class"),
 	}
+	f.features = (&fs.Features{ReadMimeType: true, WriteMimeType: true}).Fill(f)
 	if *s3ACL != "" {
 		f.acl = *s3ACL
 	}
@@ -414,7 +426,7 @@ func (f *Fs) newObjectWithInfo(remote string, info *s3.Object) (fs.Object, error
 	if info != nil {
 		// Set info but not meta
 		if info.LastModified == nil {
-			fs.Log(o, "Failed to read last modified")
+			fs.Logf(o, "Failed to read last modified")
 			o.lastModified = time.Now()
 		} else {
 			o.lastModified = *info.LastModified
@@ -476,12 +488,12 @@ func (f *Fs) list(dir string, level int, fn listFn) error {
 		if level == 1 {
 			for _, commonPrefix := range resp.CommonPrefixes {
 				if commonPrefix.Prefix == nil {
-					fs.Log(f, "Nil common prefix received")
+					fs.Logf(f, "Nil common prefix received")
 					continue
 				}
 				remote := *commonPrefix.Prefix
 				if !strings.HasPrefix(remote, f.root) {
-					fs.Log(f, "Odd name received %q", remote)
+					fs.Logf(f, "Odd name received %q", remote)
 					continue
 				}
 				remote = remote[rootLength:]
@@ -497,7 +509,7 @@ func (f *Fs) list(dir string, level int, fn listFn) error {
 		for _, object := range resp.Contents {
 			key := aws.StringValue(object.Key)
 			if !strings.HasPrefix(key, f.root) {
-				fs.Log(f, "Odd name received %q", key)
+				fs.Logf(f, "Odd name received %q", key)
 				continue
 			}
 			remote := key[rootLength:]
@@ -685,7 +697,7 @@ func (f *Fs) Precision() time.Duration {
 func (f *Fs) Copy(src fs.Object, remote string) (fs.Object, error) {
 	srcObj, ok := src.(*Object)
 	if !ok {
-		fs.Debug(src, "Can't copy - not same remote type")
+		fs.Debugf(src, "Can't copy - not same remote type")
 		return nil, fs.ErrorCantCopy
 	}
 	srcFs := srcObj.fs
@@ -739,7 +751,7 @@ func (o *Object) Hash(t fs.HashType) (string, error) {
 	etag := strings.Trim(strings.ToLower(o.etag), `"`)
 	// Check the etag is a valid md5sum
 	if !matchMd5.MatchString(etag) {
-		// fs.Debug(o, "Invalid md5sum (probably multipart uploaded) - ignoring: %q", etag)
+		// fs.Debugf(o, "Invalid md5sum (probably multipart uploaded) - ignoring: %q", etag)
 		return "", nil
 	}
 	return etag, nil
@@ -781,7 +793,7 @@ func (o *Object) readMetaData() (err error) {
 	o.bytes = size
 	o.meta = resp.Metadata
 	if resp.LastModified == nil {
-		fs.Log(o, "Failed to read last modified from HEAD: %v", err)
+		fs.Logf(o, "Failed to read last modified from HEAD: %v", err)
 		o.lastModified = time.Now()
 	} else {
 		o.lastModified = *resp.LastModified
@@ -797,18 +809,18 @@ func (o *Object) readMetaData() (err error) {
 func (o *Object) ModTime() time.Time {
 	err := o.readMetaData()
 	if err != nil {
-		fs.Log(o, "Failed to read metadata: %v", err)
+		fs.Logf(o, "Failed to read metadata: %v", err)
 		return time.Now()
 	}
 	// read mtime out of metadata if available
 	d, ok := o.meta[metaMtime]
 	if !ok || d == nil {
-		// fs.Debug(o, "No metadata")
+		// fs.Debugf(o, "No metadata")
 		return o.lastModified
 	}
 	modTime, err := swift.FloatStringToTime(*d)
 	if err != nil {
-		fs.Log(o, "Failed to read mtime from object: %v", err)
+		fs.Logf(o, "Failed to read mtime from object: %v", err)
 		return o.lastModified
 	}
 	return modTime
@@ -823,7 +835,7 @@ func (o *Object) SetModTime(modTime time.Time) error {
 	o.meta[metaMtime] = aws.String(swift.TimeToFloatString(modTime))
 
 	if o.bytes >= maxSizeForCopy {
-		fs.Debug(o, "SetModTime is unsupported for objects bigger than %v bytes", fs.SizeSuffix(maxSizeForCopy))
+		fs.Debugf(o, "SetModTime is unsupported for objects bigger than %v bytes", fs.SizeSuffix(maxSizeForCopy))
 		return nil
 	}
 
@@ -866,7 +878,7 @@ func (o *Object) Open(options ...fs.OpenOption) (in io.ReadCloser, err error) {
 			req.Range = &value
 		default:
 			if option.Mandatory() {
-				fs.Log(o, "Unsupported mandatory option: %v", option)
+				fs.Logf(o, "Unsupported mandatory option: %v", option)
 			}
 		}
 	}
@@ -945,7 +957,7 @@ func (o *Object) Remove() error {
 func (o *Object) MimeType() string {
 	err := o.readMetaData()
 	if err != nil {
-		fs.Log(o, "Failed to read metadata: %v", err)
+		fs.Logf(o, "Failed to read metadata: %v", err)
 		return ""
 	}
 	return o.mimeType
