@@ -14,7 +14,6 @@ import (
 	"github.com/ncw/rclone/fs"
 	"github.com/ncw/swift"
 	"github.com/pkg/errors"
-	"github.com/spf13/pflag"
 )
 
 // Constants
@@ -83,16 +82,17 @@ func init() {
 		},
 	})
 	// snet     = flag.Bool("swift-snet", false, "Use internal service network") // FIXME not implemented
-	pflag.VarP(&chunkSize, "swift-chunk-size", "", "Above this size files will be chunked into a _segments container.")
+	fs.VarP(&chunkSize, "swift-chunk-size", "", "Above this size files will be chunked into a _segments container.")
 }
 
 // Fs represents a remote swift server
 type Fs struct {
 	name              string            // name of this remote
+	root              string            // the path we are working on if any
+	features          *fs.Features      // optional features
 	c                 *swift.Connection // the connection to the swift server
 	container         string            // the container we are working on
 	segmentsContainer string            // container to store the segments (if any) in
-	root              string            // the path we are working on if any
 }
 
 // Object describes a swift object
@@ -128,6 +128,11 @@ func (f *Fs) String() string {
 	return fmt.Sprintf("Swift container %s path %s", f.container, f.root)
 }
 
+// Features returns the optional features of this Fs
+func (f *Fs) Features() *fs.Features {
+	return f.features
+}
+
 // Pattern to match a swift path
 var matcher = regexp.MustCompile(`^([^/]*)(.*)$`)
 
@@ -145,15 +150,15 @@ func parsePath(path string) (container, directory string, err error) {
 
 // swiftConnection makes a connection to swift
 func swiftConnection(name string) (*swift.Connection, error) {
-	userName := fs.ConfigFile.MustValue(name, "user")
+	userName := fs.ConfigFileGet(name, "user")
 	if userName == "" {
 		return nil, errors.New("user not found")
 	}
-	apiKey := fs.ConfigFile.MustValue(name, "key")
+	apiKey := fs.ConfigFileGet(name, "key")
 	if apiKey == "" {
 		return nil, errors.New("key not found")
 	}
-	authURL := fs.ConfigFile.MustValue(name, "auth")
+	authURL := fs.ConfigFileGet(name, "auth")
 	if authURL == "" {
 		return nil, errors.New("auth not found")
 	}
@@ -161,11 +166,11 @@ func swiftConnection(name string) (*swift.Connection, error) {
 		UserName:       userName,
 		ApiKey:         apiKey,
 		AuthUrl:        authURL,
-		AuthVersion:    fs.ConfigFile.MustInt(name, "auth_version", 0),
-		Tenant:         fs.ConfigFile.MustValue(name, "tenant"),
-		Region:         fs.ConfigFile.MustValue(name, "region"),
-		Domain:         fs.ConfigFile.MustValue(name, "domain"),
-		TenantDomain:   fs.ConfigFile.MustValue(name, "tenant_domain"),
+		AuthVersion:    fs.ConfigFileGetInt(name, "auth_version", 0),
+		Tenant:         fs.ConfigFileGet(name, "tenant"),
+		Region:         fs.ConfigFileGet(name, "region"),
+		Domain:         fs.ConfigFileGet(name, "domain"),
+		TenantDomain:   fs.ConfigFileGet(name, "tenant_domain"),
 		ConnectTimeout: 10 * fs.Config.ConnectTimeout, // Use the timeouts in the transport
 		Timeout:        10 * fs.Config.Timeout,        // Use the timeouts in the transport
 		Transport:      fs.Config.Transport(),
@@ -191,8 +196,9 @@ func NewFsWithConnection(name, root string, c *swift.Connection) (fs.Fs, error)
 		segmentsContainer: container + "_segments",
 		root:              directory,
 	}
+	f.features = (&fs.Features{ReadMimeType: true, WriteMimeType: true}).Fill(f)
 	// StorageURL overloading
-	storageURL := fs.ConfigFile.MustValue(name, "storage_url")
+	storageURL := fs.ConfigFileGet(name, "storage_url")
 	if storageURL != "" {
 		f.c.StorageUrl = storageURL
 		f.c.Auth = newAuth(f.c.Auth, storageURL)
@@ -295,7 +301,7 @@ func (f *Fs) listContainerRoot(container, root string, dir string, level int, fn
 					}
 				}
 				if !strings.HasPrefix(object.Name, root) {
-					fs.Log(f, "Odd name received %q", object.Name)
+					fs.Logf(f, "Odd name received %q", object.Name)
 					continue
 				}
 				remote := object.Name[rootLength:]
@@ -478,7 +484,7 @@ func (f *Fs) Purge() error {
 func (f *Fs) Copy(src fs.Object, remote string) (fs.Object, error) {
 	srcObj, ok := src.(*Object)
 	if !ok {
-		fs.Debug(src, "Can't copy - not same remote type")
+		fs.Debugf(src, "Can't copy - not same remote type")
 		return nil, fs.ErrorCantCopy
 	}
 	srcFs := srcObj.fs
@@ -528,7 +534,7 @@ func (o *Object) Hash(t fs.HashType) (string, error) {
 		return "", err
 	}
 	if isDynamicLargeObject || isStaticLargeObject {
-		fs.Debug(o, "Returning empty Md5sum for swift large object")
+		fs.Debugf(o, "Returning empty Md5sum for swift large object")
 		return "", nil
 	}
 	return strings.ToLower(o.info.Hash), nil
@@ -592,12 +598,12 @@ func (o *Object) readMetaData() (err error) {
 func (o *Object) ModTime() time.Time {
 	err := o.readMetaData()
 	if err != nil {
-		fs.Debug(o, "Failed to read metadata: %s", err)
+		fs.Debugf(o, "Failed to read metadata: %s", err)
 		return o.info.LastModified
 	}
 	modTime, err := o.headers.ObjectMetadata().GetModTime()
 	if err != nil {
-		// fs.Log(o, "Failed to read mtime from object: %v", err)
+		// fs.Logf(o, "Failed to read mtime from object: %v", err)
 		return o.info.LastModified
 	}
 	return modTime
@@ -658,11 +664,11 @@ func (o *Object) removeSegments(except string) error {
 			return nil
 		}
 		if except != "" && strings.HasPrefix(remote, except) {
-			// fs.Debug(o, "Ignoring current segment file %q in container %q", segmentsRoot+remote, o.fs.segmentsContainer)
+			// fs.Debugf(o, "Ignoring current segment file %q in container %q", segmentsRoot+remote, o.fs.segmentsContainer)
 			return nil
 		}
 		segmentPath := segmentsRoot + remote
-		fs.Debug(o, "Removing segment file %q in container %q", segmentPath, o.fs.segmentsContainer)
+		fs.Debugf(o, "Removing segment file %q in container %q", segmentPath, o.fs.segmentsContainer)
 		return o.fs.c.ObjectDelete(o.fs.segmentsContainer, segmentPath)
 	})
 	if err != nil {
@@ -671,7 +677,7 @@ func (o *Object) removeSegments(except string) error {
 	// remove the segments container if empty, ignore errors
 	err = o.fs.c.ContainerDelete(o.fs.segmentsContainer)
 	if err == nil {
-		fs.Debug(o, "Removed empty container %q", o.fs.segmentsContainer)
+		fs.Debugf(o, "Removed empty container %q", o.fs.segmentsContainer)
 	}
 	return nil
 }
@@ -711,7 +717,7 @@ func (o *Object) updateChunks(in io.Reader, headers swift.Headers, size int64, c
 		headers["Content-Length"] = strconv.FormatInt(n, 10) // set Content-Length as we know it
 		segmentReader := io.LimitReader(in, n)
 		segmentPath := fmt.Sprintf("%s/%08d", segmentsPath, i)
-		fs.Debug(o, "Uploading segment file %q into %q", segmentPath, o.fs.segmentsContainer)
+		fs.Debugf(o, "Uploading segment file %q into %q", segmentPath, o.fs.segmentsContainer)
 		_, err := o.fs.c.ObjectPut(o.fs.segmentsContainer, segmentPath, segmentReader, true, "", "", headers)
 		if err != nil {
 			return "", err
@@ -764,7 +770,7 @@ func (o *Object) Update(in io.Reader, src fs.ObjectInfo) error {
 	if isDynamicLargeObject {
 		err = o.removeSegments(uniquePrefix)
 		if err != nil {
-			fs.Log(o, "Failed to remove old segments - carrying on with upload: %v", err)
+			fs.Logf(o, "Failed to remove old segments - carrying on with upload: %v", err)
 		}
 	}
 
